$eject
;===========================
;       MAC65 equates
;===========================

;       master & key equates
;       --------------------
atari   equ     1       ; machine flag
apple   equ     !atari

oss?    equ     1       ; in house version
cbug    equ     0       ; cartridge debug (banks are for $C000 for Mosaic)
ddt?    equ     1       ; assemble with DDT
ram     equ     1       ; for ram cart assembly
eprom   equ     0       ; for eprom burning

@@base  equ     $3000   ; ram assembly addr
sorcim  equ     1       ; sorcim locals
pseude  equ     0       ; drive #3
toggle  equ     0       ; toggle list dev
dolocal equ     0       ; asm only locals

; for rom assembly (cbug=0 and @@base=$3000)

;            ______________________________
;object addr | $3000  $4000 | $5000  $6000 |
;actual addr | $a000  $b000 | $a000  $a000 |
;group/bank  | bankc   main | banka  bankb |
;            |  - - - - - - |  - - - - - - |
;            |  master chip |  slave chip  |
;            ______________________________

rom     equ     $b000
        $if cbug
romf    equ       $c000
bank1   equ       $ffc0
bank2   equ       $ffc1
bank3   equ       $ffc2
        $else
romf    equ       $a000
bank1   equ       $d500
bank2   equ       $d509
bank3   equ       $d501
ramall  equ       $d508
        $endif


; Memory areas
; ------------
fr0     equ     $d4
fr1     equ     $e0
raz1    equ     $80
raz3    equ     $fb
zfp     equ     $d2
cix     equ     $f2             ; input text index
inbuff  equ     $f3             ; indirect to lbuff
ztemp4  equ     $f7             ; syntax hex temp +2
ztemp3  equ     $f9             ; used ascii to f.p.
xrefsz  equ     ztemp3          ; asm x-ref total size

        $if oss?
tpobuf  equ       $0600         ; asm obj buf +128
        $endif                  ; & syntax text use

ram2    equ     $0480           ; non-zpage usage
ram3    equ     $0500           ; secondary non-zpage
lbuff   equ     $0580

;       syntax stack +128
;       -----------------
six     equ     ram3            ; syn save cix
sox     equ     six+1           ; syn save cox
spc     equ     six+2           ; syn save pc +2


; assembler buffer allocation buffer size must be in multiples of pages.

bufsize equ     6*256           ; set 1.25k


;       z-page memory management ptrs
;       -----------------------------
        seg     absolute
        org     raz1
lomem                   ; token buff
exbuf   ds      2       ; execution buff

objbuf                  ; asm obj buff
repbuf  ds      2       ; rep string buff

txttab  ds      2       ; start of txt
stmcur  ds      2       ; cur line ptr

txtend                  ; text end
asmram  ds      2       ; assembly buffers

stack   ds      2       ; parameter stack
macstrt ds      2       ; macro def

macend                  ; macro end
smbltp  ds      2       ; symbol tab start

smblend                 ; symbol tab end
memtop  ds      2       ; top of used mem


;       routine & table pointers
;       ------------------------
cox     ds      1       ; cur exec buf indx

srcadr                  ; search adr
msgadr  ds      2       ; pr string adr

svesa   ds      2       ; mangr sv strt adr

labpt                   ; asm ptr: = & list
saxpc                   ; syntax pc
rfpgn   ds      2       ; x-ref page # count

objix                   ; asm temp obj ptr
saxtx   ds      1       ; syntax valid ptr

usadr                   ; utility adr ptr
srlnum                  ; sv line # for srch
labof   ds      1       ; lab handler use

augx    ds      1       ; macrox argument temp

mvsiz                   ; mangr move size
slsz    ds      2       ; save/load file sz


;       working variables
;       -----------------
stklvl                  ; stntax stack ptr
mcrsk   ds      1       ; macrox stack ptr

skipfc                  ; search skip indx
modfybyt                ; byte modifier value
sverr   ds      1       ; save error code

lelnm                   ; list end line# +2
ueadr                   ; utility end adr +2
link                    ; x-ref test link +2
res     ds      1       ; asm exp resolved

curaug  ds      1       ; macrox para ptr

; next group must stay together
; * - these locations are poison

eaf     ds      1       ; edit/pass # flg*
nmfg    ds      1       ; auto flg
epf                     ; edit/rep flg*
rfpagf  ds      1       ; x-ref flags (bits)

eix     ds      1       ; err indx
entdv   ds      1       ; enter dev
lstdv   ds      1       ; list dev
refdv   ds      1       ; x-ref dev
supr    ds      1       ; supres list l. #

group   equ     $-eaf-1 ; locs in group

txtflag ds      1       ; text mode flag


;       misc locations
;       --------------
bcode                   ; asm base code
dirfg                   ; asm which directve
numi    ds      1       ; auto inc +2

bfld                    ; asm fld ofset
tem                     ; macro def temp
tem0    ds      1       ; macrox temp

asmpc                   ; asm pc +2
condptr ds      1       ; syn cond text

actptr                  ; syn actual text
hdcix   ds      1       ; temp for cix

zt1                     ; asm object list
mode                    ; asm multi adr typ
maxsort                 ; sort largest len
temp    ds      1       ; local temp

curtoke                 ; syn cur tst toke
ltemp                   ; asm lab process
cmdlg   ds      1       ; internal line len

rps1                    ; rep 1st str len
rstk    ds      1       ; conditional stk ptr

rpsp    ds      1       ; rep str ptr, list

rpll                    ; rep inbuf str len
svop                    ; ex exp sv opertor
torg    ds      1       ; disasm test origin +2

rpop                    ; rep old line ptr
svnp    ds      1       ; ex exp sv on prec

rpcl                    ; rep cur line ptr
inmac   ds      1       ; macrox in-process

ztab1   ds      1       ; z-page tab list*
ztab2   ds      1       ; z-page tab list*


;       do not relocate the next three pointers!!!
;       ------------------------------------------
                        ; bload header +4
buffer  ds      2       ; asm sliding buffer pointer
bytcnt  ds      2       ; asm byte counter for this read
filesize
        ds      2       ; asm source file size

objptr  ds      1       ; asm obj buf indx
xrefbufi
        ds      2       ; x-ref pointer
real    ds      1       ; asm master pass #
sxref   ds      1       ; save x-ref option

xrindx                  ; 'xrefbuf' index
rpoc    ds      1       ; rep occurance
ionum   ds      1       ; save iocb #
ciocmd  ds      1       ; save cio command
asmlino ds      2       ; asm current lineo +2

rpse                    ; rep 1st str end
localix ds      1       ; local label counter

rpfd                    ; rep or find flg
pcdef   ds      1       ; asm pc def flg
eofflag ds      1       ; asm eof flag

rpif                    ; rep inbuff flag
linecnt ds      1       ; lines/page cntr
column  ds      1       ; list column count
pagenum ds      2       ; page cntr +2

system  ds      1       ; internal sys flags
                        ;   bit 7:  find/rep start at label
                        ;           colunm (=0) or at first colunm
                        ;           after line #.
                        ;   bit 6:  rep re-list of line
                        ;           after replace (=0) is list.
                        ;   bit 5:  local labels appear
                        ;           in symbol table printout.

        org     zfp-2
svn2    ds      1       ; syntax use
svf2    ds      1       ; syntax use
rps2                    ; rep 2nd str len
mpf     ds      1       ; asm pass 2/3 flag
objdv   ds      1       ; object device

        org     fr0
many    ds      2       ; asm cio byte cnt+1
fwdptr                  ; sort foward ptr
lablst  ds      1       ; list label start
lablch  ds      1       ; list label index
bckptr  ds      2       ; sort back ptr

; *** start of "fre"
        ds      1               ; unused- not safe

svcpc                   ; pass 3 sv pc +2
xrfchar                 ; x-ref list char
xrfrdo  equ     $+1     ; x-ref single ref flag
        ds      2

svstm                   ; pass 3 sv stmt
lastl   ds      2       ; x-ref last line #
errors  ds      1       ; asm error count

        org     fr1
swapsort                ; sort exc flg
lablsz  ds      1       ; list label size

hilosort                ; sort len flg
lablcc  ds      1       ; list label column
sindex  ds      1       ; sort/list char index
topstk  ds      1       ; ex exp push flag

labelptr                ; sort/list lab ptr +2
aptr    ds      1       ; ex exp oper ptr
optr    ds      1       ; ex exp argument ptr

;  *** start of "fr2"

mvfa                    ; mem mgr mv fr
sltmp   ds      2       ; save/load temp
mvta                    ; mem mgr mv to
ktmp    ds      2       ; "pdum" iocb indx +1


mvlgn                   ; mangr move len
amvta                   ; asm io move to addr (temp)
zte4    ds      2       ; list decimal # temp

;  *** start of "frx"

buffers ds      2       ; asm io buffers ptr
strings ds      2       ; asm string space ptr
bufferx ds      2       ; x-ref buffer ptr

        org     raz3
klist   ds      1       ; device master list
optflag ds      1       ; asm option flg
ttix    ds      1       ; list exbuf index

tobjdex                 ; asm byte list
mnindx  ds      1       ; disasm mnemonic index
commode ds      1       ; asm temp (.byte)

        org     ram2
xcii    ds      2       ; os/a+ system use
hcntrlc ds      1       ; control-c abort char
colors  ds      2       ; save system color
errstack
        ds      3       ; error stack +3
svlomem ds      2       ; save lomem
tab1    ds      1       ; list tab setings
tab2    ds      1       ; x
tab3    ds      1       ; x
pointer ds      3       ; asm 'point' +3
oldentdv
        ds      1       ; .include save device
bufftab ds      6       ; asm buffer table


; system ".set" storage
; there are 6 sets

sysset
setbyte ds      1       ; .byte & .sbyte format
leftmar ds      1       ; left margin
rightmar
        ds      1       ; right margin
topfc   ds      1       ; top of form char
formsz  ds      1       ; physical page size
cpad    ds      1       ; comment space padding
offset  ds      2       ; offset to object
sysmax  equ     $-sysset-1      ; number of sets

runstk  ds      18      ; conditional stack
skmx    equ     $-runstk        ; run stk top

; macro stack
; -----------
svcur   equ     $       ; macrox sv stmcur
svaug   equ     $+2     ; sv param ptr
svedv   equ     $+3     ; sv enter dev
        ds      68      ; stack size
topsk   equ     $-svcur ; macro stk top

savdevi ds      6       ; asm save pointers +6
spointer
        ds      56      ; label pointers +56
xrefbuf ds      32      ; x-ref line buffer
xrmax   equ     $-xrefbuf       ; top of buffer
;fz     equ     20
fz      equ     16      ; filespec buffer size          (CART ver... was 20
asmfm   ds      fz      ; main filespec for list          in steves version..)
asmfi   ds      fz      ; include filespec
        ds      0       ; where we are

; misc working equates
; --------------------
tosa    equ     $c8     ; exexpr set stack
tose    equ     $ff-tosa        ; exexpr tos
argops  equ     $0100+tosa      ; ex exp stack
strmax  equ     80      ; reserve string space
ckbyt   equ     $fe     ; save/load ck byte
@@sx    equ     $82     ; instr stx
@@sy    equ     $80     ; instr sty
@       equ     $c0     ; for syntax tables
ch      equ     9       ; syntax change
thw     equ     ch-4    ; token hex word
thb     equ     ch-3    ; token hex byte
tdw     equ     ch-2    ; token dec word
tdb     equ     ch-1    ; token dec byte
rmode   equ     $77     ; rep on mode

        $if ddt?
dmode   equ     $00     ; ddt swap flag
        $endif

eof     equ     $88     ; atari eof
eol     equ     $9b     ; atari eol
bell    equ     $fd     ; atari bell
tab     equ     $7f     ; atari tab

cntrlc  equ     3       ; control-c auto abort
linefeed equ    $0a     ; line feed
topform equ     $0c     ; top of form

qdumy   equ     0       ; nested null
qelse   equ     1       ; search endif/else
qselse  equ     2       ; else found
qendif  equ     3       ; search else
rfnum   equ     -121    ; x-ref word type
rfnub   equ     -122    ; x-ref relative
rfpag   equ     -123    ; x-ref page number
rflab   equ     -124    ; x-ref label flag
rfeof   equ     -125    ; x-ref eof

;  *** error numbers
ntpe    equ     9       ; nesting
sofe    equ     11      ; conditional stk
nesr    equ     12      ; macro nest
dupm    equ     16      ; dup macro name
lmr     equ     18      ; linked macro def
xrerr   equ     28      ; too many x-refs
hlocer  equ     29      ; too many local fields
udfm    equ     30      ; undef macro
nesm    equ     31      ; macrox nesting
nfpr    equ     33      ; floating point #


; *** cio equates
; ---------------
aphm    equ     $0e
warmflg equ     $08     ; warm start?
icdno   equ     $0341
iccom   equ     $0342
icsta   equ     $0343
icbufa  equ     $0344
icput   equ     $0346
icbufl  equ     $0348
icaux1  equ     $034a
icaux2  equ     $034b
icsec   equ     $034c
icbyt   equ     $034e
xcio    equ     $e456
ataribye equ    $e471   ; memo pad
icax1z  equ     $2a     ; z-page iocbs
icax2z  equ     $2b

; routines in floating point roms
; -------------------------------
ilshift equ     $da5a
skpspace equ    $dba1
intlbf  equ     $da51
afp     equ     $d800
ifp     equ     $d9aa
fpi     equ     $d9d2
fascc   equ     $d8e6

dosvec  equ     $0a
brkkey  equ     $11
coldst  equ     $0244
tabmap  equ     $02a3
color1  equ     $02c5
color2  equ     $02c6
mtp     equ     $02e5           ; highest available
memlow  equ     $02e7
chr     equ     $02fc
dspflg  equ     $02fe
ssflag  equ     $02ff

consol  equ     $d01f
pokey   equ     $d200
skctl   equ     $d20f


$eject
;============================
;       DDT EQUATES
;============================

; warm start location is ddtwrm
; zero page definitions

pokmsk  equ     $10
rtclok  equ     $14

; page 2 definitions
; ------------------
vbreak  equ     $0206
vkeybd  equ     $0208
vvblki  equ     $0222
vvbkih  equ     $0223
vvblkd  equ     $0224
vvbkdh  equ     $0225
sdmctl  equ     $022f
sdlstl  equ     $0230
sdlsth  equ     $0231
gprior  equ     $026f
color4  equ     $02c8
crsinh  equ     $02f0
chact   equ     $02f3
chbas   equ     $02f4


; hardware registers
; ------------------
grafp0  equ     $d00d
gractl  equ     $d01d
kbcode  equ     $d209
irqen   equ     $d20e
dmactl  equ     $d400
dlistl  equ     $d402
vcount  equ     $d40b
nmien   equ     $d40e


; os addresses
; ------------
kchact  equ     $e002   ;note - this will load chact & chbas
kvvbki  equ     $e45f   ;standard vector
kvvbkd  equ     $e462   ;standard vector


; other definitions
; -----------------
bordr   equ     $04     ;border color
bkgnd   equ     $00     ;background color
text    equ     $0a     ;text luminance
dbchr   equ     $9c     ;flash character (control esc now)

        org     lomem+2
windf   ds      1       ;init 0
zpsflg  ds      1       ;zero page swapped flag for reset
undff   ds      1
asav    ds      1
xsav    ds      1
ysav    ds      1
spsav   ds      1
psav    ds      1
pcsav   ds      2
poksav  ds      1
dindir  ds      2
dndr2   ds      2
dscnfl  ds      1
btemp   ds      2
itemp   ds      1
chcol   ds      1
examwi  ds      2
cc      ds      1
fsize   ds      1
hexval  ds      2
mprot   ds      1
tbyte   ds      1
stemp   ds      2
itbyte  ds      1
sssiz   ds      1
sstr    ds      6


        org     $03fd           ;start of cassette buffer
grsav   ds      13
brkor   ds      1
vkeysv  ds      2       ; save keyboard irq vector
crstat  ds      1
depsw   ds      1
ddtact  ds      1       ; init 83
ssflg   ds      1
dscd1   ds      284     ; display lines for window
dscd2   ds      32      ; display line for brkps.
dscd3   ds      32      ; display line for registers
brkp0   ds      15      ; init 0

        org     $05ff-10
tbed    ds      3       ; for displaced instruction
        ds      3       ; init jmp ssret
        ds      3       ; init jmp ssbc2
;========================================
;       MAC/65 Token Definitions
;========================================
; This can't change since all MAC/65 source files
; use these tokens...

        seg     absolute
        org     0
@seol   ds      1
@if     ds      1
@else   ds      1
@endif  ds      1
@macro  ds      1
@endm   ds      1
@title  ds      1
@macrox ds      1
@page   ds      1
@word   ds      1
@error  ds      1
@byte   ds      1
@sbyte  ds      1
@dbyte  ds      1
@end    ds      1
@opt    ds      1
@tab    ds      1
@include ds     1
        ds      1       ; ds
        ds      1       ; org
        ds      1       ; equ

        ds      1       ; bra
@trb    ds      1       ; trb
        ds      1       ; tsb
        ds      1       ; .float
        ds      1       ; .cbyte
        ds      1       ; ;
        ds      1       ; .local
@set    ds      1       ; .set
@asteq  ds      1       ; *=
@equ    ds      1       ; =
@.equ   ds      1       ; .=

;       mneumonics
;       ----------
@jsr    ds      1
@jmp    ds      1
@dec    ds      1
@inc    ds      1

@ldx    ds      1
@ldy    ds      1
@stx    ds      1
@sty    ds      1
@cpx    ds      1
@cpy    ds      1
@bit    ds      1

@brk    ds      1
@clc    ds      1
@cld    ds      1
@cli    ds      1
@clv    ds      1
@dex    ds      1
@dey    ds      1

@inx    ds      1
@iny    ds      1
@nop    ds      1
@pha    ds      1
@php    ds      1
@pla    ds      1
@plp    ds      1

@rti    ds      1
@rts    ds      1
@sec    ds      1
@sed    ds      1
@sei    ds      1
@tax    ds      1
@tay    ds      1

@tsx    ds      1
@txa    ds      1
@txs    ds      1
@tya    ds      1
@bcc    ds      1
@bcs    ds      1
@beq    ds      1

@bmi    ds      1
@bne    ds      1
@bpl    ds      1
@bvc    ds      1
@bvs    ds      1

@ora    ds      1
@iand   ds      1
@eor    ds      1
@adc    ds      1
@sta    ds      1
@lda    ds      1
@cmp    ds      1

@sbc    ds      1
@asl    ds      1
@rol    ds      1
@lsr    ds      1
@ror    ds      1

@scomment
        ds      1       ; statement comment
@rockwell
        ds      1
        ds      1       ; stz
@rimpl  ds      2       ; dea/ina
        ds      2       ; phx/phy
        ds      2       ; plx/ply


;       Functions
;       ---------
        org     (ch+2) & $FE
@apost  ds      1       ; '
@mstr   ds      1       ; %$
@percent ds     1       ; %
@cpc    ds      1       ; *
        ds      2       ; 2 filler spaces
s.a     ds      1       ; A
s.q     ds      1       ; Q
valuend

@plus   ds      1       ; +
@minus  ds      1       ; -
@times  ds      1       ; *
@div    ds      1       ; /
@and    ds      1       ; &
@def    ds      1       ; .DEF
@elp    equ     @def
@cequ   ds      1       ; =
@lequ   ds      1       ; <=
@gequ   ds      1       ; >=
@nequ   ds      1       ; <>
@less   ds      1       ; >     ; crazy
@grt    ds      1       ; <     ; crazy
@umin   ds      1       ; -
@olp    ds      1       ; [
@orp    ds      1       ; ]
@bor    ds      1       ; .OR
@band   ds      1       ; .AND
@not    ds      1       ; .NOT
@or     ds      1       ; !  (or)
@xor    ds      1       ; ^  (xor)
@ref    ds      1       ; .REF
@mod    ds      1       ; \

        ds      4       ; filler
        ds      3       ; what/text/edit messages

@refl   ds      1       ; .REF
@defl   ds      1       ; .DEF
@notl   ds      1       ; .NOT
@bandl  ds      1       ;  .AND
@borl   ds      1       ;  .OR
@low    ds      1       ;  <
@high   ds      1       ;  >

@pcmax  ds      1       ; ,X)
@pcmay  ds      1       ; ),Y
@cmay   ds      1       ; ,Y
@cmax   ds      1       ; ,X
@rp     ds      1       ; )
@lcma   ds      1       ; ,
@comment equ    @lcma

@eol    ds      1
@cma    ds      1       ; ,
@pnd    ds      1       ; #
@a      ds      1       ; A
@lp     ds      1       ; (
@qu     ds      1       ; "
@doll   ds      1       ; $
@q      ds      1       ; Q
@lno    ds      1       ; NO
@no     ds      1       ; NO
@obj    ds      1       ; OBJ
@err    ds      1       ; ERR
@eject  ds      1       ; EJECT
@list   ds      1       ; LIST
@wait   ds      1       ; XREF
@mlist  ds      1       ; MLIST
@clist  ds      1       ; CLIST
@num    ds      1       ; NUM
@m      ds      1       ; M
$title(MAC65/XL Cartridge  (c) 1988 ICD Inc)
$subtitle(Cartridge main bank for MAC65)

        seg     main
;=================================
;       Process source line
;=================================

;       process text line, handle control c auto-num exit here
;       ------------------------------------------------------
ptext:  proc
        ldy     cix             ; get 1st character
        lda     (inbuff),y
        cmp     #eol            ; if eol, delete
        beq     ..dell

        sty     hdcix           ; save "cix"
        jsr     skpspace        ; is line
        lda     (inbuff),y      ; all spaces?
        cmp     hcntrlc         ; auto abort?
        bne     ..hnc           ; no
         lda    #0              ; else, stop auto
         sta    nmfg
         jmp    warm            ; go warm start

..hnc   cmp     #eol            ; null line?
        bne     ..hnsp          ; no
        lda     epf             ; replace mode?
        beq     ..nrep          ; no
..dell   jmp    dell            ; consider as delete
..nrep  jmp     noprm           ; next input

..hnsp  ldy     hdcix           ; reset "cix"
        sty     cix
        inc     cox             ; dummy for offset
        jsr     synli           ; syntax line
        ldy     #2
        dec     cox             ; pop line eol
        lda     cox             ; this is offset
        sta     (exbuf),y       ; to next line

        jsr     fndli           ; look for line
        lda     #0              ; set for not found
        bcs     ..skp           ; br: line isn't there
         jsr    getoset         ; get current length

..skp   sec                     ; find diff
        sbc     cox
        beq     samez           ; same size
        bcs     smallz          ; smaller

        jsr     exptxt          ; go expand
        lda     svesa           ; restore current
        sta     stmcur          ; line pointer
        lda     svesa+1
        sta     stmcur+1
        bne     samez           ; unc

smallz  pha                     ; set-up contract
        jsr     adoff           ; adress
        pla
        jsr     contxt          ; and contract

samez   ldy     cox             ; move new line
..mv    dey                     ; to memory
        lda     (exbuf),y
        sta     (stmcur),y
        tya
        bne     ..mv

        ldy     #3              ; check for bad line
        lda     (stmcur),y      ; non-zero
        bne     oj              ; if sucess
         sta    supr            ; do pr line #
         sta    epf             ; clr rep rts flg
         jsr    setztab         ; set zp tabs
         jsr    prli            ; print bad line

oj      bit     nmfg            ; auto?
        bpl     oy              ; no
        bvs     oy              ; special- atari quirk
         jsr    upcur           ; update next line #

oy:     lda     epf             ; are we in
        cmp     #rmode          ; replace mode?
        beq     rtrep           ; yup
         jmp    noprm           ; else, to editor

rtrep   jmp     veryr           ; to replace
        endproc



;       Get line of text from E:
;       ------------------------

input:  proc
        jsr     intlbf
inp     jsr     egli            ; get a line
        cmp     #eof            ; eof?
        beq     inp             ; yup

        lda     #0              ; clr:
        sta     cix             ; input index
        sta     saxtx           ; valid text pointer
        sta     cox             ; output index
        jsr     skpspace
        jsr     testbk          ; check break key
        beq     yline           ; no break
         ldy    #0              ; clear auto
         sty    nmfg
         dey                    ; reset flag
yline   rts
        endproc



;       Editor list and print
;       ---------------------
printx: proc
        lda     #$80            ; supress line #
         db $2c

listx:  lda     #0              ; set print line #
        sta     supr
        jsr     setztab         ; set z-tabs
                                ;  atari display
        inc     dspflg          ; control chars

        jsr     docr            ; a formality...
        jsr     gtoke
        cmp     #@pnd           ; to a device?
        bne     ..lx1           ; no

        jsr     opnout          ; handle filespec
        jsr     ..lx2           ; handle listing

closel: lda     lstdv           ; close device
        jsr     closei
zldv0   lda     #0              ; reset back
        sta     lstdv           ; to e: driver
..lx3   rts

..lx2   jsr     gtokel          ; ckeck ", lineo"
        beq     ..lx1
         jsr    gtoke           ; get next token

..lx1   jsr     lrnge           ; get s&e lineo's
..lx4   jsr     testbk          ; break abort?
        bne     ..lx3           ; yup
        jsr     lline           ; past last line?
        bcs     ..lx3           ; yes
        jsr     prli            ; print this line
        jsr     testbk          ; if middle of l.
        bne     ..lx3           ; yup
        jsr     gtadoff         ; point to next
        bcc     ..lx4           ; unc loop
        endproc


;       Line number search (find lineo in srlnum)
;       -----------------------------------------

fndli0: proc
        lda     #0
        sta     srlnum
fndli9: sta     srlnum+1

fndli:  jsr     endl?           ; at memtop?
        bcs     ..fl1           ; start at txttab

        ldy     #1              ; see if lineo >
        lda     (stmcur),y      ; last line
        cmp     srlnum+1        ; x
        bcc     ..fl2           ; less, start at stmcur
        bne     ..fl1           ; start at txttab

        dey                     ; low byte
        lda     (stmcur),y      ; x
        cmp     srlnum
        bcc     ..fl2           ; start at stmcur
        beq     ..fl5           ; at line, return

..fl1   lda     txttab+1        ; set ptr to
        ldy     txttab          ; start of
        sta     stmcur+1        ; text file
        sty     stmcur          ; x

..fl2   jsr     endl?           ; begin search
        bcs     ..fl3           ; stop if text top
        ldy     #1              ; high
        lda     (stmcur),y      ; x
        cmp     srlnum+1
        bcc     ..fl4           ; less, pt to nxt
        bne     ..fl3           ; end
        dey                     ; low
        lda     (stmcur),y      ; x
        cmp     srlnum
        bcc     ..fl4           ; less, pt to nxt
        bne     ..fl3           ; br: not found
..fl5   clc                     ; equal, set fnd
..fl3   rts

..fl4   jsr     gtadoff         ; bump stmcur
        bcc     ..fl2           ; unc
        endproc


;       get offset of stmcur in a
;       -------------------------
getoset: proc
        ldy     #2
        lda     (stmcur),y
        rts
        endproc


;       bump stmcur to next stmcur
;       --------------------------
; out:
;       cc

gtadoff: proc
        jsr     getoset         ; get offset

adoff:  clc
        adc     stmcur
        sta     stmcur
        bcc     ..ff
         inc    stmcur+1
         clc
..ff    rts
        endproc



;==================================
;       Editor - ENTER file
;       -------------------
enterx: proc
        jsr     hfile           ; handle filespec
        jsr     gtokel          ; merge?
        beq     ..ncl           ; yup, clear & boot

        jsr     gtoke           ; get enter option
        cmp     #@m             ; check merge first
        beq     ..ncn           ; yup, merge

        jsr     enttxt          ; clear memory
        lda     entdv           ; set "special" enter
        ora     #$80
        sta     entdv
        jsr     numx            ; do auto-number
        jmp     noprm           ; go get input

..ncl   jsr     enttxt          ; clear memory
..ncn   jmp     noprm           ; go for input
        endproc


;===============================
;       Editor - LOAD file
;       ------------------
loadx:  proc
        jsr     hfile           ; handl filespc
        jsr     gtokel          ; load with append?
        bne     loapp           ; yup
         jsr    enttxt          ; no, clear memory

loapp   jsr     gobj            ; pop check byte
        cmp     #ckbyt          ; save file?
        bne     ..loabt         ; no, abort it!
        jsr     gobj            ; next check byte
        cmp     #ckbyt          ; ok?
        beq     ..loap1         ; yup

..loabt lda     #200+23         ; pseudo cio err
ero     jmp     badst


..loap1 jsr     gobj            ; get file size
        sta     slsz
        jsr     gobj
        sta     slsz+1

        clc                     ; calculate new
        lda     txtend          ; mem top and ck
        adc     slsz            ; for overflow
        sta     sltmp
        tay
        lda     txtend+1
        adc     slsz+1
        bcs     er24            ; file too big!!!
         sta    sltmp+1

        cmp     mtp+1           ; ckeck for val>memtop
        bcc     loalp
        bne     er24            ; oops
        cpy     mtp
        bcc     loalp
        beq     loalp

er24    lda     #24+200         ; load too big
        bne     ero

loalp   ldx     entdv           ; get enter dev
        lda     #7              ; get chars cmd
        jsr     blkio1          ; set some cio

        lda     txtend+1        ; set "load at"
        ldy     txtend          ; adress
        jsr     io4             ; go get 'em
        cpy     #$80            ; fix: test for break
        beq     loabrk          ; don't set memtop
        jsr     getstat         ; get status

        lda     sltmp           ; get new mem top
        ldy     sltmp+1
        ldx     #txtend         ; set rest of
        jsr     cs1             ; z-page pointers

loabrk  lda     entdv           ; close dev
        jsr     closei
        lda     #0              ; reset enter device
        sta     entdv           ; to e:
        rts                     ; return
        endproc


;================================
;       Editor - BLOAD file
;       -------------------
bloadx: proc
        jsr     hfile           ; handl filespc
        jsr     gobj            ; pop ck bytes
        cmp     #$ff            ; bsave?
        bne     ..nbsv          ; no
        jsr     gobj            ; x
        cmp     #$ff            ; bsave?
        bne     ..nbsv          ; no

..top   lda     #7              ; get binary cmd
        jsr     glcxent         ; set-up
        lda     #4              ; get header
        sta     icbufl,x
        lda     #0
        sta     icbufl+1,x
        ldy     #low buffer     ; AY = address of buffer
        jsr     io4.s           ; do it!

        sec                     ; Now load the bytes
        lda     buffer+2        ; end addr -
        sbc     buffer          ; start addr
        sta     icbufl,x        ; to iocb
        lda     buffer+3
        sbc     buffer+1
        sta     icbufl+1,x
        inc     icbufl,x
        bne     ..b00
         inc    icbufl+1,x

..b00   ldy     buffer          ; the dest addr
        lda     buffer+1
        jsr     io4
        cpy     #1              ; only perfect status
        beq     ..top           ; still more to do
        jsr     getstat
        jmp     loabrk          ; tidy-up

..nbsv  lda     #25+200         ; not bsave err
        jmp     badst
        endproc


;================================
;       Editor - SAVE file
;       ------------------
savex:  proc
        jsr     fobjcx          ; handle filespc

        lda     #ckbyt          ; out check bytes
        jsr     outax

        sec                     ; calculate file
        lda     txtend          ; size for cio
        sbc     txttab
        sta     slsz
        tax                     ; low byte
        lda     txtend+1
        sbc     txttab+1
        sta     slsz+1          ; high byte
        jsr     out2            ; out file size
        jsr     blkio           ; block io set

        lda     txttab+1        ; get adr of
        ldy     txttab          ; file start
sav1    jsr     io4.s           ; go do it!

        lda     objdv           ; close this file
        jmp     closei          ; return
        endproc


;=================================
;       Editor - BSAVE file
;       -------------------
bsavex: proc
        jsr     fobjcx          ; handle filespc
        inc     cox             ; skip < sign
        jsr     outff           ; binary save header

        jsr     _getn           ; get start addr
        sta     sltmp+1         ; for bsave
        stx     sltmp
        jsr     out2            ; out start addr

        inc     cox             ; skip comma
        jsr     _getn           ; get end address
        sta     slsz+1
        stx     slsz
        jsr     out2            ; out end addr

        sec                     ; form file size
        lda     slsz
        sbc     sltmp
        sta     slsz
        lda     slsz+1          ; hi byte
        sbc     sltmp+1
        sta     slsz+1
        inc     slsz            ; fix, bump size
        bne     ..f1f
         inc    slsz+1

..f1f   jsr     blkio           ; set cio
        lda     sltmp+1         ; get file start
        ldy     sltmp           ; adress
        jmp     sav1            ; go to it
        endproc



;=====================================
;       Editor text line tests
;=====================================

;       Test for text end
;       -----------------
endl?:  proc
        lda     stmcur+1
        cmp     txtend+1
        bcc     ll0
        bne     ll0
         lda    stmcur
         cmp    txtend
ll0     rts
        endproc


;       Test for end of line range
;       --------------------------
lline:  proc
        jsr     endl?           ; at text end?
        bcs     ll0             ; yes

;       Test stmcur against lelnm
;       -------------------------
el1     proc
        ldy     #1
        lda     (stmcur),y
        cmp     lelnm+1
        bcc     ..el2
        bne     ..el2
         dey
         lda    (stmcur),y
         cmp    lelnm
         beq    ..el3
..el2   rts
..el3   clc
        rts
        endproc


;       Editor get line
;       ---------------
egli    proc
        lda     nmfg            ; num set?
        beq     egli1           ; nope
         jmp    autonum         ; yup

egli1:  jsr     stext           ; set cio text
        jmp     io1.s           ; do cio + status
        endproc



;       List device open cio for output
;       -------------------------------
opnout: proc
        lda     #8              ; aux1, output
        ldy     #1              ; iocb #1
        sty     lstdv
        bne     sc1

;       x reference for output
;       ----------------------
fref:   lda     #8              ; output
        ldy     #5              ; iocb #5
        sty     refdv
        jsr     cio
        jmp     scix

;       input dev open cio for input (hfii is .include entry)
;       -----------------------------------------------------
hfile   inc     cox             ; skip # sign
hfi:    ldy     #3              ; iocb #3
hfii    lda     #4              ; open input
        sty     entdv
        bne     sc1

;       Object dev open for output
;       --------------------------
fobjcx  inc     cox
fobj:   lda     #8              ; output
        ldy     #2              ; iocb #2
        sty     objdv

;       Do OPEN device
;       --------------
sc1     jsr     cio
        jsr     movespc
scix    jsr     getio
io1.s   jsr     io1
        jmp     getstat
        endproc


;===========================
;       Error Handler
;       -------------
puterrasm: proc
        bit     real            ; assembler entry
        bpl     puterr          ; only if pass 2
..eh3    rts                    ; else, return

puterr: ldx     #0
        stx     dspflg          ; reset display control
        stx     nmfg            ; clear auto num
        ldx     eix             ; get er stack ptr
        beq     ..eh4           ; br if first err
         cpx    #3              ; top of stack?
         bcs    ..eh1           ; yup

..eh5   cmp     errstack-1,x    ; see if error
        beq     ..eh1           ; type already reported
         dex
         bne    ..eh5

        ldx     eix
..eh4   sta     errstack,x      ; no, save it
        inc     eix             ; bump err index

..eh1   asl     a               ; cio error?
        lda     eaf             ; if 0, then not asm
        beq     ..eh2
        bcc     ..eh3           ; cc= non-cio err
        jsr     ppass           ; title & page strings
        jsr     sppr            ; list asm line
..eh2   jsr     listerr
        jmp     noprm           ; edit exit
        endproc



;       Memory manager - expand
;       -----------------------
exptxt  proc
        eor     #-1             ; compliment
        tay
        iny
        ldx     #stmcur
explow: lda     #0
expand: sty     mvsiz           ; save expand
        sta     mvsiz+1         ; size
        sec                     ; test memtop+1
        lda     memtop
        adc     mvsiz
        tay
         lda    memtop+1
         adc    mvsiz+1
         cmp    mtp+1           ; must be less
         bcc    ..exsf          ; than high mem
         bne    exxi            ; oops
        cpy     mtp             ; check low
        bcc     ..exsf
        beq     ..exsf

exxi:   lda     #1              ; mem full err
abort:  ldy     #0              ; clear junk errs
        sty     eix
        sty     epf             ; replace flag
        ldy     eaf             ; assembling?
        beq     ..ab0           ; no
         pha                    ; save error
         jsr    ppage           ; dump page string
         jsr    sppr            ; force list of line
         pla

..ab0   jsr     puterr          ; stack err
        jsr     listerr         ; if asm then
        jmp     asmexit         ; abort

;       put to buffer/ & mask 7th bit
;       -----------------------------
p20:    lda     #$20
p7f     and     #$7f            ; bit off
p58:    ldy     cix             ; cur buff index
        sta     (inbuff),y      ; store
        inc     cix             ; bump index
        beq     ..ter           ; if overflow
         rts
..ter   lda     #22             ; list tabbing err
fter:   jmp     abort

..exsf  sec                     ; form mvlng
        lda     memtop          ; move from adress
        sbc     0,x             ; mvlng=
        sta     mvlgn           ; memtop-exp adress
         lda    memtop+1        ; mvfa(l)= exp
         sbc    1,x             ; adress(l)
         sta    mvlgn+1         ; mvfa(h)= exp

         clc                    ; adress(h) + mvlng(h)
         adc    1,x             ; mvfa= !memtop
         sta    mvfa+1          ; set mvfa(h)

        lda     0,x
        sta     mvfa            ; set mvfa(l)
        sta     svesa           ; save pre-move value
        adc     mvsiz           ; mvta(l)= exp
        sta     mvta            ; adress(l)+ mvsiz(l)
         lda    1,x             ; mvta(h)= [carry +
         sta    svesa+1         ; exp adress(h)+
         adc    mvsiz+1         ; mvsiz(h)]+
         adc    mvlgn+1         ; mvlng(h). mvta=
         sta    mvta+1          ; memtop+ mvsiz

..ep1   lda     0,x             ; add mvsiz to
        adc     mvsiz           ; all table entrys
        sta     0,x             ; from expand at
        lda     1,x             ; adress to high mem
        adc     mvsiz+1
        sta     1,x
        inx
        inx
        cpx     #memtop+2       ; at limit?
        bcc     ..ep1           ; no

        ldx     mvlgn+1
        inx                     ; mvlng(h)+1 in x
        ldy     mvlgn           ; mvlng(l) in y
        bne     ..ep4           ; test 0 length
        beq     ..ep5           ; br low =0

..ep2   dey                     ; pop mvlng(l),
        dec     mvfa+1          ; mvfa(h), &
        dec     mvta+1          ; mvta(h)

..ep3   lda     (mvfa),y        ; move 255 bytes
        sta     (mvta),y
..ep4   dey
        bne     ..ep3

        lda     (mvfa),y        ; move zero byte
        sta     (mvta),y
..ep5   dex                     ; another 256 byts?
        bne     ..ep2           ; yup
        rts                     ; done!
        endproc


;       Memory manager contract
;       -----------------------
contxt: proc
        tay
        ldx     #stmcur
conlow  lda     #0

contract:
        sty     mvsiz           ; save contract size
        sta     mvsiz+1         ; x
        sec                     ; form move
        lda     memtop          ; lenght low
        sbc     0,x             ; mvlng(l)=$100-
        eor     #-1             ; memtop(l)-contract
        tay                     ; adress(l), y = 2's
        iny                     ; compliment
        sty     mvlgn           ; x
         lda    memtop+1        ; set mvlng(h)
         sbc    1,x
         sta    mvlgn+1

        lda     0,x             ; set mvfa, mvfa =
        sbc     mvlgn           ; contract adress-mvlng(l)
        sta     mvfa            ; at move mvlng(l)
         lda    1,x             ; will be added to
         sbc    #0              ; mvfa in (ind),y
         sta    mvfa+1          ; instruction
        stx     mvta            ; temp save

..cont1 sec                     ; subtract mvsiz
        lda     0,x             ; from table from
        sbc     mvsiz           ; contract adress to
        sta     0,x             ; high mem
        lda     1,x
        sbc     mvsiz+1
        sta     1,x
        inx
        inx
        cpx     #memtop+2       ; done?
        bcc     ..cont1         ; no

        ldx     mvta            ; mvta= new contract -
        lda     0,x             ; mvlng(l), during
        sbc     mvlgn           ; move mvlng(l) will
        sta     mvta            ; be added to mvta
        lda     1,x             ; in (ind),y instr
        sbc     #0
        sta     mvta+1

cont4   ldx     mvlgn+1         ; entry for buffer move
        inx                     ; mvlng(h) +1 in x
        ldy     mvlgn           ; mvlng(l) in y
        bne     ..cont3         ; test for 0
        dex
        bne     ..cont3
        rts                     ; nothing to move!

..cont2 inc     mvfa+1          ; bump mvfa(h)
        inc     mvta+1          ; bump mvta(h)
..cont3  lda    (mvfa),y        ; move 256 bytes
         sta    (mvta),y
         iny
         bne    ..cont3
        dex                     ; pop high count
        bne     ..cont2         ; branch if more to do
        rts                     ; done!
        endproc


;========================================
;       EDITOR UTILITIES PACKAGE
;========================================

$if 0

;       Utilities - change & display mem
;       --------------------------------
dx      proc
        jsr     usa             ; set default end
        jsr     getn            ; get start addr
        sta     usadr+1
        stx     usadr

        jsr     usa             ; set end adr

        jsr     gtokel          ; user end address?
        beq     ..loop          ; nope!
         jsr    getn            ; get end address
         sta    ueadr+1
         stx    ueadr

..loop  jsr     docr            ; do cr
        lda     usadr           ; print out adr
        ldy     usadr+1         ; x
        jsr     hxws            ; x
        lda     #7              ; eight bytes/line
        sta     temp

..udx   ldy     #0
        lda     (usadr),y
        jsr     hxbs            ; print byte + sp
        jsr     testbk          ; break?
        bne     ..uex           ; yes
         jsr    uip             ; point to nxt
         lda    usadr+1         ; tst for end
         ldy    usadr
         cmp    ueadr+1         ; at end?
         bcc    ..unl
         bne    ..uex
        cpy     ueadr           ; tst low
        bcc     ..unl
        beq     ..unl           ; yup
..uex    rts

..unl   dec     temp            ; do new line?
        bpl     ..udx           ; no
        bmi     ..loop          ; yup!
        endproc


;       Set end address
;       ---------------
usa     proc
        clc                     ; set up
        lda     usadr           ; display
        adc     #7              ; end adr
        sta     ueadr
         lda    usadr+1         ; high
         adc    #0
         sta    ueadr+1
        rts
        endproc
$endif


$if 0

;       Utilities - Execute change
;       --------------------------
cx      proc
        jsr     getn            ; starting address
        sta     usadr+1
        stx     usadr
        inc     cox             ; past <

..loop  jsr     testbk          ; break?
        bne     uip             ; yup
        jsr     gtokel          ; done all locations?
        beq     uip             ; yup

        cmp     #@lcma          ; skip adr?
        bne     ..uca           ; no
         jsr    uip             ; go point to nxt
         jmp    ..loop          ; and loop

..uca   jsr     getne           ; get num
        tay                     ; high should be 0
        bne     ..val           ; oops >255
         txa                    ; low byte to a
         sta    (usadr),y
         jmp    ..loop          ; yup

..val   lda     #10             ; val err
        jmp     puterr          ; quit & tell user
        endproc


;       Increment pointer
;       -----------------
uip     proc
        inc     usadr           ; inc pointer
        bne     ..uiy
         inc    usadr+1
..uiy   rts
        endproc

$endif



;       print line pointed to by stmcur
;       -------------------------------
prli    proc
        jsr     movin           ; move to inbuff
                                ;   fall thru inbout

;       move data pointed to by inbuff
;       ------------------------------
inbout: ldy     #0              ; clear index
inbout1:
        sty     cix

..inb   lda     (inbuff),y      ; get char
        pha
        cmp     #linefeed       ; line feed?
        bne     ..wilk          ; nope
        ldy     eaf             ; assembling?
        beq     ..wilk          ; nope
         jsr    tstej           ; test page eject
         pla
         pha                    ; restore a

..wilk  jsr     charo           ; print it
        pla
        inc     cix             ; bump index
        ldy     cix
        cmp     #eol            ; if eol, then end
        bne     ..inb
        rts
        endproc


;       Move string to lbuff with eol
;       -----------------------------
movespc proc
        sta     temp            ; spec len in a
        ldx     #0              ; clear index
..msp   jsr     gtoke           ; get char
        sta     lbuff,x         ; put in lbuff
        inx
        dec     temp            ; pop count
        bne     ..msp           ; loop if more

        lda     #eol            ; tack on eol
        sta     lbuff,x
        rts
        endproc


;       get command token, cmp with @eol
;       --------------------------------
gtokel: proc
        jsr     gtoke           ; get code
        cmp     #@eol           ; set status
        rts
        endproc


;       get a token from exbuff
;       ----------------------
gtoke:  proc
        ldy     cox             ; get index
        inc     cox             ; point to nxt
        lda     (exbuf),y       ; get token
        rts
        endproc


;       Do block I/O
;       ------------
; in:
;       a       = command
;       x       = device

blkio   proc
        ldx     objdv           ; iocb #
blkio2  lda     #$0b            ; put bytes

blkio1  jsr     glcx            ; set command & index
        lda     slsz            ; move file size
        sta     icbufl,x        ; to cio
        lda     slsz+1
        sta     icbufl+1,x
        rts
        endproc


;       test left margin  (return beq always)
;       -------------------------------------
leftmr: proc
        lda     leftmar         ; left margin?
        bne     space           ; do 'em
         rts

;       do spaces
;       ---------
space3: lda     #3              ; do 3 spaces
space:  sta     ttix
..sc    jsr     dosp
        dec     ttix
        bne     ..sc            ; return 0
        rts
        endproc


;       get 1 byte from dev in acc
;       --------------------------
gobj:   proc
        lda     #7              ; was 6
        jsr     glcxent         ; dev index & cmd
        jsr     io2             ; get 1 byte in a
        bmi     ..cr3           ; err!
         ldx    #-1             ; clear poss zero flg
         rts
                                ; single character out routine
                                ;  also tests control-1 key (atari)
                                ;  and re-trys timeout errors
charo:  ldx     lstdv
        inc     column          ; bump col cnt

cr1     pha                     ; save char
        jsr     gcx             ; get cio index
        stx     ktmp            ; in case fail
        pla                     ; pop char
..cr2   ldx     ssflag          ; start/stop?
        bne     ..cr2           ; yup wait

        pha                     ; if error
        tay
        ldx     ktmp
        jsr     pdum
        pla                     ; balance stack
        cpy     #138            ; did timeout occur?
        bcc     ..cr3           ; nope
        cpy     #139+1          ; or nak
        bcs     ..cr3
         dec    ssflag          ; self-inhibit!
         bcc    ..cr2           ; unc, go wait

..cr3   tya
        jmp     gst1            ; ck status
        endproc


;       put: y enters char
;       ------------------
pdum    proc
        lda     icaux1,x        ; set z-page iocb
        sta     icax1z
        lda     icaux2,x
        sta     icax2z
        lda     icput+1,x       ; rts to routine
        pha
        lda     icput,x
        pha
        tya
        ldy     #146            ; preset not impli
        rts
        endproc



;       out 2 bytes to dev in A & X
;       ---------------------------
outff:  proc
        lda     #-1             ; out $ffff
outax   tax
out2    pha
        txa
        jsr     ..out1
        pla
                                ; out 1 byte to dev in a
..out1  ldx     objdv
        jmp     cr1
        endproc




;       set CIO for input text
;       ----------------------
stext:  proc
        lda     #5

                                ; set dev index & cmd in cio
glcxent ldx     entdv
glcx    sta     ciocmd          ; cmd in a
gcx     stx     ionum           ; iocb in x

                                ; get cio index
getio   lda     ionum
        asl     a
        asl     a
        asl     a
        asl     a
        tax
        rts
        endproc


;       Do CIO
;       ------
cio:    proc
        pha                     ; save cio cmd
        sty     ionum           ; save iocb #
        jsr     getio
        lda     #$0c            ; close dev first
        jsr     io8
        ldy     #3              ; cmd for open
        sty     ciocmd
        pla                     ; get open aux1
        sta     icaux1,x
gtoke7: jsr     gtoke
        and     #$7f
        rts                     ; ret a=filspc len
        endproc


;       call 'io4' & fall to 'getstat'
;       ------------------------------
io4.s:  proc
        jsr     io4
                                ; get cio status
getstat:
        jsr     getio           ; get dev index
        lda     icsta,x
gst1    bmi     badst
..so6   rts                     ; ret to caller

badst   ldy     #0              ; reset display
        sty     dspflg          ; control
        sta     sverr           ; save err #
        cmp     #$80            ; break?
        bne     so1             ; nope
        sty     brkkey          ; simulate break
        jsr     cfile           ; close files, and
        lda     eaf             ; assembling?
        bne     ..abr           ; yup
        jmp     tef

..abr   sta     bank2           ; ^^^^^^^^^ bank 2 ^^^^^^^^
        jmp     asmexit         ; abort everything

so1     lda     ionum           ; was it enter
        cmp     entdv           ; dev?
        beq     so3             ; yup
        bne     so4

so2     sta     sverr           ; save err
        jsr     reset           ; reset for edit

so3     lda     eaf             ; ck mode
        bne     so7             ; br if in asm
        jsr     cfile           ; close
        bne     ..so5           ; br if not eof

        ldy     eaf             ; if not edit,
        bne     ..xit           ; br
        lda     #0              ; clear auto flag
        sta     nmfg            ; x
        jmp     prmpt           ; ret to edit

so4     jsr     cfile           ; close
        beq     ..xit           ; don't print if eof

..so5   jsr     puterr  ; CART does it this way!!
..xit   rts             ; everyone else branched to ..so6 in steves version!

so7     lda     sverr
        cmp     #eof            ; eof?
        bne     so4             ; no, print error

        ldx     ionum
        cpx     #4              ; we in .include?
        beq     so8             ; yup
        lda     eaf             ; don't point after
        bpl     ..xit           ; 2nd pass; return

        lda     #$25
        jsr     glcxent         ; dev index & cmd

        lda     #2              ; second byte
        sta     icbyt,x
        lda     pointer         ; move sector #
        sta     icsec,x         ; to iocb
        lda     pointer+1
        sta     icsec+1,x
        jsr     io5             ; go do it
        bmi     so2             ; if err
        jmp     popsizef        ; get file size

so8     jsr     close           ; close include dev
        lda     oldentdv        ; restore old
        sta     entdv           ; dev, and close
        sta     ionum           ; restore ionum!!!
        beq     so9             ; skip if memory file

        ldy     #5              ; restore buffer
        jsr     movedv
        ldx     #5              ; restore pointers
so0     lda     savdevi,x
        sta     buffer,x
        dex
        bpl     so0

so9     lda     #$40            ; set eof for
        sta     eofflag         ; include
        rts
        endproc



;       close routines
;       --------------
cfile   proc
        lda     #0              ; clear:
        sta     entdv           ; enter dev
        sta     lstdv           ; list dev
        sta     refdv           ; x-ref dev
        sta     objdv           ; object dev
        jsr     clsall          ; close devices
tef     lda     sverr
        cmp     #eof            ; set eof status
        rts
        endproc

;       close iocb in ionum
;       -------------------
closei: proc
        sta     ionum
close   jsr     getio
        beq     ..zno
         lda    #$0c
         jsr    io8
..zno   rts
        endproc


;       asm - close all devices
;       -----------------------
clsall: proc
        ldy     #5
        sty     ionum
..cla0  jsr     close
        dec     ionum
        bne     ..cla0
        rts



;       i/o call routine
;       ----------------
io1     proc
        ldy     #126            ; buf size =126
         db $2C
io2     ldy     #0              ; buf=0 (byte in a)
io3:    lda     #0
        sta     icbufl+1,x      ; set buff size
        tya
        sta     icbufl,x
        lda     inbuff+1
        ldy     inbuff
io4     sta     icbufa+1,x      ; set buff adr
        tya
        sta     icbufa,x
io5     lda     ciocmd          ; get cmd
io8     sta     iccom,x         ; x
        lda     #0              ; clear aux2
        sta     icaux2,x
        jmp     (xcii)          ; do i/o
        endproc


;       test for break key or "break"
;       -----------------------------
testbk: proc
        ldy     #0
        lda     brkkey
        bne     ..01
         dey
         sty    brkkey
..01    tya
        rts
        endproc


;       print dec # right justified (field size in y)
;       ---------------------------------------------
pdnumf  proc
        sta     fr0+1
        stx     fr0
        dey                     ; pop for cmp
        sty     ttix
        jsr     inii            ; integer to ascii
        ldy     #-1             ; get digit cnt
..dcl   iny
        lda     (inbuff),y
        bpl     ..dcl
        lda     ttix            ; get spaces to
        sty     ttix            ; print left of
        sec                     ; number
        sbc     ttix
        beq     ..pn0           ; no spaces
        bmi     ..pn0           ; no spaces
        jsr     space           ; leading spaces
        beq     ..pn0           ; unc, do number

                                ; print dec num in fr0
pdnum0: lda     #0              ; set high =0
pdnum:  sta     fr0+1
        stx     fr0
prdn:   jsr     inii            ; interger to ascii
..pn0   ldx     #-1
        stx     svn2            ; init index

..pn1   inc     svn2
        ldy     svn2
        lda     (inbuff),y

enters: php                     ; save stat
        bit     entdv           ; if entering
        bmi     ..pn2           ; and in auto, skip
         jsr    domk            ; out char
..pn2   plp
        bpl     ..pn1           ; br if not end
        rts
        endproc


inii:   proc
        jsr     ifp
        jmp     fascc
        endproc


;       print hex word & byte
;       ---------------------
hexw:   proc
        pha                     ; hex word in a & y
        tya                     ; high in y
        jsr     hexb
        pla
hexb    pha                     ; hex byte on a
        lsr     a
        lsr     a
        lsr     a
        lsr     a
        jsr     heo
        pla
        and     #$0f

heo     tay
        lda     hext,y
        jmp     charo
        endproc

hext:   db      '0123456789ABCDEF'


;       misc print routines
;       -------------------
hxbs    proc
        jsr     hexb            ; hex byte+space
dosp:   lda     #$20            ; do space
domk    and     #$7f            ; mask 7th bit
..och   jmp     charo

ptsc:   ldy     #high mes       ; for start-up msg
ptse:   jsr     ptst            ; text str+eol
docr:   lda     #eol            ; do eol
        bne     ..och
hxws    jsr     hexw            ; hex word+space
        jmp     dosp
        endproc


;       do 3 carrage returns
;       --------------------
dcr3    proc
        jsr     docr
        jsr     docr
        jmp     docr
        endproc

;       pr text str pointed by x & y
;       ----------------------------
ptst:   proc
        stx     msgadr          ; text string
        sty     msgadr+1
        ldy     #-1             ; set index
        sty     temp
..54    inc     temp
        ldy     temp
        lda     (msgadr),y
        bmi     domk            ; end, return thru
        jsr     domk            ; else, jsr
        jmp     ..54            ; and cont
        endproc


;==========================================
;       Assembler output test routines
;==========================================

;       test to see if we list
;       ----------------------
tstlst: proc
        lda     eaf             ; get pass #
        bmi     ..qx            ; no
        lda     optflag         ; opt flg for lst
        bpl     ..qz            ; yes

tstlss: and     #$20            ; err bit set?
        bne     ..qx            ; no
        ldx     eix             ; err indx
        beq     ..qy            ; no errs, no lst
..qz    lda     #0
..qx    rts
..qy    dex                     ; set not zero
        rts
        endproc


;       test to see if errors
;       ---------------------
tsterr: proc
        lda     optflag
        and     #$20            ; err bit set?
        bne     ..qv            ; no
         jmp    listerr         ; list errors

..qv    lda     #0              ; clr er indx
        sta     eix
        rts
        endproc


;       test for page eject
;       -------------------
tstejcr: proc
        jsr     docr            ; do a cr first
tstej:  inc     linecnt         ; inc line cnt
        lda     linecnt
        cmp     formsz          ; at max?
        bcc     tstj1           ; no
                                ; non-centronics entry
ject    lda     optflag
        lsr     a               ; no eject opt?
        bcs     tstj0           ; yes

;       eject page
;       ----------
        jsr     tstlst
        bne     tstj0           ; ret

ppass:  lda     fr0             ; stack fr0
        pha
        lda     fr0+1
        pha
        lda     topfc           ; centronics if 0
        beq     ..ctr0
         jsr    charo           ; do top of form
         jmp    ..ctr1

..ctr0  jsr     dcr3
..ctr1  jsr     ptitle
        jsr     ppage
        jsr     dcr3
        pla                     ; unstack fr0
        sta     fr0+1
        pla
        sta     fr0
tstj0   lda     #0              ; clr pg cnt
        sta     linecnt         ; stat = 0
tstj1   rts
        endproc


;       Directive - .PAGE
;       -----------------
@@page: proc
        ldx     #1              ; get string statement
        jsr     msrg            ; move string
        jsr     tstlst          ; see if list
        bne     @ph             ; no list
         jsr    spp3            ; list line
                                ;   bit linecnt just ejected?
                                ;   branch @ph
         jsr    page55          ; eject nxt page
         dec    linecnt         ; reset to -1
@ph     rts                     ; else, ret
        endproc


;       do blank lines until new page
;       -----------------------------
page55  proc
        lda     topfc           ; centronics if 0
        beq     page56
        jmp     ject            ; and page eject

@pk     jsr     docr
page56  jsr     tstej           ; bump line cnt
        lda     linecnt         ; test for zero
        bne     @pk             ; do blank line
        rts
        endproc


;       Print .TITLE & .PAGE
;       --------------------
ptitle  proc
        lda     refdv           ; doing x-refs?
        bne     ..nwr           ; no
         lda    sxref           ; option no x-ref?
         beq    ..nyr           ; no
..nwr   lda     #rfpag          ; flag page
        jsr     xrout

..nyr   jsr     leftmr          ; test left margin
        inc     pagenum         ; bump page count
        bne     ..pi0
         inc    pagenum+1

..pi0   ldy     #high pames     ; print 'page  '
        ldx     #low pames
        jsr     ptst

        ldx     pagenum         ; print page number
        lda     pagenum+1
        jsr     pdnum           ; pr page #
        jsr     space3          ; three spaces

        ldx     strings         ; print title string
        ldy     strings+1
        jmp     ptse            ; print & ret
        endproc


;       print page string & current file
;       --------------------------------
ppage   proc
        lda     #0              ; init column
        sta     column
        jsr     leftmr          ; test margin
        clc                     ; form pagep addr
        ldy     strings+1
        lda     strings
        adc     #strmax
        tax                     ; low in x
        bcc     ..ppq
         iny
..ppq   jsr     ptst            ; print page string
        jsr     dosp            ; trail a space
        sec                     ; get # of fill spaces
        lda     rightmar
        sbc     column
        bcc     ..pof           ; no fill
        sbc     #fz             ; and size of buff
        bcc     ..pof           ; no fill
         jsr    space           ; do fill

                                ; now, dump current filespec
..pof   lda     svedv           ; get macro saved dev
        ldy     mcrsk           ; in a macro?
        bne     ..im            ; yes, a= dev#
         lda    entdv           ; else, get cur dev

..im    cmp     #4              ; include = iocb #4
        ldy     #high asmfm     ; set for main
        ldx     #low asmfm
        bcc     ..pw            ; it's main
         $if (high asmfm) <> (high asmfi)
          ldy   #high asmfi
         $endif
         ldx    #low asmfi      ; else, include file
..pw    jmp     ptse            ; pr str & eol
        endproc


pames   x_db    'PAGE '
ssm     x_db    'SYMBOLS'
ssml    equ     $-ssm

errasm: x_db    'ASSEMBLY ERRORS: '
bytasm: x_db    ' BYTES FREE'
chartype
        db      ' ?=?%?=?'



;       Assembler listing & object routine
;       ----------------------------------
        proc
..nrg   lda     #19
        jmp     abort

putinbuf0:
        sta     tpobuf          ; format cnt
putinbuff:
        bit     eaf             ; get pass
        bpl     ..pf1           ; num 2
..pf2    jsr    dopc            ; update pc
         sta    asmpc+1
         stx    asmpc
         rts
                                ; PASS 2: do object & listing
..pf1   bit     optflag         ; test for obj
        bvs     ..nbj           ; nope!
        jsr     gpc             ; get offsetted pc
        sta     ztemp4
        sty     ztemp4+1

        lda     objix           ; ptr =0?
        beq     ..nbj           ; yup
        lda     pcdef           ; is pc def?
        beq     ..nrg           ; no, err
        ldy     objdv           ; to a dev?
        bne     ..pf4           ; yup

        dey                     ; set y to $ff
..pf3   iny
        cpy     objix           ; done?
        bcs     ..nbj           ; y
         lda    tpobuf+1,y
         sta    (ztemp4),y      ; put in mem
         bcc    ..pf3           ; jump always

..pf4   ldy     #-1             ; to device
        sty     svf2            ; temp
..pf5   inc     svf2
        ldy     svf2
        cpy     objix           ; at end?
        bcs     ..nbj
         lda    tpobuf+1,y      ; fill obj
         jsr    inbuffo         ; buffer
         bne    ..pf5           ; jump always

                                ; do listing output
..nbj   lda     #0
        sta     tobjdex
        jsr     tstlst
        bne     ..pf2           ; no list

        jsr     pobjl           ; pr obj l.
        lda     #16             ; cal space fill
        sec                     ; x
        sbc     column          ; never zero
        jsr     spcem           ; spaces & macro m
        jsr     asmlst          ; pr input l.

..al4   lda     objix           ; if obj =0 then
        beq     ..al3           ; end
         jsr    pobjl           ; else, do obj l.
         jsr    docr            ; and eol
         jmp    ..al4           ; repeat
..al3   jmp     tsterr          ; ret thru
        endproc



;       print line of object
;       --------------------
pobjl   proc
        jsr     tstej           ; eject?
        jsr     leftmr          ; test left margin
        lda     #0
        sta     column          ; init column
        jsr     papc            ; print pc w/space
        lda     objix
        beq     ..p00           ; all done!
        lda     tpobuf          ; get format
        sta     zt1

..p01   ldy     tobjdex
        lda     tpobuf+1,y      ; get byte
        jsr     hexb            ; pr hex
        inc     asmpc           ; set for next
        bne     ..p02
         inc    asmpc+1

..p02   inc     tobjdex
        dec     objix
        beq     ..p00           ; done
         dec    zt1             ; dec bytes on l.
         bne    ..p01
..p00   rts
        endproc


papc:   proc
        jsr     gpc             ; get pc
        jmp     hxws
        endproc


gpc     proc
        clc
        lda     asmpc
        adc     offset          ; add in offset
        tax                     ; save in x
        lda     asmpc+1
        adc     offset+1
        tay
        txa
        rts
        endproc



;       get token line (faster 12/16/83)
;       --------------------------------
objx3:  proc
        lda     #3              ; get 3 bytes (line header)
objx:   sta     many            ; save bytes to get
newget  lda     bytcnt+1        ; empty buffer?
        bne     ..ij0           ; can't be!
        ora     bytcnt
        beq     get_1k          ; get next block
        cmp     many            ; enough in there?
        bcs     ..ij0           ; yes

                        ; last line is short...
                        ;  do as two call to "objx"
        sec
        lda     many
        sbc     bytcnt          ; calculate remainder
        pha                     ; save that
        lda     bytcnt          ; get greatest #
        jsr     objx
        pla
        bne     objx            ; unc, get remainder

                        ; get specified number of bytes
..ij0   clc                     ; set-up move to addr
        lda     cox             ; so 'many' can be counter
        adc     exbuf
        sta     amvta
         lda    exbuf+1
         adc    #0
         sta    amvta+1

        ldy     #0              ; now, the fast move
..ij1   lda     (buffer),y
        sta     (amvta),y
        iny
        cpy     many
        bcc     ..ij1
        pha                     ; save last byte!!!

                        ; now, adjust pointers, counters
        lda     bytcnt          ; cs, pop byte count
        sbc     many
        sta     bytcnt
        lda     bytcnt+1
        sbc     #0
        sta     bytcnt+1

        clc                     ; bump buffer ptr
        tya                     ; y= 'many'
        adc     buffer
        sta     buffer
        bcc     ..ij2
         inc    buffer+1
         clc

..ij2   tya                     ; adjust 'cox'
        adc     cox             ; store new cox
        sta     cox
        pla                     ; recover last byte
                                ; cc= sucess
        rts                     ; a=last byte moved

                        ; get next block from device

get_1k  lda     filesize+1      ; test for eof
        ora     filesize
        beq     reof

        ldx     filesize
        lda     filesize+1      ; see if get will
        cmp     #high bufsize   ; be less than buffer
        bcc     getk2
         lda    #high (bufsize-1)       ; set # of
         ldx    #low (bufsize-1)        ; bytes to get

getk2   sta     slsz+1
        stx     slsz
        sec                     ; adjust
        lda     filesize        ; number of bytes
        sbc     slsz            ; remaining for
        sta     filesize        ; this file
         lda    filesize+1
         sbc    slsz+1
         sta    filesize+1

        lda     #7              ; 'get block' cmd
        ldx     entdv           ; input dev
        jsr     blkio1          ; set some cio
        lda     buffers+1       ; set buffer addr
        sta     buffer+1
        ldy     buffers
        sty     buffer
        jsr     io4             ; do cio

        cpy     #$80            ; break abort?
        beq     doquit          ; yup
        bcc     ..tstc          ; get byte count
        cpy     #eof            ; cio eof?
        beq     ..tstc          ; yup, ignore

         jsr    getstat         ; handle error
doquit   jmp    asmexit         ; finish up

                        ; copy number of bytes moved
..tstc  lda     icbufl,x
        sta     bytcnt
         lda    icbufl+1,x
         sta    bytcnt+1
        jmp     ..ij0           ; unc

                        ; handle eof
reof    lda     entdv           ; simulate call to cio
        sta     ionum
        lda     #eof
        jsr     badst           ; handle eof
        sec                     ; set eof status
        rts
        endproc


;       move device buffers
;       -------------------
; in:
;       y       = move direction

movedv  proc
        ldx     #3
..mvbf  lda     bufftab,y
        sta     mvfa,x
        dey
        dex
        bpl     ..mvbf

        lda     #low (bufsize ^ -1) + 1
        sta     mvlgn
         lda    #high bufsize
         sta    mvlgn+1
        jmp     cont4
        endproc


;       put object buffer to dev
;       ------------------------
runout: proc
        ldy     #0              ; pre-set <256
        lda     objptr          ; buffer size
        bne     runo0           ; br: less
         iny                    ; set = 256
runo0   sta     slsz            ; set-up for blkio
        sty     slsz+1
        jsr     blkio
        lda     objbuf+1        ; buff address
        ldy     objbuf
        jmp     io4.s           ; go do it
        endproc


;       Directive - .INCLUDE
;       --------------------
@@include: proc
        inc     cox             ; skip pound sign
        lda     entdv
        cmp     #4              ; already include?
        beq     inest
        sta     oldentdv        ; save iocb

        tay
        beq     fincl           ; don't move if memory
        jsr     movedv          ; y = 3
        ldx     #5              ; save primary pointers
..incsv lda     buffer,x
        sta     savdevi,x
        dex
        bpl     ..incsv

fincl   lda     #0              ; clear byte count
        sta     bytcnt
        sta     bytcnt+1
        jsr     intlbf          ; reset inbuff
        ldy     #4              ; iocb #4
        jsr     hfii            ; do cio

        ldy     #low asmfi      ; CART version...
        jsr     hstr

        jsr     validf          ; test ck bytes
        ldx     #high $ff00     ; (line #65280)
        stx     lastl+1         ; force x-ref word token
        jmp     spp3            ; list

inest   lda     #21             ; include err
        jmp     sp3er
        endproc


;       move filespec string for listing
;       --------------------------------
; in:
;       Y       = low addr of buffer

hstr:   proc
        $if (( high (asmfi- high asmfi))) <> (( high (asmfm- high asmfm)))
         $error "asmfi, asmfm page"
        $endif

        ldx     #high asmfi     ; CART does it this way!
        stx     srcadr+1
        sty     srcadr
        ldy     #0
..mm    lda     lbuff,y
        sta     (srcadr),y
        iny
        cmp     #eol            ; done?
        beq     ..ml            ; yup
        cpy     #fz
        bcc     ..mm            ; no

..ml    dey                     ; flag last char
        ora     #$80
        sta     (srcadr),y
        rts
        endproc



;       dump x-ref data
;       ---------------
xrout:  proc
        ldx     #0              ; byte in a
        sta     (xrefbufi,x)    ; store
        inc     xrefbufi        ; bump pointer
        bne     ..qrfo
         inc    xrefbufi+1      ; bump pointer h

..qrfo  lda     xrefbufi+1
        cmp     asmram+3        ; top of buff?
        bcc     ..qrfr          ; nope
         lda    xrefbufi        ; try low
         cmp    asmram+2
         bcs    rfwrite         ; dump this block
..qrfr  rts

rfwrite:
        sec                     ; form buff size
        lda     xrefbufi
        sbc     bufferx
        tax
         lda    xrefbufi+1
         sbc    bufferx+1
        stx     slsz            ; for blkio
         sta    slsz+1
        txa
        ora     slsz+1          ; nothing there?
        beq     ..qrfr          ; nope!

        clc
        txa                     ; update file size
        adc     xrefsz
        sta     xrefsz
         lda    slsz+1
         adc    xrefsz+1
         sta    xrefsz+1

        tya
        pha                     ; save y
        ldx     refdv
        jsr     blkio2          ; set-up cio
        lda     bufferx+1
        sta     xrefbufi+1      ; reset buffer
         ldy    bufferx
         sty    xrefbufi
        jsr     io4.s           ; do cio
        pla
        tay                     ; restore y
v00     rts
        endproc



wrtxref: proc
        bit     eaf             ; only if pass 2
        bmi     v00
        lda     xrindx          ; something there?
        beq     v00             ; nope!
        lda     refdv           ; and doing it...
        beq     v00

        ldy     #1
        lda     (exbuf),y       ; line # relative?
        cmp     lastl+1
        bne     yrfw            ; nope
        dey
        lda     (exbuf),y       ; low
        sec
        sbc     lastl
        beq     yrfb            ; if 0
        cmp     #120
        bcs     yrfb            ; two byte relative
        eor     #-1             ; compliment
        adc     #1
        bne     yrrv

yrfb    pha                     ; save offset
        lda     #rfnub          ; relative >=120
        jsr     xrout
        pla
yrrv    jsr     xrout           ; out it
        jmp     ..uref0         ; y=0

yrfw    lda     #rfnum          ; line # word flag
        jsr     xrout
        ldy     #1
..yxr   lda     (exbuf),y       ; hi/lo format
        jsr     xrout
        dey
        bpl     ..yxr
        iny                     ; y=0

..uref0 lda     xrefbuf,y       ; out references
        jsr     xrout
         lda    xrefbuf+1,y
         jsr    xrout
        iny
        iny
        cpy     xrindx          ; done?
        bcc     ..uref0

        ldy     #0              ; reset index
        sty     xrindx
                        ; set last line #
        lda     (exbuf),y
        sta     lastl
         iny
         lda    (exbuf),y
         sta    lastl+1
        rts                     ; ret
        endproc



;       note starting sector of entdv
;       -----------------------------
donote: proc
        lda     #$26            ; set for note command
        jsr     glcxent
        jsr     io5             ; do cio
        lda     icsec,x         ; save sector # for
        sta     pointer         ; pass 2 or x-ref use
        lda     icsec+1,x
        sta     pointer+1
        jmp     getstat         ; ret & check status
        endproc


;======================================================================


;       symbol table print out
;       ----------------------
@@symbol: proc
        jsr     asmmess         ; asm end message
        lda     #0              ; init to no x-ref
        sta     rfpagf
        lda     refdv           ; doing x-ref?
        bne     drefs           ; yup!
        lda     optflag         ; check list option
        bmi     ..10
         jmp    ps1
..10    rts                     ; no list, no table!

                        ; dump x-refs
drefs   lda     #($c1*2)&255    ; set pre-shifted value
        asl     optflag         ; bit 7 to carry (list bit)
        ror     a               ; to bit 7 of acc
        lsr     optflag         ; clearing bit 7
        sta     rfpagf          ; format flag:
                        ;  a=$41 print page #'s
                        ;  a=$c1 no page #'s
        lda     #1              ; contract macro
        ldy     #0              ; parameter stack
        ldx     #stack+2
        jsr     contract
                                ; contract macro table
        sec
        lda     macend
        sbc     macstrt
        tay                     ; low =y
         lda    macend+1
         sbc    macstrt+1
        ldx     #macend
        jsr     contract
                                ;  zap buffers
        lda     #high (bufsize*2+bufsize/2+strmax+(strmax-ssml-1))
        ldy     #low  (bufsize*2+bufsize/2+strmax+(strmax-ssml-1))
        ldx     #asmram+2
        jsr     contract

                        ; now, test file size
        clc                     ; can file fit?
        lda     xrefsz
        sta     slsz            ; set file size
        adc     memtop
        tax                     ; save low
         lda    xrefsz+1
         sta    slsz+1          ; file size high
         adc    memtop+1
         cmp    mtp+1
         bcc    xrok            ; good, it'll fit
         bne    ..oops
        cpx     mtp
        bcc     xrok            ; good, it'll fit

                        ; file is too big
..oops  lda     #0
        sta     rfpagf          ; no xref's
        beq     ps1
                        ; burst it in
xrok    ldx     entdv
        lda     #7              ; get binary cmd
        jsr     blkio1          ; set some cio
        jsr     setbf           ; load addr & buffer addr
        jsr     io4.s           ; do the io/status

                        ; dump symbol table
ps1     ldx     #ssml           ; move "symbols"
        ldy     #strmax+ssml-1
..00    lda     ssm-1,x         ; message
        sta     (strings),y
         dey
         dex
         bne    ..00

        stx     refdv           ; more references
;       stx     sxref           ; and shadow            not in CART!

        lda     #$40            ; reset asm pass 2
        sta     eaf
        jsr     page55          ; eject new page

        sta     bank3           ; ^^^^^^^^ bank 3 ^^^^^^^
        jsr     sortsymbols     ; sort labels
        sta     bank2           ; ^^^^^^^^ bank 2 ^^^^^^^

        ldx     #'@'            ; start past locals
        lda     system          ; special?
        and     #$20
        beq     ..l             ; nope
         dex                    ; yes, do locals
..l     stx     sindex          ; as index

..psn3  lda     sindex          ; get index
        inc     sindex          ; bump
        cmp     #'Z'+1          ; done?
        bcc     ..psn1          ; no
         jmp    docr

..psn1  asl     a               ; get table index
        tax
        ldy     spointer-'?'*2+1,x      ; high address
        bmi     ..psn3          ; nothing there.

        jsr     formtop         ; set topsort

..ps2   lda     #$80            ; just did cr?
        cmp     lablcc
        beq     ps.1            ; yes, don't do twice!

         jsr    tstejcr         ; new letter, new line

ps.1    jsr     leftmr          ; left margin
        jsr     testbk          ; break?
        beq     ..ps3           ; no
         rts

..ps3   lda     #4              ; four columns
        sta     lablcc

..ps4   dec     lablcc
        bmi     ..ps2           ; new line
        bit     lablcc          ; special for x-ref
        bvs     ps.1            ; yup

        ldx     #6              ; start at 6th char
        stx     lablch
                                ; get offset
        lda     (labelptr-6,x)  ; x=6 (fun!)
        sta     lablsz
        tay
        dey
        lda     (labelptr),y    ; test for macro
        cmp     #eol            ; string
        beq     ..ps6           ; yup, so skip
        cmp     #'('            ; or internal local
        beq     ..ps6           ; yup, so skip
        stx     lablst          ; char index (x=6)

        ldy     #3              ; get define byte
        lda     (labelptr),y
        and     #7
        tax                     ; do leading type
        lda     chartype,x      ; character
        jsr     charo

        ldy     #1
        lda     (labelptr),y    ; print label
        pha                     ; address
        iny
        lda     (labelptr),y
        tay
        pla
        jsr     hxws            ; hex + space

                        ; out label
..ps5   ldy     lablst          ; get index
        inc     lablst
        cpy     lablsz          ; past last?
        bcs     pad             ; yup, pad spaces
         lda    (labelptr),y
                                ; label a local?
        cpy     #6              ; only first character
        bne     ..nolo
        cmp     #'@'
        bcs     ..nolo
        lda     #':'            ;local leading char
..nolo
        jsr     domk            ; print char
        inc     lablch          ; bump char count
        bne     ..ps5           ; unc

                        ; pad spaces
pad     jsr     dosp            ; space
        inc     lablch          ; bump char count
        lda     lablch
        sec                     ; at next column?
        sbc     #80/4
        beq     ..ps7           ; yup
        bmi     pad
        sta     lablch
        dec     lablcc          ; pop columns
        jmp     pad
                        ; test last or point to next
                        ; & handle x-refs
..ps7   lda     rfpagf          ; do x-ref?
        beq     ..ps6           ; nope
        lda     sindex          ; get current alpha
        cmp     #'@'            ; locals: no references
        beq     ..ps6
         jsr    findr           ; find & dump x-refs

..ps6   ldy     #5
        lda     (labelptr),y
        bpl     ..ps8           ; not last
         jmp    ..psn3          ; last, get next

..ps8   jsr     formlab
        sty     labelptr
        sta     labelptr+1
        jmp     ..ps4
        endproc



;       dump x-refs
;       -----------
findr   proc
        asl     lablcc          ; double columns
        dec     lablcc          ; minus 2
        dec     lablcc
        ldx     #0              ; init for each pass:
        stx     lastl           ; current line #s
        stx     lastl+1
        stx     rfpgn           ; page #
        stx     rfpgn+1
        stx     xrfrdo          ; no reference flag
        sec
        lda     labelptr        ; form search link
        sbc     smbltp
        sta     link
         lda    labelptr+1
         sbc    smbltp+1
         sta    link+1
        jsr     setbf           ; load addr & buffer addr

                        ; x-ref search/dump loop
                        ; ----------------------
..xloop lda     #$20            ; set leading char
        sta     xrfchar
..xlopz jsr     grefbyt         ; byte at a time!
        tax                     ; find sign
        bmi     ..xlo1          ; if token type

        cmp     link+1          ; a high link
        beq     ..xz0           ; yes, check low
         jsr    gref1           ; just skip
         bne    ..xloop         ; unc

..xz0   jsr     grefbyt         ; get low link
        cmp     link            ; test
        bne     ..xloop         ; no match

                        ; print reference: page #,line #
        bit     lablcc          ; fill needed?
        bpl     ..nfl           ; nope
         jsr    leftmr          ; test margin
         lda    #20             ; for looks
         jsr    space
         lda    #4              ; set 5 columns
         sta    lablcc

..nfl   lda     xrfchar         ; out leading char
        sta     xrfrdo          ; shows referenced
        jsr     charo

        bit     rfpagf          ; page #s?
        bmi     ..onlyl         ; nope
         lda    rfpgn+1         ; print page #
         ldx    rfpgn
         ldy    #3              ; format size
         jsr    pdnumf
         lda    #','            ; and comma
         jsr    charo

..onlyl ldx     lastl           ; print line #
        lda     lastl+1
        jsr     pdnum           ; print it
        sec                     ; format it
        lda     #6
        sbc     svn2
        jsr     space           ; two spaces
        dec     lablcc          ; pop columns
        bpl     ..xloop         ; same row

        jsr     tstejcr         ; new line
                                ; high bit indicates
                                ; space fill needed.
        lda     #$80            ; five columns
        sta     lablcc
        bmi     ..xloop         ; unc

                        ; handle x-ref tokens
..xlo1  cmp     #rfnum          ; single byte offset?
        bcc     ..xlo3          ; no
        beq     ..xlo2
        eor     #-1             ; find true difference
        adc     #0              ; cs- add 1
..xlo0  clc                     ; add into lastl
        adc     lastl
        sta     lastl
        bcc     ..xlopz         ; cont
         inc    lastl+1
..xlo7  bne     ..xlopz         ; unc

..xlo4  cmp     #rfpag          ; page flag?
        bne     ..xlo5          ; no
        inc     rfpgn           ; just bump & cont
        bne     ..xlopz
         inc    rfpgn+1
         bne    ..xlo7          ; unc

..xlo2  jsr     grefbyt
        sta     lastl+1         ; hi first
         jsr    grefbyt
         sta    lastl
        bne     ..xlo7          ; unc

..xlo3  cmp     #rfnub          ; two byte offset?
        bne     ..xlo4
         jsr    grefbyt
         bne    ..xlo0          ; go add

..xlo5  cmp     #rflab          ; label flag?
        bne     ..xlo6
        lda     #'#'
        sta     xrfchar
        bne     ..xlo7          ; unc

..xlo6  lda     #0              ; must be eof flag
        cmp     xrfrdo          ; any references?
        beq     ..xlo8          ; no, do cr
         bit    lablcc          ; just did cr?
         bmi    gf_xit          ; yup (return)
..xlo8  sta     lablcc          ; this forces a cr
        rts
        endproc


;       get x-ref byte in a, return ne
;       ------------------------------
grefbyt proc
        ldx     #0
        lda     (buffer,x)
gref1   inc     buffer
        bne     gf_xit
         inc    buffer+1
gf_xit  rts
        endproc


;       set buffer addr
;       ---------------
setbf   proc
        lda     memtop+1        ; load addr
        ldy     memtop
        sta     buffer+1        ; init buffer addr
        sty     buffer
        rts
        endproc



;       assembler list entry point
;       --------------------------
asmlst: proc
asmlst1 ldy     #2              ; line length
        lda     (exbuf),y
        ldx     exbuf           ; get pointer to
        ldy     exbuf+1         ; token line

        sta     bank3           ; ^^^^^^^^^ bank 3 ^^^^^^^
        jsr     _edtv           ; convert it
        jsr     rightmr         ; print to eol & ret
        sta     bank2           ; ^^^^^^^^^ bank 2 ^^^^^^^
        rts
        endproc


;       editor list entry point
;       -----------------------
movin:  proc
        jsr     getoset         ; get inter len
        ldx     stmcur          ; mv stmcur for
        ldy     stmcur+1        ; conversion

edtv    sta     bank3           ; ^^^^^^^^^ bank 3 ^^^^^^^
        jsr     _edtv
        sta     bank1           ; ^^^^^^^^^ bank 1 ^^^^^^^
        rts
        endproc


;       dump symbols routines
;       ---------------------
formtop: proc
        clc
        lda     smbltp          ; form topsort
        adc     spointer-'?'*2,x        ; cc
        sta     labelptr
        tya
        adc     smbltp+1
        sta     labelptr+1
        rts
        endproc


;       form from labelptr a enters hi
;       ------------------------------
; out:
;       AY      = whatever

formlab: proc
        tax                     ; save high
        clc
        dey                     ; point to low
        lda     smbltp          ; add base to
        adc     (labelptr),y    ; offset
        tay                     ; save low
        txa                     ; do high
        adc     smbltp+1
        rts
        endproc


;       mac tab routines & ifs/getslen
;       ------------------------------
settab0: proc
        lda     txttab          ; pnt to first
        sta     stmcur          ; line in memory
         lda    txttab+1
         sta    stmcur+1
        lda     #4              ; init obj ptr
        sta     objptr

settab: lda     #8              ; set listing
        sta     tab1            ; tabs
        lda     #12
        sta     tab2
        lda     #20
        sta     tab3

setztab:
        lda     tab1            ; set z-page tabs
        sta     ztab1
        lda     tab2
        sta     ztab2
        rts
        endproc



;       get string length and increment 'cox' past string
;       -------------------------------------------------
; note:
;       'ifs' is used by .if

getslen: proc
        jsr     gtoke
ifs:    and     #$7f            ; this is length
        tay                     ; save
        clc                     ; adjust cox
        adc     cox
        sta     cox
        tya                     ; ret: a =length
        rts
        endproc


;       put in buffer table entries
;       ---------------------------
;       a       = table token index

lookup: proc
        ldy     #high function
        ldx     #low function   ; function
        sec                     ; subt relative
        sbc     #ch+1           ; table offset

lookup1:
        sta     bank1           ; ^^^^^^^^^ bank 1 ^^^^^^^ where tables are

        sta     fr0+2           ; save table index
        sty     fr0+1           ; start of table (h)
        txa                     ; low of table
        ldx     fr0+2           ; entry counter

..loo1  sta     fr0             ; table low
        dex                     ; pointing @ entry?
        bmi     ..loo4          ; yup, put in buff

        ldy     #-1             ; find end of entry
..loo5  iny
        lda     (fr0),y         ; x
        bpl     ..loo5          ; keep going

        sec                     ; up-date pointer
        tya                     ; new ptr = y+1+
        adc     fr0             ; old ptr
        bcc     ..loo1
         inc    fr0+1           ; bump high
        bcs     ..loo1          ; unc

..loo4  ldx     fr0+2           ; recover token

loo3:   ldy     #0              ; clear index
        sta     bank1           ; ^^^^^^^^ bank 1 ^^^^^^^^ where tables are
        sty     fr0+2

..loo2  ldy     fr0+2           ; move string to
        inc     fr0+2           ; holding buffer
        lda     (fr0),y
        pha                     ; save stat
        jsr     p7f             ; store it
        pla
        bpl     ..loo2          ; br; not done
        sta     bank3           ; ^^^^^^^^^^^ bank 3 ^^^^^^^ to calling bank
        rts
        endproc


;       put object to buffer
;       --------------------
object: proc
        ldy     objix
        sta     tpobuf+1,y
        inc     objix
        rts
        endproc


;       object buffer handler
;       ---------------------
; note:
;       fills buffer, then writes

inbuffo proc
        ldy     objptr
        sta     (objbuf),y
        inc     objptr
        beq     ..wrt           ; write this buf
         rts
..wrt   jmp     write
        endproc
$title(MAC65/XL Cartridge  (c) 1988 ICD Inc)
$subtitle(Misc segamnets for MAC65)

        seg     switch
;====================================
;       Bank switching vectors
;====================================

;       call listerr
;       ------------
listerr: proc
        lda     romf+$0fff      ; get bank #
        pha                     ; & save
        sta     bank1           ; get list bank
        jsr     _listerr        ; do it
        pla
        tax
        sta     bank1,x
        rts
        endproc


;       call asm
;       --------
asmx:   proc
        sta     bank2
        jmp     _asmx
        endproc


;       editor get numbers
;       ------------------
_getne: proc
        sta     bank2
        jsr     getne
        jmp     ..j

_getn:  sta     bank2
        jsr     getn
..j     sta     bank1
        rts
        endproc



$eject
        seg     carvec          ; cart vectors at $BFFA
;=====================================
;       Cartridge vector point
;       ----------------------
        dw      cart
        db      0
        db      $05
        dw      carti


;========================================
;       Cartridge bank identifiers
;       --------------------------
        seg     banka_v
        db      bank1 & 15

        seg     bankb_v
        db      bank2 & 15

        seg     bankc_v
        db      bank3 & 15


$eject
        seg     copy_right
;=================================
;       Copyright message
;=================================
mes:    db      $7d,'MAC/65',tab,tab, 'Version 1.0',$B1
mes1    db      '(c) 1984 Stephen D. Lawrow, OSS, Inc', $AE
        db      $7F

dos:    jmp     (dosvec)



$eject
        seg     initial
;==================================================
;       Initz and cold/warm cart entry points
;==================================================
        jmp     warm            ; warm entry without car

;       cartridge init
;       --------------
carti   proc
        lda     coldst          ; power-up?
        beq     ..ci            ; nope!
        lda     #$ca            ; set up default colors
        sta     colors
        lda     #$94
        sta     colors+1
        lda     #cntrlc         ; abort=control-c
        sta     hcntrlc

cartrtn:
        lda     #high xcio      ; set (cio)
        sta     xcii+1
        lda     #low xcio
        sta     xcii
..ci    $if cbug
         nop
        $else
         rts
        $endif
        endproc


;       cartridge run
;       -------------
cart    proc
        sta     bank1           ; enable syntax bank
        $if ddt?
         ldx    warmflg         ; cartridge cold?
         beq    ..ddt           ; yup, don't check ddt
         lda    zpsflg          ; in ddt?
         bne    ..ddt           ; nope
         jmp    ddtwrm
..ddt
        $endif

        cld                     ; clear dec mode
        cli                     ; enable interrupts
        ldx     #-1             ; set stack
        txs
        stx     brkkey          ; clear break key
        inx                     ; x=0  list to e:
        stx     lstdv
        stx     entdv           ; enter dev =0
        stx     eaf             ; set edit mode

;       $if (low mes) <> 0      ; CAUTION assume that MES starts at page bound
;        ldx    #low mes        ; print start-up msg
;       $endif

        jsr     ptsc            ; at cart entry
        ldx     #low mes1       ; part 2 of msg
        jsr     ptsc

;       $if (high mes) <> (high mes1)   ; CAUTION assume that MES and MES1 on
;        $error "start-up msg error"    ;  same page!
;       $endif

        lda     warmflg         ; cart cold start?
        bne     warm            ; nope!
         jsr    init            ; set mem ptrs & edit mode
        endproc


;       warm start
;       ----------
warm:   proc
        jsr     settab          ; set list tabs
        jsr     reset           ; init some vars

prmpt:  lda     #3              ; enable keyboard
        sta     skctl
        sta     bank1
        jsr     docr
        ldx     #low edit       ; print prompt
        ldy     #high edit
        bit     txtflag         ; we in text mode?
        bpl     ..wrm0          ; nope
         ldx    #low txtmd      ; else, print text:

;       $if (high txtmd) <> (high edit)
;        ldy    #high txtmd
;       $endif

..wrm0  jsr     ptse
        lda     colors          ; set color
        sta     color1
        lda     colors+1
        sta     color2

noprm:  sta     bank1           ; syntax bank
        ldx     #4              ; set tabs!
..wrm1  lda     oss,x
        sta     tabmap,x
         dex
         bpl    ..wrm1

        txs                     ; set stack
        inx                     ; x=0
        stx     epf             ; clear rep mode
        stx     eaf             ; set edit mode
        stx     dspflg          ; display ctrl chars
        stx     cpad            ; comment pad=0
        jsr     sd1             ; no sound
        jsr     input           ; get line of input
        bne     noprm           ; br if break key

very:   jsr     lnum?           ; numed line?
        bcs     pedit           ; no, try command
         jmp    ptext           ; process text
        endproc


;       process editor command
;       ----------------------
pedit   proc
        jsr     konvert         ; fix case
        jsr     skpspace
        ldx     #2              ; go search
        lda     #high command   ; command table
        ldy     #low command
        jsr     search
        bcs     what            ; no good!

        stx     cix             ; new text index
        lda     svf2            ; table entry
        pha                     ; save for exec
        ldy     #0              ; get syntax adr
        lda     (srcadr),y      ; from keyword table
        sta     saxpc
        iny
        lda     (srcadr),y
        sta     saxpc+1
        jsr     stc             ; syntax it
        pla                     ; pop table entry
        bcs     what            ; abort if bad
        jsr     exec            ; execute
        jmp     prmpt           ; and re-prompt
what    jmp     _what           ; print "what!"
        endproc



;       cold start initialization
;       -------------------------
init    proc
        ldx     memlow          ; get lomem
        stx     svlomem         ; and save
        ldx     memlow+1
        stx     svlomem+1

;       editor new
;       ----------
newx:   lda     #0              ; new clears
newtxt: sta     txtflag         ; text mode

enttxt: ldx     svlomem         ; set lomem
        ldy     svlomem+1
        stx     lomem
        sty     lomem+1
        iny                     ; for exbuff
        txa                     ; x
        ldx     #objbuf         ; set other ptrs
        jsr     cs1
        ldx     #txttab         ; expand by 256
        ldy     #0              ; for obj buff
        lda     #1
        jmp     expand          ; expand & return
        endproc


;       clean-up for return to editor
;       -----------------------------
reset:  proc
        lda     #0
        ldx     #group
..rss   sta     eaf,x           ; clear all key
        dex
        bpl     ..rss

                        ; reset upper pointers
reset1: ldx     #txtend+2
        lda     txtend
        ldy     txtend+1        ; fall to cs1

                        ; init table pointers
cs1:    sta     0,x             ; set adr low
        inx
        sty     0,x             ; set adr high
        inx
        cpx     #memtop+2       ; at limit?
        bcc     cs1
        sty     aphm+1          ; and aphmi
        sta     aphm
        rts
        endproc
$title(MAC65/XL Cartridge  (c) 1988 ICD Inc)
$subtitle(Cartridge bank A for MAC65)


        seg     banka
;==================================
;       line statement tables
;       ---------------------
statement:
        db      eol | $80
        x_db    '.IF'
        x_db    '.ELSE'
        x_db    '.ENDIF'
        x_db    '.MACRO'
        x_db    '.ENDM'
        x_db    '.TITLE'
        x_db    ' '
        x_db    '.PAGE'
        x_db    '.WORD'
        x_db    '.ERROR'
        x_db    '.BYTE'
        x_db    '.SBYTE'
        x_db    '.DBYTE'
        x_db    '.END'
        x_db    '.OPT'
        x_db    '.TAB'
        x_db    '.INCLUDE'
        x_db    '.DS'
        x_db    '.ORG'
        x_db    '.EQU'

; 65C02 instructions

        x_db    'BRA'           ; branch always
        x_db    'TRB'
        x_db    'TSB'
        x_db    '.FLOAT'
        x_db    '.CBYTE'
        x_db    ';'
        x_db    '.LOCAL'
        x_db    '.SET'
        x_db    '*='
        x_db    '='
        x_db    '.='


;       This table is global because DDT needs it!
;       ------------------------------------------
mnetab: x_db 'JSR','JMP','DEC','INC'
        x_db 'LDX','LDY','STX','STY','CPX','CPY','BIT'
        x_db 'BRK','CLC','CLD','CLI','CLV','DEX','DEY'
        x_db 'INX','INY','NOP','PHA','PHP','PLA','PLP'
        x_db 'RTI','RTS','SEC','SED','SEI','TAX','TAY'
        x_db 'TSX','TXA','TXS','TYA','BCC','BCS','BEQ'
        x_db 'BMI','BNE','BPL','BVC','BVS'
        x_db 'ORA','AND','EOR','ADC','STA','LDA','CMP'
        x_db 'SBC','ASL','ROL','LSR','ROR'
        db $80          ; comment

        x_db 'STZ'
        x_db 'DEA','INA'
        x_db 'PHX','PHY'
        x_db 'PLX','PLY'
        db 0


;============================
;       function table
;============================
; order in table is relevent!

function:
        x_db    "'"     ; ascii literal
        x_db    '%$'    ; macro string
        x_db    '%'     ; macro parameter
        x_db    '*'     ; current asm pc
        x_db    ' ',' '
        x_db    'a'
        x_db    'q'
        x_db    '+'     ; addition
        x_db    '-'     ; subtraction
        x_db    '*'     ; mutliplication
        x_db    '/'     ; division
        x_db    '&'     ; logical and
        x_db    '.DEF'  ; unary lab def
        x_db    '='     ; comparison
        x_db    '<='    ; less or equal
        x_db    '>='    ; greater or equal
        x_db    '<>'    ; not equal
        x_db    '>'     ; less than
        x_db    '<'     ; greater than and
        x_db    '-'     ; unary minus
        x_db    '['     ; oper left para
        x_db    ']'     ; oper right para
        x_db    '.OR'   ; basic-like or
        x_db    '.AND'  ; basic-like and
        x_db    '.NOT'  ; basic-like not
        x_db    '!'     ; logical or
        x_db    '^'     ; logical eor
        x_db    '.REF'  ; label reference
        x_db    '\'     ; modulo

        db      $80,$80,$80,$80 ; filler

; inbedded tables (save 3 bytes)
; ------------------------------
edit:   x_db    'Edit'
wh      x_db    'What!'
txtmd:  x_db    'Text:'

        x_db    '.REF '         ; label ref list
        x_db    '.DEF '         ; define list
        x_db    '.NOT '         ; not list
        x_db    ' .AND '        ; and list
        x_db    ' .OR '         ; or list
        x_db    ' <'            ; low byte
        x_db    ' >'            ; high byte

; begin terimators
; ----------------
        x_db    ',X)'           ; indirect ,x
        x_db    '),Y'           ; indirect ,y
        x_db    ',Y'            ; indexed ,y
        x_db    ',X'            ; indexed ,x
        x_db    ')'             ; indirect abs
        x_db    ','             ; comma syntax
        db      eol | $80       ; end of line
        x_db    ','             ; comma list
        x_db    '#'             ; imm/dev follows
        x_db    'A'             ; accumulator adressing
        x_db    '('             ; left para
        x_db    '"'             ; text string
        x_db    '$'             ; hex follows
        x_db    'Q'             ; rep querry
        x_db    'NO'            ; option off
        x_db    'NO '           ; list no
        x_db    'OBJ'           ; object
        x_db    'ERR'           ; error
        x_db    'EJECT'         ; page eject
        x_db    'LIST'          ; asm listing
        x_db    'XREF'          ; x-ref
        x_db    'MLIST'         ; macro list
        x_db    'CLIST'         ; conditional l.
        x_db    'NUM'           ; assembler auto-num
        x_db    'M'             ; merge
        db      0       ; end of table


;       statement syntax adress table
;       -----------------------------
linetab equ     $-2
        dw      ify-1           ; if
        dw      elsey-1         ; else
        dw      elsey-1         ; endif
        dw      macroy-1        ; macro define
        dw      elsey-1         ; endm
        dw      pagey-1         ; title
        dw      macroxy-1       ; macro exec
        dw      pagey-1         ; page
        dw      wordy-1         ; word
        dw      pagey-1         ; error
        dw      bytey-1         ; byte
        dw      bytey-1         ; sbyte
        dw      wordy-1         ; dbyte
        dw      endly-1         ; end
        dw      opty-1          ; opt
        dw      taby-1          ; tab
        dw      includy-1       ; include
        dw      equy-1          ; ds
        dw      equy-1          ; org
        dw      equy0-1         ; equ
        dw      equy-1          ; bra
        dw      equy-1,equy-1   ; trb & tsb
        dw      floaty-1        ; float
        dw      bytey-1         ; cbyte
        dw      endly-1         ; comment
        dw      endly-1         ; local
        dw      sety-1          ; system set
        dw      equy-1          ; origin
        dw      equy0-1         ; equate
        dw      equy0-1         ; set equate

        dw      jsry-1          ; jsr instr
        dw      jmpy-1          ; jmp instr
        dw      incy-1          ; dec instr
        dw      incy-1          ; inc instr
        dw      ldxy-1          ; ldx instr
        dw      ldyy-1          ; ldy instr
        dw      stxy-1          ; stx instr
        dw      incy-1          ; sty instr
        dw      cpxy-1          ; cpx instr
        dw      cpxy-1          ; cpy instr
        dw      bity-1          ; bit instr
        dw      brky-1          ; type implied
        dw      bccy-1          ; type branch
        dw      oray-1          ; type multi
        dw      asly-1          ; type accumulator



;       command statement table
;       and syntax table adress
;       -----------------------
command:
        dw      eoly-1          ; bye
         x_db   'BYE'
        dw      listy-1         ; list
         x_db   'LIST'
        dw      eoly-1          ; new
         x_db   'NEW'
        dw      dely-1          ; del
         x_db   'DEL'
        dw      findy-1         ; find
         x_db   'FIND'
        dw      asmy-1          ; asm
         x_db   'ASM'
        dw      includy-1       ; save
         x_db   'SAVE'
        dw      entery-1        ; load
         x_db   'LOAD'
        dw      entery-1        ; enter
         x_db   'ENTER'
        dw      includy-1       ; bload
         x_db   'BLOAD'
        dw      bsavey-1        ; bsave
         x_db   'BSAVE'
        dw      numy-1          ; num
         x_db   'NUM'
        dw      listy-1         ; print
         x_db   'PRINT'
        dw      numy-1          ; ren
         x_db   'REN'
        dw      repy-1          ; rep
         x_db   'REP'
        dw      eoly-1          ; size
         x_db   'SIZE'
        dw      lomemy-1        ; lomem
         x_db   'LOMEM'
        dw      eoly-1          ; cp
         x_db   'CP'
        dw      eoly-1          ; dos
         x_db   'DOS'

        $if 0
         dw     dy-1    ; d - display
          x_db  'D'
         dw     cy-1    ; c - change
          x_db  'C'
        $endif

        dw      qy-1            ; convert
         x_db   '?'
        dw      eoly-1          ; text command
         x_db   'TEXT'
        dw      eoly-1          ; enter DDT
         x_db   'DDT'
        db 0,0,0


;       editor routine table
;       --------------------
extabl  dw      ataribye-1      ; bye
        dw      listx-1         ; list
        dw      newx-1          ; new
        dw      delx-1          ; del
        dw      findx-1         ; find
        dw      asmx-1          ; asm
        dw      savex-1         ; save
        dw      loadx-1         ; load
        dw      enterx-1        ; enter
        dw      bloadx-1        ; bload
        dw      bsavex-1        ; bsave
        dw      numx-1          ; num
        dw      printx-1        ; print
        dw      renx-1          ; ren
        dw      repx-1          ; rep
        dw      sizex-1         ; size
        dw      lomemx-1        ; lomem
        dw      dos-1           ; cp
        dw      dos-1           ; dos

        $if 0
         dw     dx-1    ; display
         dw     cx-1    ; change
        $endif

        dw      qx-1            ; convert
        dw      textx-1         ; text mode
        dw      ddt-1           ; dunion's ddt



;       Delete line
;       -----------
dell:   proc
        jsr     fndli           ; find line
        bcs     os              ; br if not there
        jsr     getoset         ; move size
        pha
        jsr     adoff           ; set from adr
        pla
        jsr     contxt          ; go contract
        lda     #$40            ; show rep that
        sta     rpif            ; line was deleted
os      jmp     oy              ; unc, test rep
        endproc


;       print 'what'
;       ------------
_what:  proc
        lda     #$35            ; sound effect
        ldx     #$ac
        jsr     sd
        ldx     #low wh
        ldy     #high wh
        jsr     ptse
        jmp     prmpt           ; go re-prompt
        endproc


;       sound!
;       ------
sd1:    proc
        txa                     ; sound off x=0
sd      sta     pokey           ; channel 1
        stx     pokey+1
        rts
        endproc


;       execute editor command
;       ----------------------
exec:   proc
        asl     a
        tay                     ; get index to table
        lda     extabl+1,y      ; get adr
        pha                     ; and stack
        lda     extabl,y
        pha
        lda     cox             ; save for rep
        sta     cmdlg
        ldy     #0              ; reset out-put
        sty     cox             ; index
        rts                     ; go to it
        endproc


;       editor text mode
;       ----------------
textx   proc
        lda     #$ff
        jmp     newtxt
        endproc


;       execute syntax
;       --------------
stc:    proc
        lda     saxpc+1         ; init to
        sta     spc+1           ; first level
        lda     saxpc
        sta     spc
        lda     cox
        sta     sox
        lda     cix
        sta     six
        ldy     #0              ; clr:
        sty     curtoke         ; current token
        sty     condptr         ; cond text index
        sty     actptr          ; current text index
        sty     stklvl          ; set at first level

sloop   jsr     getop
        bmi     sbra            ; syntax branch
        cmp     #1              ; absolute?
        bcc     sabs            ; yes, abs internal
        bne     cksng           ; go check

        jsr     sabs            ; go abs external
        bcc     sloop           ; sucessful test
        jmp     sterm           ; look for term

cksng   cmp     #5
        bcc     sret            ; go ret
        beq     pop             ; go pop stack

        jsr     stest           ; go test
        bcc     sloop           ; sucessful test
        jmp     sterm           ; look for term

pop     lda     stklvl          ; don't pop below
        beq     sloop           ; first level
        sbc     #4              ; cs, next lower level
        tax
        lda     sox+4,x         ; copy input &
        sta     sox,x           ; output indices to
        lda     six+4,x         ; this level
        sta     six,x
        stx     stklvl
        jmp     sloop           ; and get nxt

sbra    sec                     ; compute new
        ldx     #0              ; table addr
        sbc     #@+1
        bcs     bck             ; a back branch
         dex
bck     clc
        adc     saxpc           ; form new pc addr
        pha                     ; save
        txa
        adc     saxpc+1
        pha                     ; save
        jmp     spush           ; stack current

sabs    jsr     getop           ; get adress
        pha
        jsr     getop
        pha
        bcc     spush           ; br if internal
         rts                    ; go absolute

v0      lda     #15
        jmp     fter            ; fatal err fix

spush   ldx     stklvl          ; inc stack pointer
        inx
        inx
        inx
        inx
        bmi     v0              ; stack overflow!
        stx     stklvl
        lda     cix             ; save current ptrs
        sta     six,x
        lda     cox
        sta     sox,x
        lda     saxpc
        sta     spc,x
         lda    saxpc+1
         sta    spc+1,x
        pla                     ; restore new addr
        sta     saxpc+1
        pla
        sta     saxpc
        jmp     sloop           ; continue...

sret    ldx     stklvl          ; at first level?
        bne     sret1           ; no, else
        rts                     ; syntax all done!

sret1   lda     spc,x           ; restore old
        sta     saxpc           ; table adress
        lda     spc+1,x
        sta     saxpc+1
        dex                     ; pop stk ptr
        dex
        dex
        dex
        stx     stklvl          ; save it
        bcs     sterm           ; br if bad
         jmp    sloop           ; else, continue


sterm   jsr     getop
        bmi     sterm           ; skip branch
        cmp     #2              ; abs reference?
        beq     v1              ; go 'or'
        bcs     stwo
        jsr     incpt           ; skip absolute
        jsr     incpt           ; adress
        bne     sterm           ; unc

stwo    cmp     #3              ; return?
        beq     sret            ; yup
        cmp     #4+1            ; an 'or-return'?
        bcs     sterm           ; no, keep skipping

        jsr     restr           ; restore last valid
        clc                     ; set sucess
        jmp     sret            ; and go return

v1      jsr     restr
        jmp     sloop           ; continue...
        endproc



;       routines getop & incpt
;       ----------------------
getop   proc
        jsr     incpt           ; bump syntax pc
        ldx     #0
        lda     (saxpc,x)       ; get code
        rts

incpt   inc     saxpc
        bne     ..2
         inc    saxpc+1
..2     rts
        endproc


restr   proc
        ldy     cix
        jsr     valid           ; set valid text
        ldx     stklvl          ; restore indices
        lda     six,x
        sta     cix
        lda     sox,x
        sta     cox
        rts
        endproc


;       syntax externals
;       ----------------
; test for external assembly test,
;  or internal token match

;       6 = test for eol/comment
;               note: 6 is strict comment
;               6,6 is loose comment
;       7 = test for hex #
;       8 = test for dec #
;       9 = special change token

stest   proc
        cmp     #ch+1           ; internal tst?
        bcs     tsttok          ; yup

        tay                     ; get table indx
        lda     ..exh-6,y       ; get routine
        pha
        lda     ..exl-6,y
        pha
        rts                     ; and go to it!

; external assembly test table
; ----------------------------
..exl   db      low s@@end-1, low s@@hex-1
        db      low s@@dec-1, low s@@ch-1

..exh   db      high (s@@end-1), high (s@@hex-1)
        db      high (s@@dec-1), high (s@@ch-1)

; change token
; ------------
s@ch    dec     cox             ; change level before
s@@ch   jsr     getop           ; get change token
        dec     cox             ; pop output indx
        clc                     ; set sucess

ptoke   ldy     cox             ; put in buff
        sta     (exbuf),y
        inc     cox
        beq     ..tob           ; overflow!
         rts
..tob   lda     #8              ; buff ovrflo
        jmp     fter            ; fatal err fix
        endproc


;       test for internal token match
;       -----------------------------
tsttok  proc
        jsr     skpspace
        cpy     actptr
        beq     nsearch         ; don't search
        sty     actptr          ; actual text
        lda     #high function
        ldy     #low function
        ldx     #0
        jsr     search
        bcs     notfnd          ; not there

        stx     condptr         ; text conditional
        lda     svf2            ; table indx
        adc     #ch+1           ; first function
        sta     curtoke

nsearch ldy     #0              ; comp this token
        lda     (saxpc),y
        cmp     curtoke
        beq     match           ; match
        sec                     ; else, set fail
        rts

match   ldx     condptr
        stx     cix
        jsr     ptoke
        clc
        rts

notfnd  lda     #0              ; set null token
        sta     curtoke
        rts
        endproc


;       search for statement
;       --------------------
ssrch   proc
        jsr     skpspace        ; table search
        ldx     #0
        lda     #high statement
        ldy     #low statement
        jsr     search
        stx     cix
        lda     svf2
        rts
        endproc


search: proc
        stx     skipfc          ; skip factor
        ldx     #-1             ; set -1 entry
        stx     svf2            ; x
        sta     srcadr+1        ; adress table high

nx99    sty     srcadr          ; adress table low
        inc     svf2            ; inc entry
        ldx     cix             ; pointer in buffer
        ldy     skipfc          ; skip
        lda     (srcadr),y      ; end of tab?
        beq     u0              ; yup

v8      lda     lbuff,x         ; get buff char
        and     #$7f            ; seven bits
        eor     (srcadr),y      ; compare
        asl     a               ; save end stat
        beq     v7              ; br if match,
        bcs     v9              ; if symbol end

nx00    iny                     ; fail, look for
        lda     (srcadr),y      ; symbol end
        bpl     nx00

v9      sec                     ; y=y+1
        tya                     ; pnt to nxt symb
        adc     srcadr
        tay                     ; y=low
        bcc     nx99            ; save time!
        inc     srcadr+1        ; if necessary
        bne     nx99            ; unc

v7      iny                     ; set for nxt
        inx                     ; compare
        bcc     v8              ; if not end
        clc                     ; else, sucess
        rts

u0      sec                     ; fail
        rts
        endproc


;       external routine dec num test
;       -----------------------------
s@@dec  proc
        jsr     skpspace        ; skip spaces
decasy  sty     svn2            ; save current text indx
        jsr     afp             ; to floatint pt
        bcs     badnum          ; br if bad
        jsr     fpi             ; cvrt to intger
        bcc     ..ok            ; go save token

badnum  ldy     svn2            ; bad #, restore text
        sty     cix             ; carry was set
        jmp     valid           ; go do valid pointer

                        ; external hex or dec num test
hxdc    jsr     skpspace        ; hex or dec #
        lda     (inbuff),y      ; get character
        cmp     #'$'            ; dollar sign?
        bne     decasy          ; no, try dec num
        inc     cix             ; skip '$'
        bne     s@@hex          ; unc to tst hex

..ok    lda     #tdb            ; set as byte
hexent  ldx     fr0+1           ; is it byte?
        bne     ..ok0           ; no, do word
        jsr     ptoke           ; save num type
        lda     fr0             ; and save num
        jmp     ..h0            ; go

..ok0   sec
        sbc     #1              ; set as word
        jsr     ptoke           ; num type
ok9     lda     fr0             ; save num
        jsr     ptoke
        lda     fr0+1
..h0    jsr     ptoke
        clc                     ; sucess
        rts
        endproc


;       external test hex num
;       ---------------------
s@@hex  proc
        jsr     skpspace
        sty     svn2            ; save inbuff index
        lda     #0              ; clr result
        sta     ztemp4
        sta     ztemp4+1
        jsr     getnib          ; first must
        bcs     badnum          ; be valid
..ar    clc                     ; add in result
        adc     ztemp4+1
        sta     ztemp4+1
        bcc     ..v6
        inc     ztemp4
..v6    jsr     getnib          ; get next
        bcs     ..as            ; done!

        ldx     #4              ; shift into position
..sh    $if atari
         jsr    ilshift
        $else
         asl    ztemp4+1
         rol    ztemp4
        $endif
        bcs     badnum          ; oops- bad num
        dex
        bne     ..sh
        beq     ..ar            ; unc get nxt

..as    lda     ztemp4          ; move result
        sta     fr0+1
        lda     ztemp4+1        ; (sic)
        sta     fr0
        dec     cix             ; point back
        lda     #thb            ; set hex byte
        jmp     hexent          ; go process
        endproc


getnib  proc
        ldy     cix             ; get hex nibble
        inc     cix
        lda     (inbuff),y
        sec
        sbc     #'0'
        bmi     ..x9
        cmp     #$0a
        bcc     ..x8
        sbc     #7
        cmp     #$0A
        bcc     ..x9
        cmp     #$10
        bcc     ..x8
..x9    sec
..x8    rts
        endproc



;       external test for string
;       to delimiter as follows:
;       ------------------------
; ffnd - finds same as first
; fsav - end in <
; fspc - end in ,
; frep - end same as ffnd
; fstr - end in '
;
ffnd    proc
        lda     #$00            ; no null length & get from tpobuf
        sta     temp
        jsr     skpspace
        lda     (inbuff),y      ; get char
        cmp     #eol            ; eol?
        beq     bad6            ; illegal!
        bne     fv1             ; look for same

fsav    lda     #'<'            ; get test char
        bne     fv4             ; unc

fspc    jsr     skpspace
        lda     #','
fv4     ldy     #0              ; no null
        sty     temp            ; x
        jsr     fv0             ; go find it
        bcs     cllen           ; go put str len
qq      dec     cix             ; point back to ,
        rts

frep    dec     cix             ; point to last
        lda     #$c0            ; set for ''
        ldx     #'"'            ; was delimiter ''?
        cpx     svf2
        beq     ..fep           ; yes, get from tpobuf
         lda    #$40
..fep   sta     temp            ; allow null
        jmp     fv2

fstr    lda     #0              ; dis-allow null
        sta     temp            ; x
        lda     #'"'            ; double quotes
        jsr     fv0             ; find them
        bcc     qq              ; sucess
        rts
        endproc



fv0     proc
        dec     cix             ; back up 1 before
fv1     sta     svf2            ; save delimiter
fv2     tax                     ; save
        lda     cox             ; save output indx
        sta     svn2            ; x
        inc     cox             ; dummy for len
fv3     inc     cix             ; get char
        ldy     cix

        bit     temp            ; where to get chars from
        bpl     ginb
        lda     tpobuf,y        ; rep only
        bne     ginc            ; unc, unless 'cntrl-,'

ginb    lda     (inbuff),y
ginc    cmp     #eol            ; eol?
        beq     bad6            ; yup, error
        cmp     svf2            ; delimiter?
        beq     fv5             ; yup, compute len
        jsr     ptoke           ; save str char
        bne     fv3             ; unc

fv5     inc     cix             ; index to nxt
cllen   clc                     ; get str len
        lda     cox
        sbc     svn2
        bne     fv6             ; non-zero
        bit     temp            ; allow zero?
        bvc     ..cf            ; no, error
fv6     ldy     svn2            ; indx len in out
        ora     #$80            ; high bit on
        sta     (exbuf),y       ; str len +128
        clc                     ; sucess
        rts
bad6    sec                     ; fail
        rts

                        ; external tst for ascii char after single quote
tstasc  ldy     cix             ; buff indx
        lda     (inbuff),y      ; get char
        cmp     #eol            ; if eol, error
        beq     ..cf
..ts2   jsr     ptoke           ; else, save it
        inc     cix             ; point to nxt
..ts1   clc                     ; sucess
        rts

                        ; external for mandatory label
miss    ldy     #3              ; byte should
        lda     (exbuf),y       ; be minus
        bmi     ..ts1           ; if label is
        bpl     ..cf            ; present

                        ; external for mandatory no label
miss0   jsr     miss            ; call tst lab &
        bcs     ..ts1           ; reverse stat
..cf    sec
        rts


;       external test for single unary
;       ------------------------------
only1
        ldy     cox
        dey                     ; get last token
        lda     (exbuf),y       ; can be used
        cmp     #@defl          ; only once in an
        beq     ..j             ; expression
        cmp     #@low
        beq     ..j
        cmp     #@high
        beq     ..j
        clc
..j     rts


;       test for asm comma
;       ------------------
tcoma
        jsr     skpspace
        sty     saxtx
        lda     (inbuff),y
        cmp     #','
        bne     ..cf            ; err
        lda     #@cma           ; put in buffer
        bne     ..ts2           ; unc
        endproc


illegal db      ' ,[]+-*/\!&^"'

s@@end  proc
        ldx     #12             ; set strict count
        ldy     #1              ; test for loose
        lda     (saxpc),y
        cmp     #6
        bne     ..k1            ; do strict comment

        jsr     incpt           ; loose comment type
        ldx     #0
..k1    jsr     skpspace
        lda     (inbuff),y
..k2    cmp     illegal,x
        beq     valid           ; no good
        dex
        bpl     ..k2            ; again or legal
yup     cmp     #eol            ; eol?
        bne     comment         ; no, do comment
        inc     cox             ; dummy bump
        clc                     ; sucess!

                        ; up-date valid text pointer
valid   php                     ; save status
        cpy     saxtx           ; update valid text
        bcc     yop
         sty    saxtx
yop     plp                     ; get status
        rts
        endproc



;       syntax source line
;       ------------------
synli:  proc
        jsr     konvert         ; do lower case
        ldy     cix
        sty     svn2
        sty     saxtx           ; set valid text
        lda     (inbuff),y
        cmp     #$20            ; manadatory!
        beq     good            ; so far...

error   lda     #3              ; oops
        sta     cox             ; reset output buf
        ldy     saxtx           ; show where
        lda     (inbuff),y      ; syntax failed
        cmp     #eol            ; failed at end?
        bne     feol            ; no
        iny                     ; bump eol 1
        sta     tpobuf,y
        dey
        lda     #$20            ; space
feol    eor     #$80            ; flag it
        sta     tpobuf,y        ; save it
        lda     #0              ; err line token
        beq     ..mc            ; unc

comms   lda     #@scomment      ; statement
        bne     ..mc            ; comment
comment lda     #@comment
..mc    jsr     ptoke
..mc0   ldy     cix             ; move string
        lda     tpobuf,y        ; to exec buff
        jsr     ptoke
        inc     cix
        cmp     #eol            ; done?
        bne     ..mc0
        clc                     ; set sucessful
..ble   rts

                        ; check label/statement/comment
good    inc     cix             ; get nxt char
        inc     saxtx           ; valid text ptr
        ldy     cix
        lda     (inbuff),y
        cmp     #';'            ; semi-colon
        beq     comms
        cmp     #'*'            ; or asterisk ok
        beq     comms           ; for comment
        bit     txtflag         ; in text mode?
        bmi     comms           ; yup, do comment

        tax                     ; save a
        tya                     ; save cix for
        pha                     ; possible error

        lda     #high (nlabel-1)        ; set for no label
        ldy     #low (nlabel-1)
        cpx     #$20
        beq     ..nola
         dey                    ; YA = (label-1)

..nola  sta     saxpc+1
        sty     saxpc
        jsr     stc             ; do syntax
        pla
        bcc     ..ble           ; go rts
        sta     cix             ; reset inbuff pointer
        jmp     error
        endproc


;       fnd func. & synt
;       ----------------
funt    proc
        jsr     skpspace
        sty     svn2
        sty     saxtx           ; set valid ptr

        jsr     ssrch           ; srch statements

        bcc     fnt
        lda     #@macrox        ; asume macro call

fnt     jsr     ptoke           ; save statement
        cmp     #@seol          ; for label eol
        beq     ysgd

                        ; divide syntax by instr type
        cmp     #@brk           ; a single instruction type?
        bcc     fun1            ; yup
        ldx     #@brk           ; pre-set indx
..yl    cmp     ..itp-@brk,x    ; find type
        bcc     fun0            ; yup
        inx                     ; bump x
        cpx     #@brk+3         ; done?
        bcc     ..yl            ; no

        cmp     #@rockwell      ; new instr?
        bcc     fun0            ; nope
        bne     fun2
                                ; the stz instr
        ldx     #@bit           ; same syntax
         db $2c
fun2    ldx     #@brk           ; get implied type

fun0    txa
fun1    asl     a               ; double for table index
        tay                     ; to y
        lda     linetab,y       ; get syntax routine
        sta     saxpc           ; adress
        lda     linetab+1,y
        sta     saxpc+1
        pla                     ; pop old return
        pla
        jsr     stc             ; do syntax
        bcs     nogd            ; bad!

ysgd    clc                     ; sucess
nogd    rts

..itp   db      @bcc,@ora,@asl
        endproc



;       label text routines (tstlabi for prefix labels)
;       -----------------------------------------------
tstlab  proc
        jsr     skpspace        ; skip spaces
tstlabi ldy     cix
        sty     svn2            ; save inbuff
        sty     svf2

        lda     (inbuff),y      ; first character
        jsr     tstletter       ; must be alpha
        bcs     nolab           ; it isn't...

loop1   ldy     cix             ; get inbuff index
        inc     cix             ; bump pointer
        lda     (inbuff),y      ; get character
        jsr     tstletter       ; is it letter?
        bcc     loop1           ; yup, do next
        jsr     tstnum          ; no, number type?
        bcc     loop1           ; yup, get next
        lda     cox             ; save cox for macro
        tax
        dec     cix             ; pop input indx
        lda     cix             ; form lab len
        sbc     svn2            ; cs
        ora     #$80            ; flag label-type
        jsr     ptoke           ; store in outbuf

loop2   ldy     svn2            ; move to outbuf
        inc     svn2
        cpy     cix             ; done?
        bcs     cka             ; yes, test for 'a'
        lda     (inbuff),y      ; move character
        jsr     ptoke
        bne     loop2           ; unc

cka     txa                     ; test for label 'a'
        tay
        lda     (exbuf),y       ; check length
        cmp     #$81            ; one?
        bne     done            ; no, go sucess
        iny
        lda     (exbuf),y       ; check character
        cmp     #'A'            ; a?
        bne     done            ; no, sucess

        stx     cox             ; reset exbuf index
nolab   lda     svf2            ; reset inbuf index
        sta     cix             ; carry was set
        rts
        endproc



;       test letter & number
;       --------------------
tstletter proc
        cmp     #':'            ; colon
        beq     done
        cmp     #'_'            ; underline
        beq     done
        cmp     #'?'
        bcc     nochar
        cmp     #'['
        rts
done    clc                     ; sucess
        rts

..tst   cmp     #'.'
        beq     done
nochar  sec
        rts
                        ; test macro name label
tstlab0 jsr     tstlab          ; ck label
        bcs     nochar          ; bad
        txa                     ; x= cox at label
        tay
        lda     (exbuf),y       ; get lab size
        adc     #1              ; bump by 1 cc
        sta     (exbuf),y       ; store
        lda     #$20            ; trailing space
        jmp     ptoke           ; and ret

tstnum  cmp     #'0'            ; test decimal
        bcc     ..tst           ; number
         cmp    #'9'+1
        rts
        endproc



konvert: proc
        ldy     #$7f
sasis   lda     (inbuff),y
        sta     tpobuf,y
        dey
        bpl     sasis

        lda     epf             ; skip if rep mode
        cmp     #rmode
        beq     konv
                        ; change lower to upper
nextup  iny
        lda     (inbuff),y      ; get char
        cmp     #eol            ; done?
        beq     konv
        and     #$7f            ; normalize

        cmp     #'"'            ; double quote?
        bne     trylow
        sta     (inbuff),y      ; save quote
tryq    iny
        lda     (inbuff),y
        cmp     #eol            ; eol?
        beq     konv
        and     #$7f            ; search for
        cmp     #'"'            ; other '
        bne     tryq

trylow  cmp     #'a'            ; test lower
        bcc     putbck
        cmp     #'z'+1
        bcs     putbck
         adc    #-$20           ; subtract $20

putbck  cmp     #tab            ; tab?
        bne     khar            ; nope
        lda     #$20            ; rep with space
        sta     tpobuf,y        ; echo in back-up

khar    sta     (inbuff),y
        cmp     #'''            ; single quote?
        bne     nextup          ; no
        iny
        lda     (inbuff),y      ; next an eol?
        cmp     #eol
        bne     nextup          ; no
konv    rts
        endproc



;       execute syntax & table search syntax macros
;       -------------------------------------------

nully   db      1, high (bad6-1), low (bad6-1)  ; always return an error

; .float
; ------
floaty  db      1, high (fspc-1), low (fspc-1)  ; first # string
        db      $C0 + (fop0-$), 6, 3

fop0    db      @lcma, ch, @cma
        db      1, high (fspc-1), low (fspc-1)
        db      $C0 + (fop0-$), 4

; .tab & .set
; -----------
taby    db      0, low (sx-1), high (sx-1)
        db      1, high (tcoma-1), low (tcoma-1)

sety    db      0, low (sx-1), high (sx-1)
        db      1, high (tcoma-1), low (tcoma-1)
        db      $C0 + (equy-$), 3

; =, .=, *=
; ---------
equy0   db      1, high (miss-1), low (miss-1)  ; = & .= have label
equy    db      0, low (sx-1), high (sx-1)      ; expression
endy    db      6,3                             ; implied

; macro defination
; ----------------
macroy  db      1, high (miss0-1), low (miss0-1)        ; no label!
        db      1, high (tstlab0-1), low (tstlab0-1)    ; macro name
endly   db      6,6,3                                   ; loose comment

; .else & .endif & .endm
; ----------------------
elsey   db      1, high (miss0-1), low (miss0-1)        ; no label!
        db      $C0 + (endly-$), 3

; .if
; ---
sre     db      $C0 + (srf-$), @ref, ch, @refl
        db      1, high (tstlab-1), low (tstlab-1)      ; label reference
        db      3

srf     db      @not, ch, @notl
        db      4

ify     db      0, low (sx-1), high (sx-1)              ; expr
        db      $C0 + (endly-$), 3

; .opt
; ----
opty    db      $C0 + (j1-$)
        db      $C0 + (endy-$), 3

j1      db      $C0 + (j2-$)
        db      $C0 + (j3-$)
        db      $C0 + (j4-$), 3

j2      db      @lno, ch, @no
        db      4

j3      db      @obj, 2, @err, 2, @wait, 2
        db      @eject, 2, @list, 2, @num, 2
        db      @mlist, 2, @clist, 3

j4      db      1, high (tcoma-1), low (tcoma-1)
        db      $C0 + (j1-$), 4

; test byte modifier
; ------------------
modf    db      @plus                           ; plus
        db      0, low (sx-1), high (sx-1)      ; expression
        db      1, high (tcoma-1), low (tcoma-1)
        db      4

; macro execution
; ---------------
macroxy db      1, high (tstlab0-1), low (tstlab0-1)    ; macro name
        db      $C0 + (svr1-$), 6, 3

svr1    db      0, low (svr-1), high (svr-1)    ; exp or string
        db      $C0 + (byt1-$), 4               ; byte syntax

; .page & .title
; --------------
pagey   db      @qu
        db      1, high (fstr-1), low (fstr-1)
        db      @qu, 6, 2, 6, 3

; db & .sbyte
; -----------
bytey   db      $C0 + (modf-$)
        db      $C0 + (svr-$)
        db      $C0 + (byt1-$), 6, 3

byt0    db      5                                       ; pop syntax
byt1    db      1, high (tcoma-1), low (tcoma-1)
        db      $C0 + (svr-$)
        db      $C0 + (byt0-$), 4

; dw & .dbyte
; -----------
wordy   db      $C0 + (sx-$)
        db      $C0 + (wor1-$), 6, 3
wor0    db      5                                       ; pop syntax
wor1    db      1, high (tcoma-1), low (tcoma-1)
        db      $C0 + (sx-$)
        db      $C0 + (wor0-$), 4

; string or expression
; --------------------
svr     db      $C0 + (sx-$), 2, @qu                    ; expression
        db      1, high (fstr-1), low (fstr-1)          ; quoted string
        db      @qu, 2, @mstr
        db      0, low (bsx5-1), high (bsx5-1)          ; macro string
        db      3


;       syntax instructions
;       -------------------
jmpy    db      @lp
        db      $C0 + (sx-$)
        db      0, low (gp9-1), high (gp9-1)
        db      $C0 + (brky-$), 2

jsry    db      $C0 + (sx-$), 6, 3
cpxy    db      $C0 + (@pndt-$)
bccy    db      $C0 + (sx-$), 6, 3
brky    db      6, 6, 3

;       syntax expression
;       -----------------
sx      db      0, low (sre-1), high (sre-1)            ; ref label
        db      2
        db      0, low (bsx-1), high (bsx-1)            ; expr
        db      3

@pndt   db      @pnd, 4                         ; test for '#'

gp0     db      @lp
        db      $C0 + (sx-$)
        db      $C0 + (gp1-$), 2

gp2     db      @pnd
        db      $C0 + (sx-$), 2

gp3     db      $C0 + (sx-$)
        db      $C0 + (gcmax-$), 3

gp5     db      @pnd
        db      $C0 + (sx-$), 2

gp6     db      $C0 + (sx-$), @cmay, 2
        db      $C0 + (sx-$), 3

bity                                            ; new addr modes
incy    db      $C0 + (gp3-$), 6, 3
ldyy    db      $C0 + (gp2-$), 6, 3
asly    db      $C0 + (gp3-$), 6, 2
        db      @a, 6, 3

oray    db      $C0 + (sx-$), @cmay, 6, 2
        db      $C0 + (gp0-$), 6, 3
ldxy    db      $C0 + (gp5-$), 6, 3
stxy    db      $C0 + (gp6-$), 6, 3

gp1     db      @pcmay, 2
gp9     db      @pcmax, 2, @rp, 3
gcmax   db      @cmax, 4


;       syntax editor commands
;       ----------------------
tsq     db      @q,2,s.q,ch,@q,3
tsa     db      @a,2,s.a,ch,@a,3

dely    db      8
        db      $C0 + (a03-$), @eol, 3

a03     db      @lcma, 8, 4     ; dec num

findy   db      0, low (a11-1), high (a11-1)            ; filespec
        db      1, high (ffnd-1), low (ffnd-1)          ; string
        db      $C0 + (a02-$)                           ; line numbers
        db      $C0 + (a01-$)                           ; options

eoly    db      @eol, 3                 ; eol
repy    db      1, high (ffnd-1), low (ffnd-1)
        db      1, high (frep-1), low (frep-1)
        db      $C0 + (a02-$)
        db      $C0 + (a00-$), @eol, 3

a00     db      @lcma
        db      $C0 + (tsq-$), 2

a01     db      @lcma
        db      $C0 + (tsa-$), 4

asmy    db      $C0 + (a05-$), 2                ; (source)! or
        db      $C0 + (a04-$)                   ; (source or ,)
        db      $C0 + (le-$), 2                 ; (list)! or
        db      $C0 + (a04-$)                   ; (source or ,)
        db      $C0 + (l4-$)                    ; (list or ,)
        db      $C0 + (a05-$), 2                ; (object)! or

        db      $C0 + (a04-$)                   ; (source or,)
        db      $C0 + (l4-$)                    ; (list or ,)
        db      $C0 + (a04-$)                   ; (object or ,)
        db      $C0 + (a05-$), 2, @eol, 3       ; (xref)! or !

a02     db      8
        db      $C0 + (a03-$), 4
l4      db      $C0 + (ll-$), @lcma, 2
        db      $C0 + (a04-$), 3
le      db      $C0 + (ll-$), @eol, 2
        db      $C0 + (a05-$), 3

ll      db      @pnd, @minus                    ; rep @pnd
        db      1, high (s@ch-1), low (s@ch-1)  ; with @minus
        db      @minus, 3

a04     db      $C0 + (bloady-$), @lcma, 2
        db      @lcma, 3

a05     db      $C0 + (bloady-$), @eol, 3
a11     db      $C0 + (bloady-$), @lcma, 4
a13     db      8, @minus, @eol, 2
numy    db      $C0 + (a02-$)
        db      @eol, 3

bloady  db      @pnd
        db      1, high (fspc-1), low (fspc-1)
        db      3

entery  db      $C0 + (bloady-$)
        db      $C0 + (a06-$), @eol, 3

a06     db      @lcma, @m, 2
        db      @lcma, @a, 4

listy   db      $C0 + (a05-$), 2
        db      $C0 + (a11-$)
        db      $C0 + (a13-$), 3

bsavey  db      @pnd
        db      1, high (fsav-1), low (fsav-1)
        db      $C0 + (a07-$), 3

a07     db      @grt, 7, @lcma

lomemy  db      7, @eol, 3      ; lomem

dy      db      7
        db      $C0 + (a08-$), @eol, 3

a08     db      @lcma, 7, 4

cy      db      7, @grt
        db      $C0 + (a09-$), @eol, 3

a09     db      @lcma
        db      $C0 + (a09-$), 2, 7
        db      $C0 + (a10-$), 3

a10     db      @lcma
        db      $C0 + (a09-$), 4

qy      db      @doll, 7, @eol, 2, 8, @eol, 3           ; ?
includy db      @pnd
        db      1, high (fspc-1), low (fspc-1)
        db      @eol, 3


;       syntax expression
;       -----------------
bsx6    db      @def, ch, @defl, 2
        db      @grt, ch, @low, 2
        db      @less, ch, @high, 3

bsx3    db      @minus, ch, @umin, 2
        db      @not, ch, @notl, 2
        db      1, high (only1-1), low (only1-1)        ; single unarys
        db      $C0 + (bsx6-$), 3

;  *** entry point

bsx     db      @olp                            ;[ exp ]
        db      $C0 + (bsx-$), @orp

        db      $C0 + (bsx1-$), 2               ; op exp
        db      $C0 + (bsx3-$)                  ; unary exp
        db      $C0 + (bsx-$), 2
        db      $C0 + (bsx4-$)                  ; const op exp
        db      $C0 + (bsx1-$), 3

bsx1    db      $C0 + (bsx2-$)
        db      $C0 + (bsx-$), 4

bsx2    db      @minus, 2, @plus, 2
        db      @cpc, ch, @times, 2
        db      @mod, 2
        db      @div, 2, @and, 2, @or, 2
        db      @xor, 2, @cequ, 2, @lequ, 2
        db      @gequ, 2, @nequ, 2
        db      @grt, 2, @less, 2
        db      @bor, ch, @borl, 2
        db      @band,ch, @bandl, 3

bsx4    db      @cma,2
        db      1, high (tstlab-1), low (tstlab-1)      ; label
        db      2
        db      1, high (hxdc-1), low (hxdc-1)          ; hex/decimal #
        db      2, @cpc, 2, @apost
        db      1, high (tstasc-1), low (tstasc-1)
        db      2, @percent
        db      $C0 + (bsx5-$), 3

bsx5    db      8, 2, @lp
        db      0, low (bsx-1), high (bsx-1)            ; param # exp
        db      @rp, 3

;       master syntax source line
;       -------------------------               ;===== CAUTION ============
label   db      $C0 + (labels-$)                ; NLABEL must follow LABEL...
nlabel  db      1, high (funt-1), low (funt-1)  ;  assumed in assembly!!!
        db      3                               ; LABEL must not be on page
                                                ;  boundary!!
labels  db      1, high (tstlabi-1), low (tstlabi-1)

;       inbedded table (save 1 byte)
;       ----------------------------
oss:    db      $04, $46, $04, $10, $42


;======================================
;       Error messages
;       --------------
eres:   x_db    'ERROR - '
errtab  x_db    1,'MEM FULL'
        x_db    2,'BAD DEL'
        x_db    3,'BR RANGE'
        x_db    4,'NOT Z PAGE/IMM'
        x_db    5,'UNDEF'
        x_db    6,"EXP TOO C'PLX"
        x_db    7,'DUP'
        x_db    8,"O'FLOW"
        x_db    9,'IF/ENDIF'
        x_db    10,'# >255'
        x_db    12,'MACRO ERR'
        x_db    13,'PHASE'
        x_db    14,'*= UNDEF'
        x_db    17,'# >65535'
        x_db    18,'NO ENDM'
        x_db    21,'NESTED INCLUDE'
        x_db    23,'FILE TYPE'
        x_db    24,'LOAD TOO BIG'
        x_db    27,'BAD SET'
        x_db    xrerr,'TOO MANY X-REFS'
        x_db    hlocer,'TOO MANY LOCAL FIELDS'
        x_db    32,'BAD PARAMETER'
        x_db    nfpr,'NOT FP #'

        x_db    130,'NO DEV HANDLER'
        x_db    137,'TRUNCATED REC'
        x_db    138,'TIMEOUT'
        x_db    139,'NAK'
        x_db    144,'WRITE PROTECTED'
        x_db    160,'DRIVE #'
        x_db    161,'TOO MANY FILES'
        x_db    162,'DISK FULL'
        x_db    165,'FILE NAME'
        x_db    167,'LOCKED'
        x_db    170,'FILE NOT FND',0
errf    x_db    '(SEE MANUAL)'


;       print error message
;       -------------------
li1     rts
_listerr:
        lda     eix             ; any errors?
        beq     li1             ; nope, return

        inc     errors          ; bump error count
        pha                     ; save count
        lda     lstdv           ; to device?
        beq     li0             ; nope, just screen
        pha                     ; save device
        lda     #0              ; list to e:
        sta     lstdv           ; x
        jsr     le1             ; do it
        pla                     ; pop list device
        sta     lstdv
li0     pla                     ; pop error count
        sta     eix

        lda     eaf             ; ck mode
        beq     le1             ; br if edit
        jsr     tstej           ; ck eject
le1     lda     #bell           ; ring bell
        jsr     charo
        ldx     #low eres       ; print error
        ldy     #high eres      ; message string
        jsr     ptst

le2     dec     eix             ; pop index
        bmi     le0             ; if done
        ldx     eix
        lda     errstack,x      ; get er code
        beq     usrer           ; user error
        cmp     #200            ; pseudo cio err?
        bcc     le3             ; no
        sbc     #200            ; adjust code

le3     pha                     ; print err code #
        tax
        lda     #0
        jsr     pdnum
        lda     #','            ; print comma
        jsr     charo

        pla                     ; search message table
        jsr     same            ; same messages?

        sta     fr1+2
        lda     #high errtab
        sta     fr1+1
        lda     #low errtab
le4     sta     fr1
        ldy     #0
        lda     (fr1),y
        beq     le7             ; use default mess
        cmp     fr1+2
        beq     le8             ; fnd mess, go print

le5     iny                     ; find start of
        lda     (fr1),y         ; next message
        bpl     le5
        sec
        tya
        adc     fr1             ; add y+1 offset
        bcc     le4             ; to ptr
        inc     fr1+1
        bne     le4             ; unc

usrer   clc
        lda     strings
        adc     #strmax*2
        tax
        ldy     strings+1
        bcc     le9             ; error message
        iny
        bcs     le9             ; unc

le0     ldx     #0
        stx     eix             ; clear err #
        jmp     docr            ;***  done

le8     ldx     fr1             ; bump to start
        ldy     fr1+1           ; of mess
        inx
        bne     le9
        iny
        bne     le9             ; unc

le7     ldx     #low errf       ; default
        ldy     #high errf      ; message
le9     jsr     ptst            ; print string
        jsr     dosp            ; space
        jmp     le2             ; get next

same    proc
        ldy     #8
..sa1   cmp     smg,y
        beq     ..sa2
        dey
        bpl     ..sa1
        rts
..sa2   lda     sng,y
        rts
        endproc


smg     db 11,31,30,16
        db 25,15,20,22
        db 19

sng     db 9,12,12,12
        db 23,8,8,8
        db 14


;       text line # test
;       ----------------
lnum?:  proc
        jsr     afp             ; dec #?
        bcc     glnm            ; yup
        lda     #0              ; nope, reset
        sta     cix             ; text pointer & ret
        rts

glnm    lda     nmfg            ; in auto-num?
        beq     glnm1           ; no
        bit     entdv           ; entering auto?
        bmi     glnm1           ; yes

        lda     cix             ; handle atari e: quirk
        pha
        jsr     afp             ; second line #?
        pla
        bcc     glnm2           ; yes, forget 1st line#
        sta     cix             ; no, restore input index
        bcs     glnm1           ; unc

glnm2   lda     #$c0            ; bit 6 => no auto incr
        sta     nmfg
glnm1   jsr     fpi             ; f.p. to interger

        bcs     lerr            ; if not interger
        ldy     fr0+1           ; set search line #
        lda     fr0
        sty     srlnum+1
        sta     srlnum
        jmp     ok9             ; go store num
lerr    lda     #17
        jsr     puterr
        endproc


;       editor auto number
;       ------------------
numx:   proc
        jsr     anump           ; set user para's
        bcc     nm3             ; go if have slno
        jsr     fndli0

nm1     jsr     endl?           ; statements end?
        bcs     nm2             ; yup

        ldy     #0              ; save this lineo
        lda     (stmcur),y
        sta     srlnum
        iny
        lda     (stmcur),y
        sta     srlnum+1
        jsr     gtadoff         ; point to next
        bcc     nm1             ; unc, try again

nm2     clc                     ; add incr
        lda     srlnum          ; to last lineo
        adc     numi            ; and detect
        sta     lelnm           ; overflow
        lda     srlnum+1
        adc     numi+1
        sta     lelnm+1         ; x
        bcs     numerr          ; oops!

nm3     lda     #$80            ; sucess, set
        sta     nmfg            ; auto flag here
        rts

numerr  lda     #20
        jmp     puterr
        endproc


;       update cur line # for auto/ren
;       ------------------------------
upcur:  proc
        clc                     ; add numi
        lda     lelnm           ; to lelnm
        adc     numi            ; and detect
        sta     lelnm           ; overflow
         lda    lelnm+1         ; x
         adc    numi+1          ; x
         sta    lelnm+1
        bcs     numerr          ; no good!
        rts
        endproc


;       number set-up for rep & num
;       ---------------------------
; out:
;       cc      = starting lineo

anump   proc
        lda     #10             ; set default increment
        sta     lelnm           ; to 10
        sta     numi
        lda     #0              ; high=0
        sta     lelnm+1
        sta     numi+1

        jsr     _getn           ; number or carry
        bcs     anumset         ; use defaults
        sta     numi+1          ; use user
        stx     numi            ; increments

        jsr     gtokel          ; another number?
        beq     anumset         ; no

        lda     numi            ; user inc is now
        sta     lelnm           ; starting lineo
        lda     numi+1
        sta     lelnm+1         ; x

        jsr     _getn           ; get user num inc
        sta     numi+1
        stx     numi

        clc                     ; set starting
        rts                     ; lineo
anumset sec
rj2     rts
        endproc



;       editor renumber statement lines
;       -------------------------------
renx    proc
        jsr     anump           ; set num paras
        jsr     fndli0          ; get first line

rj1     jsr     endl?           ; statements end?
        bcs     rj2             ; yup

        ldy     #0              ; set this lineo
        lda     lelnm           ; equal to
        sta     (stmcur),y      ; current lineo
        lda     lelnm+1
        iny
        sta     (stmcur),y      ; high byte
        jsr     upcur           ; update current
        jsr     gtadoff         ; point to nxt
        bcc     rj1             ; unc, again
        endproc



;       editor delete
;       -------------
delx    proc
        jsr     lrng1           ; set s&e lines
        bcc     dl1             ; first l. fnd
deler   lda     #2              ; invalid del
        jmp     puterr          ; x

dl1     ldy     stmcur          ; save stmcur
        ldx     stmcur+1        ; x
        lda     lelnm           ; move end line to
        sta     srlnum          ; search line ptr
        lda     lelnm+1         ; to calculate move size
        sty     lelnm           ; put start adress
        stx     lelnm+1         ; here for now

        jsr     fndli9          ; search end line #
        bcs     nobump          ; don't inc stmcur
        jsr     gtadoff         ; contract at next line

nobump  sec                     ; get move size
        lda     stmcur
        sbc     lelnm           ; x
        tay                     ; low byte in y
        lda     stmcur+1        ; x
        sbc     lelnm+1         ; high byte in a
        bmi     deler           ; fatal!!!
        ldx     #stmcur         ; table index
        jmp     contract        ; contract & ret
        endproc


;       editor replace
;       --------------
repx    proc
        jsr     getslen         ; get 1st string length
        sta     rps1            ; save
        jsr     getslen         ; get 2nd string length
        sta     rps2            ; save
        lda     #$80            ; set rep mode
        bne     fx1
        endproc


;       editor find
;       -----------
findx   proc
        jsr     gtoke
        cmp     #@pnd           ; list to device?
        bne     fndx1           ; no
        jsr     opnout          ; handle filespec
        inc     cox             ; skip comma

                        ; special procedure for filespec
        sec
        lda     cmdlg           ; adjust exbuf
        sbc     cox
        tax                     ; move size
        sec
        lda     exbuf           ; form move to
        sbc     cox             ; address
        sta     inbuff          ; temp pointer
        lda     exbuf+1
        sbc     #0
        sta     inbuff+1
        ldy     cox             ; now, move 'em
..efr   lda     (exbuf),y
        sta     (inbuff),y
        iny
         dex
         bpl    ..efr
        inx                     ; reset cox  x=0
        stx     cox

        jsr     fndx2           ; execute find
        jmp     closel          ; finish & return
        endproc


;       go find
;       -------
fndx1   proc
        dec     cox
fndx2   jsr     getslen         ; get string length
        sta     rps1            ; save
        lda     #0              ; set find mode

fx1     sta     rpfd
        inc     dspflg          ; for control chars
        jsr     getpa           ; set s&e lineos

        ldy     cmdlg
fx2     lda     (exbuf),y       ; move command line
        sta     (repbuf),y      ; to safe place
        dey
        bpl     fx2

fx3     lda     #$80            ; clear inbuff flag
        sta     rpif
        jsr     lline           ; past last line?
        bcs     fx0             ; yup
fx4     jsr     testbk          ; break?
        bne     fx0             ; yup

        bit     rpif            ; did rep cause delete
        bvs     fx3             ; yup

fx5     jsr     isrch           ; go match string
        bcs     fx6             ; no match!
        lda     rpfd            ; check mode
        bne     rx1             ; br if rep
        jsr     inbout          ; display line
        lda     rpoc            ; first occurance?
        beq     fx0             ; yup

fx6     jsr     gtadoff         ; point to next
        jmp     fx3
fx0     rts
        endproc


                        ; replace module
rx1     lda     rpoc            ; in querry?
        bpl     rx4             ; no
        jsr     inbout          ; list line

rx2     lda     #'?'            ; prompt
        jsr     charo
        jsr     intlbf          ; fail safe!
        lda     #0              ; force $500 buf
        sta     inbuff
        jsr     egli1           ; get line
        jsr     intlbf          ; reset buff
        jsr     testbk          ; abort?
        beq     rx3             ; no
..rex   jmp     prmpt           ; ret to edit

rx3     lda     lbuff-$80       ; get responce
        cmp     #'Y'            ; yes?
        beq     rx4             ; exchange!
        cmp     hcntrlc         ; aborting?
        beq     ..rex
        cmp     #eol            ; no exchange?
        bne     rx2             ; no, go reprompt

        inc     rpsp            ; point to nxt,
        jmp     fx5             ; continue

rx4     lda     rps1            ; form string diff
        sec
        sbc     rps2
        sta     temp            ; save
        lda     rpll            ; form new
        sec                     ; inbuff end
        sbc     temp
        pha                     ; save
        cmp     rpll            ; check against old
        beq     equal           ; br if =
        bcs     movup           ; br if >

        lda     rpop            ; form move from
        adc     rps2            ; inbuf index
        tax
        ldy     rpsp            ; mvta index
mvdw    inx                     ; bump mvfa
        iny                     ; bump mvta
        lda     lbuff,y         ; move until eol
        sta     lbuff,x
        cmp     #eol
        bne     mvdw            ; another byte
        beq     equal           ; unc

movup   tax                     ; mvta inbuf index
        ldy     rpll            ; mvfa inbuf index

mup     lda     lbuff,y         ; move byte
        sta     lbuff,x
        dex                     ; pop indices
        dey
        cpy     rpsp            ; at cur ptr?
        bne     mup             ; no

equal   pla                     ; new inbuff
        sta     rpll            ; string length
        lda     rpsp            ; form new current
        sbc     temp            ; inbuff ptr (cs)
        sta     rpsp
        inc     rpsp            ; and bump it

        lda     rps2            ; string 2 length
        beq     rx6             ; br if zero
        sta     temp            ; save
        ldy     rpse            ; start 2nd string -1
        ldx     rpop            ; old inbuf ptr

rx5     iny                     ; bump
        inx
        lda     (repbuf),y      ; move new char
        sta     lbuff,x         ; to inbuf
        dec     temp            ; pop cnt
        bne     rx5             ; br if not done

rx6     lda     #rmode          ; syntax this
        sta     epf             ; line
        lda     #0              ; clear syntax pointers
        sta     cix             ; current input
        sta     saxtx           ; syntax valid
        sta     cox             ; current output
        jsr     skpspace
        jsr     konvert         ; handle lower case
        jmp     very            ; hidden jsr

veryr:  lda     #0              ; ret here unless
        sta     epf             ; syntax err
        bit     system          ; display the line?
        bvs     ..nd
        jsr     inbout          ; yes, display line

..nd    lda     rpoc            ; first occur?
        beq     ss0             ; yup, ret
        jmp     fx4             ; no, do nxt
        endproc



;       inbuff string search
;       --------------------
; out:
;       cc = found; cs = not fnd

isrch   proc
        lda     rpif            ; inbuf have line?
        beq     ss1             ; yup
        jsr     setztab         ; reset zp tabs
        jsr     movin           ; move to inbuff
        ldy     cix             ; inbuf length +1
        dey                     ; pop
        sty     rpll            ; save
        ldy     #0
        sty     rpif            ; clear inbuf flg

ss1     dec     rpsp            ; init inbuf index
ss2     ldy     #0              ; clear pointer to rep buff
        sty     rpcl
        lda     rpsp            ; save current pointer
        sta     rpop

ss3     inc     rpcl            ; pt to nxt char
        ldy     rpcl            ; check ptr
        cpy     rpse
        bcs     ss5             ; at end, fnd

        lda     (repbuf),y      ; get char
        inc     rpsp            ; inc inbuf ptr
        ldy     rpsp
        cpy     rpll            ; check ptr
        bcs     ss0             ; at end, not fnd

        cmp     lbuff,y         ; they match?
        beq     ss3             ; yup, get nxt
        ldy     rpop            ; old inbuf index
        iny                     ; bump
        sty     rpsp            ; as new srch index
        bne     ss2             ; unc
ss5     clc
ss0     rts
        endproc




;       init for find & replace
;       -----------------------
getpa   proc
        jsr     lrng1           ; set s&e lineo
        jsr     gtokel          ; if eol, then set
        beq     gpa2            ; first occurance

        jsr     gtoke
        ldx     #$ff            ; set for querry
        cmp     #@a             ; is it "a"?
        bne     gpa1            ; no
        ldx     #1              ; set all occurance
         db $2C

gpa2    ldx     #0              ; first occurance
gpa1    stx     rpoc            ; set q, a, or 1st

        ldx     rps1            ; get string length
        inx                     ; bump and save as
        stx     rpse            ; end for string test
        lda     #0              ; list line #
        sta     supr
        rts
        endproc



;       auto-number
;       -----------
an0     proc
        sta     nmfg            ; reset bit 6
        jsr     upcur           ; do the auto update
                        ;  fall for re-try
autonum:
        lda     lelnm           ; search for
        sta     srlnum          ; cur line
        sta     fr0
        lda     lelnm+1
        sta     srlnum+1
        sta     fr0+1
        bit     entdv           ; enter with auto?
        bmi     an1             ; yup, skip!
        jsr     fndli
                        ;   if not in memory,
        bcs     an1             ; do this num

        lda     #$80            ; maybe special case?
        bit     nmfg            ; if bit 6: special
        bvs     an0
        asl     a               ; a=0, clear
        sta     nmfg            ; auto-num & ret
        jmp     prmpt

an1     jsr     prdn            ; print lineo
        lda     #$A0            ; and space
        jsr     enters
        ldy     svn2            ; clean up
        lda     (inbuff),y      ; inbuf lineo
        and     #$7f            ; x
        sta     (inbuff),y
        iny
        lda     #$20            ; lineo + space
        sta     (inbuff),y
        iny
        sty     svn2
        clc
        tya
        adc     inbuff
        sta     inbuff
        lda     #127            ; form new buff
        sec                     ; start adr
        sbc     svn2
        tay
        jsr     stext           ; set cio text
        jsr     io3             ; do cio
        jsr     getstat         ; ck stat
        jmp     intlbf          ; ret
        endproc



;       set-up line number range
;       for list, print, find, & rep
;       ----------------------------
lrng1   proc
        jsr     gtoke
lrnge:  ldy     #0              ; default to
        sty     srlnum          ; line 0 through
        sty     srlnum+1        ; line 65535
        dey
        sty     lelnm
        sty     lelnm+1

        jsr     _getne          ; get num or carry
        bcs     ..lr2           ; no number
        sta     srlnum+1        ; set start &
        sta     lelnm+1         ; end line
        stx     srlnum          ; numbers equal
        stx     lelnm

        jsr     gtokel          ; done?
        beq     ..lr2           ; yup, all set

        cmp     #@minus         ; for "list exp,"
        bne     ..lr8
        lda     #-1             ; set default lelnm
        sta     lelnm
        sta     lelnm+1
        bcs     ..lr7           ; unc

..lr8   jsr     _getn           ; number or carry
        bcs     ..lr1
        sta     lelnm+1         ; save end
        stx     lelnm           ; line number
        bcc     ..lr7           ; unc, find start l.

..lr1   dec     cox             ; pop to spec
..lr2   dec     cox             ; pop to eol/spec
..lr7   jmp     fndli           ; fall thru to "fndli"
        endproc



;       editor size
;       -----------
sizex   proc
        lda     lomem           ; print lomem
        ldy     lomem+1
        jsr     hexw
        jsr     space3          ; three spaces

        lda     txtend          ; print top of text
        ldy     txtend+1
        jsr     hexw
        jsr     space3

        lda     mtp             ; print high mem
        ldy     mtp+1

si      jsr     hexw            ; print hex word
        jmp     docr            ; do eol & ret
        endproc



;       editor hex/dec-dec/hex convert
;       ------------------------------
qx      proc
        lda     #'='            ; print leading =
        jsr     charo
        jsr     gtoke
        cmp     #@doll          ; dollar sign?
        bne     tohex           ; nope

        jsr     _getn
        jsr     pdnum           ; go print it
        jmp     docr            ; do eol & ret

tohex   pha                     ; dec to hex convert
        lda     #'$'            ; leading $ sign
        jsr     charo
        pla
        jsr     _getne          ; get number
        tay                     ; swap for
        txa                     ; print routine
        jmp     si              ; go print it
        endproc



;       editor lomem
;       ------------
lomemx  proc
        jsr     _getn           ; get new lomem
        stx     svlomem         ; adress
        sta     svlomem+1
        jmp     newx            ; clear all pointers
        endproc
$title(MAC65/XL Cartridge  (c) 1988 ICD Inc)
$subtitle(Cartridge bank B for MAC65)

        seg     bankb

;=================================
;       operator routines
;       -----------------
; note:
; special note:  all ops' addresses
; are all on the same page.
;
        db      -1      ; so that high (symbol-1) are all on same page!!
@xtimes jmp     xtimes  ;  @xtimes is assumed to be first (for $if later!)
@xdiv   jmp     xdiv
@xmod   jmp     xmod


;       Addition
;       --------
@xplus  proc                    ; addition
        jsr     opop
        tay
        txa
        clc
        adc     fr0
        sta     fr0
         tya
         adc    fr0+1
         sta    fr0+1
        rts
        endproc


;       Minus and subtract
;       ------------------
@xum    proc
        lda     #0              ; unary minus
        tay                     ; a & y =0
        beq     ..jsb           ; join subtract

@xminus jsr     opop            ; subtraction
        tay
        txa

..jsb   sec
        sbc     fr0
        sta     fr0
         tya
         sbc    fr0+1
         sta    fr0+1
        rts
        endproc


;       Logical and
;       -----------
@xand   proc
        jsr     opop            ; logical and
        and     fr0+1
        sta     fr0+1
         txa
         and    fr0
         sta    fr0
        rts
        endproc


;       Equal and logical not
;       ---------------------
@xceq   proc
        jsr     @xminus         ; comp =
@xnot   lda     fr0+1           ; not
        bne     falser
        lda     fr0
        bne     falser

truer   ldx     #1              ; push true
sf1     lda     #0
sf0     sta     fr0+1
        stx     fr0
        rts

falser  ldx     #0              ; push false
        beq     sf1
        endproc


;       Not equal
;       ---------
@xneq   proc
        jsr     @xminus         ; comp <>
        lda     fr0+1
        bne     truer
        lda     fr0
        bne     truer
        beq     falser
        endproc


;       test label def
;       -------------
@xdef   proc
        ldy     #3              ; tst label def
        lda     (srcadr),y
        and     #1
        eor     #1
        tax
        lda     #0
        sta     res             ; reset resolved
        beq     sf0
        endproc


@xorp   proc
        ldy     aptr            ; right paren ]
        pla
        pla
        ldx     argops,y
        jmp     xs7
        endproc


;       Left paran
;       ----------
@xelp   proc
        pla                     ; left "(" special
        pla                     ; pop return addr
        lda     fr0+1           ; this is param #
        ldx     fr0
        jsr     valuep          ; get actual value
        sta     fr0+1           ; return to fr0
        stx     fr0
        jmp     xs9             ; and continue
        endproc


;       Logical or
;       ----------
@xorb   proc
        jsr     opop            ; logical or
        ora     fr0+1
        sta     fr0+1
         txa
         ora    fr0
         sta    fr0
@xolp   rts                     ; left paren [
        endproc


;       Logical XOR
;       -----------
@xxor   proc
        jsr     opop            ; logical eor
        eor     fr0+1
        sta     fr0+1
        txa
        eor     fr0
        sta     fr0
        rts
        endproc


;       Compare
;       -------
cmp     proc
        jsr     opop            ; compare
        cmp     fr0+1
        bcc     cm0
        bne     cm0
        cpx     fr0
cm0     rts
        endproc


;       comparison operators
;       --------------------
@xgrt   proc
        jsr     cmp             ; greater than
        beq     @fl
        bcc     @fl
        bcs     @tr

@xls    jsr     cmp             ; less than
        bcc     @tr
        bcs     @fl

@xleq   jsr     cmp             ; less or equ
        beq     @tr
        bcc     @tr
        bcs     @fl
        endproc

;       BASIC and/or
;       ------------
@xband  proc
        jsr     opop            ; basic and
        stx     fr1             ; low
        ora     fr1             ; low or high (in a)
        beq     @fl
        bne     ..xbj           ; check high

@xbor   jsr     opop            ; basic or
        stx     fr1
        ora     fr1
        bne     @tr
..xbj   lda     fr0
        ora     fr0+1
        bne     @tr
        beq     @fl
        endproc


;       Greater or equal
;       ----------------
@xgeq   proc
        jsr     cmp             ; greater or equ
        bcc     @fl

@tr     jmp     truer
@fl     jmp     falser
        endproc


;       High/low operator
;       -----------------
@xhigh  proc
        lda     fr0+1           ; do high byte
        sta     fr0

@xlow   lda     #0              ; do low byte
        sta     fr0+1
        rts
        endproc


;       Reference
;       ---------
@xref   proc
        ldy     #3              ; label reference
        ldx     #0
        bit     eaf             ; get pass #
        bvs     @2ref           ; go if pass 2
        lda     mpf             ; check mini 2
        bne     @2ref           ; yup
        bit     res             ; test referenced
        stx     res             ; inhibit undefined
        bmi     @2ref           ; not referenced
        jsr     s3by            ; save 3 bytes
        bne     @tr             ; unc

@2ref   lda     (srcadr),y      ; see if ref
        stx     res             ; inhibit undefined
        and     #$10            ; in pass 1
        bne     @tr             ; yup
        beq     @fl
        endproc


        $if (@xref-@xtimes) > 254
         $error "operators not on same page"
        $endif

semd    proc
        lda     #0              ; set up for
        sta     fr0+2           ; mult & div
        sta     fr0+3
        ldy     #16
        rts
        endproc


;       Multiply
;       --------
xtimes  proc
        jsr     opop            ; multiply
        sta     fr1+1
        stx     fr1
        jsr     semd
..m1    lsr     fr0+3
        ror     fr0+2
        ror     fr0+1
        ror     fr0
        dey
        bmi     ..dv9           ; all done
        bcc     ..m1

        clc
        ldx     #-2
..m2    lda     fr1+2,x
        adc     fr0+4,x
        sta     fr0+4,x
        inx
        bne     ..m2
        beq     ..m1

xmod    jsr     xdiv            ; modulo
        lda     fr0+2           ; remainder is result
        sta     fr0
        lda     fr0+3
        sta     fr0+1
        rts

xdiv    lda     fr0             ; division
        sta     fr1
        lda     fr0+1
        sta     fr1+1
        jsr     opop
        sta     fr0+1
        stx     fr0
        jsr     semd
..dv1   asl     fr0
        rol     fr0+1
        rol     fr0+2
        rol     fr0+3
        sec
        lda     fr0+2
        sbc     fr1
        tax
        lda     fr0+3
        sbc     fr1+1
        bcc     ..dv2
        sta     fr0+3
        stx     fr0+2
        inc     fr0
..dv2   dey
        bne     ..dv1
..dv9   rts
        endproc



;       going on stack precedance
;       -------------------------
onprec  db $09          ; +
        db $09          ; -
        db $0a          ; *
        db $0a          ; /
        db $08          ; &
        db $0f          ; operator "("
        db $07          ; =
        db $07          ; <=
        db $07          ; >=
        db $07          ; <>
        db $07          ; >
        db $07          ; <
        db $0f          ; - unary
        db $0f          ; [
        db $04          ; ]
        db $00,$00,$00
        db $08          ; !
        db $08          ; ^
        db $00
        db $0b          ; \

;addressing mode type by terimator
; the next 8 bytes are an inbedded
; table, only .ref is used but not
; needed

endm    db 3            ; ,x)      1 free
        db 2            ; ),y        |
        db 6            ; ,y         |
        db 5            ; ,x         |
        db 4            ; )          |
        db 7            ; @comment   |
        db 7            ; @eol     7 free
        db $fb          ; @cma      .ref

        db $0f          ; .def
        db $0f          ; .not
        db $06          ; .and
        db $05          ; .or
        db $0f          ; < low byte
        db $0f          ; > high byte
        db 0,0,0,0,0,0,0,0,0


;       comming off precedance
;       ----------------------
ofprec  db $09          ; +
        db $09          ; -
        db $0a          ; *
        db $0a          ; /
        db $08          ; &
        db $01          ; operator "("
        db $07          ; =
        db $07          ; <=
        db $07          ; >=
        db $07          ; <>
        db $07          ; >
        db $07          ; <
        db $0d          ; - unary
        db $02          ; [
        db $0e          ; ]

; string displ table (3 bytes)
; ----------------------------
mstb    db $00,strmax,strmax*2
        db $08          ; !
        db $08          ; ^
        db $00
        db $0b          ; \

;option mask table
; the next 8 bytes are an inbedded
; table, only .ref is used but not
; needed

opttab  db $40          ; obj      1 free
        db $20          ; err        |
        db $01          ; eject      |
        db $80          ; list       |
        db $02          ; xref       |
        db $04          ; mlist      |
        db $08          ; clist    7 free
        db $10          ; num       .ref

        db $02          ; .def
        db $0d          ; .not
        db $06          ; .and
        db $05          ; .or
        db $02          ; < low byte
        db $02          ; > high byte   CART version 1.01 this way
;       db $04          ; > high byte   Steve's version like this
        db 0,0,0,0,0,0,0,0,0



;       operator tab low bytes
;       ----------------------
opertrl db      low @xplus-1    ; +
        db      low @xminus-1   ; -
        db      low @xtimes-1   ; *
        db      low @xdiv-1     ; /
        db      low @xand-1     ; &  binary
        db      low @xelp-1     ; operator "("
        db      low @xceq-1     ; =  compare
        db      low @xleq-1     ; <=
        db      low @xgeq-1     ; >=
        db      low @xneq-1     ; <>
        db      low @xgrt-1     ; >
        db      low @xls-1      ; <
        db      low @xum-1      ; -  unary
        db      low @xolp-1     ; [
        db      low @xorp-1     ; ]
        db      0,0,0           ; null entries
        db      low @xorb-1     ; !  binary
        db      low @xxor-1     ; ^ eor
        db      0               ; ref null
        db      low @xmod-1     ; \

;inbedded routine for .ref
; save 3 bytes (7 bytes long)

s3by    lda     #$10            ; show referenced
        ora     (srcadr),y
        sta     (srcadr),y
        rts

        db      low @xref-1     ; .ref
        db      low @xdef-1     ; .def
        db      low @xnot-1     ; .not
        db      low @xband-1    ; .and
        db      low @xbor-1     ; .or
        db      low @xlow-1     ; < low byte
        db      low @xhigh-1    ; > high byte


;       execute directive tab
;       ---------------------

xdirtb  dw      @@if-1
        dw      @@else-1
        dw      @@endif-1
        dw      @@macro-1
        dw      @@endm-1
        dw      @@title-1
        dw      @@macrox-1
        dw      @@page-1
        dw      @@word-1
        dw      @@error-1
        dw      @@byte-1
        dw      @@sbyte-1
        dw      @@dbyte-1
        dw      @@end-1
        dw      @@opt-1
        dw      @@tab-1
        dw      @@include-1
        dw      @@ds-1
        dw      @@aster-1
        dw      @@equal-1

        dw      @@bra-1
        dw      @@trb-1, @@trb-1

        dw      @@float-1
        dw      @@cbyte-1
        dw      @@semi-1
        dw      @@local-1
        dw      @@set-1
        dw      @@aster-1
        dw      @@equal-1, @@.equal-1
        dw      @@jsr-1, @@jmp-1


;       adr field tab
;       -------------
fldtab  db      0               ; null
        db      $08             ; immediate
        db      $10             ; ind,y
        db      $00             ; ind,x
        db      $10             ; zp ind
        db      $14             ; zp,x
        db      $14             ; zp,y
        db      $04             ; zero page
        db      $1c             ; abs,x
        db      $18             ; abs,y
        db      $0c             ; absolute

        db      $08             ; accumulator
        db      $00             ; immediate
        db      0,0,0           ; null  (3)
        db      $14             ; zp,x
                        ; next 2 bytes are
trb0                    ;  for trb & tsb
        db      $14             ; zp,y
        db      $04             ; zero page
        db      $1c             ; abs,x
        db      $1c             ; abs,y
        db      $0c             ; absolute


;       op tab
;       ------
                ; instr & type
@@cd    db $c6          ; dec 2
        db $e6          ; inc 2
        db $a6          ; ldx 2
        db $a4          ; ldy 2
        db @@sx         ; stx 1
        db @@sy         ; sty 1
        db $e4          ; cpx 2
        db $c4          ; cpy 2
        db $20          ; bit 1

                ; implied
        db $00          ; brk
        db $18          ; clc
        db $d8          ; cld
        db $58          ; cli
        db $b8          ; clv
        db $ca          ; dex
        db $88          ; dey
        db $e8          ; inx
        db $c8          ; iny
        db $ea          ; nop
        db $48          ; pha
        db $08          ; php
        db $68          ; pla
        db $28          ; plp
        db $40          ; rti
        db $60          ; rts
        db $38          ; sec
        db $f8          ; sed
        db $78          ; sei
        db $aa          ; tax
        db $a8          ; tay
        db $ba          ; tsx
        db $8a          ; txa
        db $9a          ; txs
        db $98          ; tya

                ; branch
        db $90          ; bcc
        db $b0          ; bcs
        db $f0          ; beq
        db $30          ; bmi
        db $d0          ; bne
        db $10          ; bpl
        db $50          ; bvc
        db $70          ; bvs

                ; multi
        db $01          ; ora 1
        db $21          ; and 1
        db $41          ; eor 1
        db $61          ; adc 1
        db $81          ; sta 1
        db $a1          ; lda 1
        db $c1          ; cmp 1
        db $e1          ; sbc 1
        db $06          ; asl 2
        db $26          ; rol 2
        db $46          ; lsr 2
        db $66          ; ror 2

        db -1           ; r65c02 instrs
        db $80          ; bra (stz slot)
        db $3a          ; dea
        db $1a          ; ina
        db $da          ; phx
        db $5a          ; phy
        db $fa          ; plx
        db $7a          ; ply


;       instruction length table
;       ------------------------
lentab  db 1,2,2,2
        db 2,2,2,2
        db 3,3,3


;       assembler portion
;       -----------------
_asmx:  proc
        ldx     #tosa-1         ; set asm stack
        txs
        jsr     asminit         ; initialize
        lda     #$80            ; set 1st pass
        sta     real            ; master pass cntrl
        sta     eaf             ; internal pass #

        jsr     gtokel          ; test @eol
        beq     ax3             ; yup, all set
        cmp     #@lcma          ; a comma?
        beq     ax1             ; yup, check list device

                        ; *** source file
        jsr     hfi             ; open for read

        ldy     #low asmfm      ; CART version!!
        jsr     hstr

        jsr     validf          ; test check bytes
        jsr     donote          ; 'note' start
        jsr     gtokel          ; test @eol
        beq     ax3             ; yup, go assemble

ax1     jsr     gtoke           ; next token
        cmp     #@lcma          ; a comma?
        beq     ax2             ; yup, check object device

                        ;  *** list file
        cmp     #@minus         ; inhibit list?
        bne     ..opn           ; must be a "#"
        lda     #$80            ; master list control
        sta     klist
        bne     ..il            ; unc
..opn   jsr     opnout          ; use list dev
..il    jsr     gtokel          ; test @eol
        beq     ax3             ; yup, all set

ax2     jsr     gtoke           ; next token
        cmp     #@lcma          ; comma?
        beq     axx             ; do cross reference

                        ;  *** object file
        jsr     fobj            ; use obj dev
                        ;  out binary file
        jsr     outff           ; check bytes
        jsr     gtokel          ; test @eol
        beq     ax3             ; all set

                        ;   *** x-ref file
axx     inc     cox             ; skip # sign
        jsr     mstrlitr        ; filespec in table & lbuff
        jsr     fref            ; now, cio stuff
        lda     asmram+2        ; set initial addr
        ldx     asmram+3
        sta     bufferx
        sta     xrefbufi
        stx     bufferx+1
        stx     xrefbufi+1

        ldx     #asmram+2       ; now, allocate it
        lda     #high bufsize /2
        ldy     #0
        jsr     expand

ax3     lda     #$40+$10        ; no obj, no num
        ldy     objdv
        beq     ..4
        lda     #$10            ; no num
..4     ora     klist           ; and master list
        sta     optflag

                        ; main loop
asml    jsr     asmline         ; assemble a line
        jsr     wrtxref         ; handle x-refs
        lda     eofflag         ; end of pass?
        bpl     asml            ; nope

        lda     sxref           ; pop x-ref?
        beq     ..rx            ; no
         sta    refdv           ; restore x-ref

..rx    asl     real            ; done assembling?
        bne     asme            ; yup
        lda     #$40            ; no, set pass 2
        sta     real
        sta     eaf
        jsr     reasm           ; warm init

                        ;  out title, page...
        jsr     ppass
        beq     ax3             ; unc, go do it!

asme    lda     refdv           ; doing x-ref?
        beq     asmf            ; no
        lda     #rfeof          ; eof marker
        jsr     xrout
        jsr     rfwrite         ; empty buffer
        lda     refdv
        jsr     closei          ; close x-ref dev
        ldy     #3              ; re-open as entdv
        lda     #4
        sty     entdv
        jsr     cio
        clc                     ; form filespec addr
        lda     smbltp
        adc     #6
        tay
        lda     smbltp+1
        adc     #0
        jsr     io4.s           ; do open

asmf    jsr     @@symbol        ; do symbol table
nullx
asmexit:
        jsr     clsall          ; close devices
        jsr     settab0         ; reset list tabs
        jsr     reset           ; reset pointers
        jmp     prmpt           ; return to editor
        endproc



;       initialize assembler
;       --------------------
asminit proc
        jsr     reset1
                        ;  start string space
        lda     asmram+2
        ldx     asmram+3
        sta     strings
        stx     strings+1
                        ;  set io buffer addr
        clc
        adc     #low (strmax*3)
        sta     buffers         ; lows
        sta     bufftab
        sta     bufftab+4
        sta     bufftab+2

        txa
        adc     #high (strmax*3)
        sta     buffers+1
        sta     bufftab+1
        sta     bufftab+5
                        ;   buffersav addr
        clc
        adc     #high bufsize
        sta     bufftab+3
                        ;    now, allocate area
        lda     #high ((bufsize*2) + (strmax*3))
        ldy     #low  ((bufsize*2) + (strmax*3))
        ldx     #asmram+2
        jsr     expand

        lda     #topform        ; init top of form
        sta     topfc
        lda     #80             ; right margin
        sta     rightmar
        lda     #66-8           ; lines/page
                        ;   (physical size-heading)
        sta     formsz
        ldy     #0
        sty     lstdv           ; clr all dev
        sty     entdv
        sty     objdv
        sty     refdv
        sty     leftmar         ; left margin
        sty     klist           ; master list on

        lda     #-1
        ldx     #55             ; clear label pointers
..cst   sta     spointer,x
        dex
        bpl     ..cst

        lda     #$a0
        sta     asmfm           ; main file list
        sta     (strings),y     ; title string (y=0)
                        ;   expand macro stack
        lda     #1              ; by 1 page (y=0)
        ldx     #stack+2
        jsr     expand

reasm   ldy     #1              ; init to zero
        lda     #0              ; parameters on stack
        sta     (stack),y
         dey
         sta    (stack),y

        ldx     #1
..ccl   sta     asmpc,x         ; pc
        sta     offset,x        ; object offset
        sta     xrefsz,x        ; x-ref file size
        sta     pagenum,x       ; set pg # to 0
        sta     bytcnt,x        ; byte count
        sta     lastl,x         ; last line number

        $if xrindx = (sxref+1)
          sty   sxref,x
        $endif
        $if eofflag = (pcdef+1)
          sty   pcdef,x
        $endif
        $if eix = (epf+1)
          sty   epf,x           ; clr err indx
        $endif
        $if localix = (asmlino+2)
          sty   asmlino+1,x     ; auto-num
        $endif

        dex
        bpl     ..ccl
        stx     linecnt         ; lines on page=-1
                                ;   y =0
        $if xrindx <> (sxref+1)
          sty   xrindx          ; x-ref buff index
          sty   sxref           ; no x-ref hold
        $endif
        $if eofflag <> (pcdef+1)
          sty   pcdef           ; pc def flg
          sty   eofflag ; clr pass end
        $endif
        $if eix <> (epf+1)
          sty   eix             ; clr err indx
          sty   epf             ; v1.100 safety
        $endif
        $if localix <> (asmlino+2)
          sty   localix         ; first local field
          sty   asmlino+1       ; asm auto-num, hi
        $endif

        sty     curaug          ; macro params
        sty     mpf             ; mini-pass flag
        sty     inmac           ; in-process macro
        sty     rstk            ; run-time stk ptr
        sty     mcrsk           ; macro stk ptr
        sty     errors          ; assembly errors!

        lda     #$a0
        ldy     #strmax
        sta     (strings),y     ; set pagep

        $if (strmax*2) = $a0
          tay
        $else
          ldy   #strmax*2
        $endif

        sta     (strings),y     ; set usrmess
        sta     asmfi           ; include file list

        lda     #low 100
        sta     asmlino
        lda     #4
        sta     setbyte         ; set byte listing
        jmp     settab0         ; go do tabs
        endproc


;       assemble a line
;       ---------------
asmline proc
        jsr     gscl            ; get a line
        bcs     enpass          ; br end of pass
        lda     #0
        sta     objix           ; obj index
        sta     xrindx          ; x-ref index
        jsr     gtoke
        beq     spp3            ; error l.
        bpl     ..nlab
        jsr     labint          ; prefix label

        ldy     #2              ; ck for label, eol
        lda     cox
        cmp     (exbuf),y
        bcs     spp9            ; yup, go list
        jsr     gtoke
..nlab  cmp     #@scomment      ; comment?
        beq     spp3            ; yup
        jmp     opcode          ; handle instr
spp9    jmp     putinbuff

                        ; set end of source status
enpass  bit     eofflag         ; who issued eof
        bvc     true_           ; not .include
         lda    #0              ; else, clr eof
          db $2c
true_   lda     #$80            ; set eof stat
        sta     eofflag
        jmp     write           ; clean buffer
        endproc



;       list entry for non-code sorce
;       -----------------------------
; out:
;       result in fr0 for spp2
; spp1: label eol/*=/end/macro def
; spp2: .if and =,
; spp3: comment/error/command

spp1    proc
        bit     real            ; really pass 1?
        bmi     spq             ; yes
        jsr     spp0            ; see if list
        bne     spq             ; no
                        ;  fatal err list entry
sppr:   jsr     papc            ; print asm pc +space
        lda     #11             ; spaces
        bne     spcey           ; now do line

spp2    bit     real            ; really pass 1?
        bmi     spq             ; yes
        jsr     spp0
        bne     spq             ; as above
        lda     #4
        jsr     space           ; do spaces
        lda     #'='            ; do =
        jsr     charo
        ldy     fr0+1
        lda     fr0
        jsr     hexw            ; pr result
        lda     #7
        bne     spcey           ; pr 7 spaces

sp3er:  jsr     puterr          ; list w/error

spp3:   bit     real            ; really pass 1?
        bmi     spq             ; yes
        jsr     spp0            ; listing?
        bne     spq
        lda     #16             ; sixteen spaces
spcey   jsr     spcez           ; ck list macro
        jsr     asmlst          ; list source
        jmp     tsterr          ; ret thru here
        endproc



;       if in macro, out an "m" & do left margin
;       ----------------------------------------
spcez   proc
spcem:  bit     inmac           ; in macro?
        bpl     spceu           ; nope

        sec
        sbc     #1              ; pop spaces by 1
        jsr     space           ; and do 'em
        lda     #'M'            ; do 'm
        jmp     charo           ; ret
spceu   jmp     space           ; just pad spaces
        endproc


tstlstm proc
        lda     inmac           ; active macro?
        beq     spq             ; nope
        lda     optflag
        jsr     tstlss          ; ck errs
        beq     spq             ; err line, list
        lda     optflag         ; ck macro list opt
        and     #4
spq     rts
        endproc


;       test for list and eject (hairy!)
;       --------------------------------
spp0    proc
        jsr     tstlst
        bne     spd             ; no listing
        jsr     tstlstm         ; macro list
        bne     spd
        jsr     tstej           ; test eject
        jsr     leftmr          ; check left margin
spd     php
        jsr     tstlstm         ; macro line #
        beq     spb
        jsr     bckup
spb     plp
        rts
        endproc


;       reset asmlino
;       -------------
bckup   proc
        bit     eaf             ; check pass #
        bmi     ..b5            ; still in #1
        ldy     #1              ; if: in macrox &
..b6    lda     (exbuf),y       ; opt = no mlist
        sta     asmlino,y
        dey                     ; reset asmlino
        bpl     ..b6
..b5    rts                     ; ret <>0
        endproc



;       get a line from something
;       -------------------------
        proc
..abrk  jmp     asmexit
gscl    bit     brkkey          ; break key?
        bpl     ..abrk          ; yup

        lda     entdv
        beq     memory
retry   ldy     #0              ; set exbuf index
        sty     cox
        jsr     objx3           ; get line # +offset
        bcs     eoof            ; oops!
        adc     #-3             ; sub 3; a =offset
        jsr     objx
        bcc     puteol          ; sucess
eoof    sec
        rts

memory  lda     inmac           ; inprocess macro?
        bmi     notxt           ; don't test text end
        jsr     endl?
        bcs     eoof
notxt   ldy     #2
        lda     (stmcur),y
        tay
        dey
mvel    lda     (stmcur),y
        sta     (exbuf),y
        dey
        bne     mvel
        lda     (stmcur),y
        sta     (exbuf),y

        jsr     gtadoff         ; update "stmcur"
puteol  ldy     #2
        lda     (exbuf),y
        iny
        sty     cox
        tay                     ; y returns with
        lda     #@eol           ; line len
        sta     (exbuf),y
        bit     real            ; pass 2?
        bmi     ..wk1           ; no

        lda     optflag         ; num option
        and     #$10
        bne     ..wk1           ; no num

        sty     bfld            ; safe temp for y
        ldy     #1
..wk2   lda     asmlino,y       ; do auto num
        sta     (exbuf),y
        dey
        bpl     ..wk2
        inc     asmlino         ; bump line num
        bne     ..wk0
         inc    asmlino+1

..wk0   ldy     bfld            ; restore y
..wk1   clc
        rts
        endproc



;       print assembly end message
;       --------------------------
asmmess: proc
        jsr     tstej           ; test eject
        jsr     tstejcr         ; blank line
        jsr     leftmr          ; handle left margin

        ldy     #high errasm
        ldx     #low errasm
        jsr     ptst
        ldx     errors
        jsr     pdnum0          ; out # of errors
        jsr     space3          ; spaces
        sec                     ; form free bytes
        lda     mtp
        sbc     memtop
        tax
         lda    mtp+1
         sbc    memtop+1
        jsr     pdnum
        ldy     #high bytasm    ; final message
        ldx     #low bytasm
        jmp     ptse            ; ret thru
        endproc


;       branches
;       --------
@@bra   proc
        ldy     #@rockwell      ; branch always
xbra    lda     @@cd-@dec,y     ; get instr
        sta     tpobuf+1        ; in obj buff
        jsr     xscan           ; get result
        txa                     ; some malupation
        sec                     ; cal rel brnch
        sbc     asmpc
        tay
        lda     fr0+1
        sbc     asmpc+1
        tax
        tya
        sec
        sbc     #2
        sta     tpobuf+2        ; operand in buff

        tay                     ; now, test range
        txa
        sbc     #0
        bne     ..xr1
        tya
        bmi     ..xr2
        bpl     ..xr3
..xr1   tax
        inx
        bne     ..xr2
        tya
        bmi     ..xr3

..xr2   lda     #3              ; br range err
        jsr     puterrasm
..xr3   lda     #2
        sta     objix           ; buffer size
        jmp     putinbuf0
        endproc



;       process mnenorics/directives
;       ----------------------------
opcode  proc
        tay
        cmp     #@dec           ; directive?
        bcs     opc1            ; no

        asl     a               ; go execute directive
        tax
        lda     xdirtb-@if*2+1,x
        pha
         lda    xdirtb-@if*2,x
         pha
        rts

opc1    cmp     #@brk           ; check for other types
        bcc     multi
        cmp     #@ora           ; multi type
        bcs     mvlti
        cmp     #@bcc           ; branch?
        bcs     xbra

aimp    lda     @@cd-@dec,y     ; do implied type
        sta     tpobuf+1        ; in obj buff
        lda     #1
        sta     objix           ; buff size
        jmp     putinbuf0

                        ; test/reset bit instrs
@@trb   lda     trb0-@trb,y     ; get base code
        pha                     ; and save
        jsr     xscan           ; get addr
        pla
        jmp     df2

mvlti   cmp     #@rockwell      ; new instr?
        bcc     multi           ; nope
        bne     aimp            ; new implied type
                        ;  else, stz
        jsr     xscan
        lda     stzt-5,y        ; get op
        sty     mode            ; save mode
        jmp     finup           ; unc


multi   lda     @@cd-@dec,y     ; multi
        tax                     ; save inter-code
        and     #$e3            ; get base code
        sta     bcode           ; less adress field
        txa
        and     #$1c            ; which field?
        beq     fld1            ; field 1
         lda    #11
fld1    sta     bfld            ; ofset to field 2
        jsr     gtoke           ; test prefix adress
        cmp     #@a             ; acc?
        bne     notac
        ldy     #0              ; field indx acc
        beq     skk
notac   cmp     #@pnd           ; immd?
        bne     cklp            ; test (
        jsr     xscan           ; scan exp
        ldy     #1
        bne     skk
cklp    cmp     #@lp            ; for indirect
        beq     sk
        dec     cox             ; adj indx
sk      jsr     xscan
skk     sty     mode            ; y=mode
        cpy     #4              ; was it (zp)?
        bne     skj             ; no

        lda     bcode           ; special ***
        and     #$f0            ; save base
        ora     #2              ; add new addr field
        bne     df1             ; unc, go join

skj     clc
        tya
        adc     bfld
        tay

        lda     bcode           ; test special case op-codes
ckstxy  cmp     #@@sx           ; is it stx?
        bne     ckstyy          ; nope
        cpy     #9              ; it ,y?
        bne     df1             ; no
        ldy     #6              ; reset to z-page
        bne     df2             ; unc
ckstyy  cmp     #@@sy           ; it sty?
        bne     ckzpy           ; no
        cpy     #8              ; it ,x?
        bne     df1             ; no
        ldy     #5              ; reset to z-page
        bne     df2             ; unc
ckzpy   cpy     #6              ; it zp,y?
        bne     df1             ; no
        ldy     #9              ; set to abs,y
df2     sty     mode            ; store adressing mode
df1     ora     fldtab,y

finup   sta     tpobuf+1        ; op code
        lda     fr0             ; lo
        sta     tpobuf+2
        lda     fr0+1           ; high
        sta     tpobuf+3
        ldy     mode            ; instr len
        lda     lentab,y
        sta     objix
        cmp     #2              ; test for legal
        bne     nzero
                                ;   a =2
        ldy     fr0+1
        beq     nzero           ; z page ok
        iny                     ; try for $ff
        beq     nzero           ; ok
        asl     a               ; addr err (a =4)
        jsr     puterrasm       ; asm vector
        lda     objix           ; restore count
nzero   jmp     putinbuf0
        endproc



;       stz op codes
;       jmp addr fields
;       ---------------

ijt     db $30,$20,-1   ;jmp: ind,x  ind
;stzt   db $74,-1       ;stz: zp,x  null        ; steve's version!!
stzt    db $74,0        ;stz: zp,x  null        ; CART version!!!
        db $64,$9e      ;stz: zp  abs,x
        db $00,$9c      ;jmp: abs stz: abs


;       directive jsr & jmp
;       -------------------
@@jsr   proc
        lda     #$20            ; base op code
..ii    pha                     ; save instr
        jsr     xscan           ; get result
        sta     tpobuf+3        ; high byte
        pla                     ; get instr
        ora     ijt-3,y         ; for jmp
        sta     tpobuf+1        ; opcode
        stx     tpobuf+2        ; low byte
        lda     #3
        sta     objix           ; size of instr
        jmp     putinbuf0

@@jmp   ldy     cox             ; test for (
        lda     (exbuf),y
        cmp     #@lp
        bne     ..ij
         inc    cox             ; skip the (
..ij    lda     #$4c            ; get base code
        bne     ..ii            ; unc
        endproc



;       directives
;       db, .sbyte, .cbyte, .dbyte, .word
;       ---------------------------------

@@cbyte proc
        lda     #$40            ; character bytes
        bne     setdrf

@@word  lda     #$80            ; set reverse adress
         db $2c
@@dbyte lda     #$81            ; set high/low adress
        ldx     #2              ; list 2 bytes
        bne     setdr

@@sbyte lda     #0              ; screen adjust
         db $2c
@@byte  lda     #1              ; no screen adjust
setdrf  ldx     setbyte         ; list format

setdr   stx     tpobuf          ; list format
        sta     dirfg           ; directive type

        lda     #0              ; init modifier
        sta     modfybyt
        jsr     gtoke           ; modifier?
        cmp     #@plus
        beq     evalm           ; yes
        dec     cox             ; pop exbuf index
        bne     eval            ; unc

                        ; get modifier value
evalm   jsr     xscan           ; get result
        stx     modfybyt        ; and save it!

eval0   inc     cox             ; past comma
eval    jsr     gtoke           ; entry point
        cmp     #@qu            ; string literal?
        beq     str@            ; yup
        cmp     #@mstr          ; macro string
        beq     strm@           ; yup

        dec     cox
        jsr     xscan           ; scan exp
        sty     commode         ; save end status
        ldy     dirfg
        bmi     words           ; two byte
        bne     ..aj0
         txa                    ; adj to screen
         jsr    adjust          ; code
         tax

..aj0   txa
        jsr     object          ; put in buff
        ldy     fr0+1           ; test for byte
        beq     inrnge          ; ok if =0
        iny                     ; other case
        beq     inrnge          ; ok if =-1
        lda     #10             ; val > 255 er
        jsr     puterrasm       ; asm vector

inrnge  ldy     commode         ; at end?
        bmi     eval0           ; y minus if comma

bytend  lda     modfybyt        ; do modifier
        beq     nomod           ; skip if nothing
        ldy     objix
modloop clc
        lda     tpobuf,y
        adc     modfybyt
        sta     tpobuf,y
        dey
        bne     modloop

nomod   jmp     putinbuff

words   cpy     #$81            ; two byte process
        bne     dword           ; it is word

        lda     fr0
        ldx     fr0+1
dword   pha
        txa
        jsr     object
        pla
        jsr     object
        bne     inrnge          ; unc go test nxt
        endproc


;       string bytes, literal
;       ---------------------
str@    proc
        jsr     gtoke           ; str literal
        and     #$7f            ; str len
        tax
jj8     jsr     gtoke
        ldy     dirfg           ; screen code?
        bne     ..aj1
         jsr    adjust
..aj1   jsr     object
        dex
        bne     jj8
        inc     cox             ; past other "

                        ; is directive .cbyte?
strend  bit     dirfg           ; .cbyte?
        bvc     ..noc           ; nope!
                        ;   y has index from "object"
                        ;   a has last byte too
         eor    #$80            ; toggle last string char
         sta    tpobuf+1,y      ; and store

..noc   jsr     endexp?
        tya                     ; y= - if comma
        bpl     bytend
        jmp     eval0           ; another exp
        endproc


;       macro strings %$
;       ----------------
strm@   proc
        dec     cox             ; point back to %$
        jsr     xscan           ; set srcadr adr
        ldx     #0              ; string ofset
        lda     (srcadr,x)
        sta     temp
        ldy     #6
mmstr   lda     (srcadr),y      ; move string
        sta     lbuff,x         ; to lbuff
        inx
        iny
        cpy     temp            ; done?
        bcc     mmstr           ; move another

        cmp     #eol            ; was last eol?
        bne     mvout           ; no!

        dex                     ; strip the eol
mvout   stx     temp            ; save length
        ldx     #0
..mob   lda     lbuff,x         ; get the byte
        ldy     dirfg           ; a screen byte?
        bne     ..aj2           ; no
         jsr    adjust          ; do screen code
..aj2   jsr     object          ; out it
        inx
        cpx     temp            ; done?
        bcc     ..mob           ; no
        bcs     strend          ; yup, unc
        endproc


;       adjust ascii to atari screen
;       ----------------------------
adjust  proc
        tay                     ; set stat
        php                     ; save stat
        and     #$7f
        cmp     #'a'-1
        bcs     ..f01
        cmp     #$20            ; space
        bcc     ..f00
        sbc     #$20
         jmp    ..f01

..f00   adc     #$40
..f01   plp                     ; get stat
        bpl     ..f02
         ora    #$80
..f02   rts
        endproc



;       directive opt
;       -------------
@@opt   proc
        jsr     gtoke
        cmp     #@no            ; negate option?
        beq     off@
        cmp     #@wait          ; really x-ref
        bne     ..ot
        ldx     sxref
        beq     ..ot2           ; nothing to restore
        stx     refdv

..ot    cmp     #@obj           ; object?
        bne     ..ot1
         pha
         jsr    setobuf         ; set obj buff
         pla
..ot1   cmp     #@err           ; err?
        bne     ..ot2
        ldy     #0              ; clr err indx
        sty     eix
..ot2   tay
        lda     opttab-@obj,y
        eor     #$ff
        and     optflag
        ora     klist           ; master list control
        sta     optflag
        jmp     ..ot4

off@    jsr     gtoke           ; turn-off option
        cmp     #@wait          ; really x-ref
        bne     oqt

        ldx     refdv           ; save x-ref device
        beq     oqt             ; don't over write
        stx     sxref
        ldx     #0              ; and inhibit x-refs
        stx     refdv

oqt     cmp     #@obj           ; object?
        bne     ..ot3
         pha
         jsr    write           ; write cur buf
         pla
..ot3   tay
        lda     opttab-@obj,y
        ora     optflag
        sta     optflag
..ot4   jsr     gtoke           ; get nxt
        cmp     #@cma           ; another option?
        beq     @@opt           ; yup
        bne     ..z5            ; unc, go list


;       directive end
;       -------------
@@end   lda     entdv           ; get dev
        bne     ..z5            ; ignore if dev
         jsr    write           ; out buffer
         lda    #$80
         sta    eofflag         ; set eof
..z5    jmp     spp1
        endproc



;       set assembler system variable
;       -----------------------------
@@set   proc
        jsr     write           ; for .set 6,exp
        jsr     xscan           ; get id number
        txa
        cmp     #sysmax         ; illegal?
        bcc     ..sys0          ; nope
         lda    #27             ; invalid .set
         jmp    sp3er

..sys0  pha                     ; save for later
        inc     cox             ; skip comma
        jsr     xscan           ; get set value
        pla                     ; get index
        tax                     ; save sys #
        lda     fr0             ; get para 2
        cpx     #1
        beq     sysmar          ; left margin, set 1
        bcc     sysbyte         ; if set 0
        cpx     #4              ; page size?
        bne     ..sys1          ; if not
                        ;   form size, set 4
        sbc     #8              ; cs
        cmp     #20             ; must be >20
        bcs     ..sys1          ; ok
        bcc     ..sys3

sysbyte tay                     ; byte format
        beq     ..sys2
        cmp     #5
        bcc     ..sys1
..sys2  lda     #4

sysmar  and     #$1f            ; left margin
..sys1                  ;   just store sets
        sta     sysset,x
        cpx     #6              ; set offset?
        bne     ..sys3
         lda    fr0+1
         sta    sysset+1,x
         jsr    setobuf         ; set obj buff
         jmp    spp2            ; different list
..sys3  jmp     spp3


;       comment statements
;       ------------------
@@semi  ldy     #3              ; labeled line?
        lda     (exbuf),y
        bpl     ..sys3          ; no
        jmp     spp9            ; list with label

;       .float directive
;       ----------------
fmor    jsr     gtokel          ; done or comma
        beq     ..fd            ; yup
        cmp     #@comment       ; end of stmt?
        beq     ..fd            ; yup

@@float jsr     gtoke           ; string size
        jsr     processlaa      ; move to lbuff
        lda     #eol            ; term ascii #
        sta     lbuff,x
        lda     #0              ; reset lbuff index
        sta     cix
        jsr     intlbf          ; init lbuff

        ldx     #-6             ; save conflicting ram
..scfr  lda     buffers+6,x
        pha
        inx
        bne     ..scfr
        jsr     afp             ; to floating point

        ldx     #5              ; restore 'em now
..rcfr  pla
        sta     buffers,x
        dex
        bpl     ..rcfr
        bcc     ..fpg           ; if legal
         lda    #nfpr           ; not floating point!
         jsr    puterr

..fpg   ldx     #-6             ; move to obj buff
..fpm   lda     fr0+6,x
        jsr     object
        inx
        bne     ..fpm           ; if more...
        beq     fmor            ; unc

..fd    lda     setbyte         ; listing format
        jmp     putinbuf0       ; list & return
        endproc


;       directive macrox
;       ----------------
@@macrox proc
        jsr     gtoke           ; handle name
        jsr     processlab
        lda     (srcadr),y      ; defined?
        and     #4
        bne     c1              ; yup

d4      lda     #nesm           ; nesting error
         db $2c
        lda     #udfm           ; undef macro
        jmp     sp3er           ; list w/error

c1      ldx     mcrsk           ; get stk ptr
        cpx     #topsk          ; at top of stk?
        bcs     d4              ; yup

        lda     stmcur          ; save next statement
        sta     svcur,x         ; pointers for
         lda    stmcur+1        ; return
         sta    svcur+1,x
        lda     entdv           ; save enter dev
        sta     svedv,x
        lda     curaug          ; save aug ptr
        sta     svaug,x

        txa                     ; cc from "c1"
        adc     #4              ; update stack level
        sta     mcrsk

        ldy     #2
        lda     (srcadr),y      ; get macro
        sta     stmcur+1        ; exec adress
         dey
         lda    (srcadr),y
         sta    stmcur
        jsr     parset
        jsr     spp3            ; list here
        jsr     wrtxref         ; references

                        ; test & set pass 3
        lda     eaf             ; ck asm pass #
        bmi     p0000           ; br if pass 1
        bit     mpf             ; ck mini-pass #
        bmi     p0000           ; br if pass 3

        sta     mpf             ; set mini pass 2
        asl     a
        sta     eaf             ; set asm pass 1

        lda     asmpc           ; save cur pc
        sta     svcpc
         lda    asmpc+1
         sta    svcpc+1
        lda     stmcur          ; save macro start
        sta     svstm           ; adress
        lda     stmcur+1
        sta     svstm+1

p0000   lda     #-1             ; inprocess macro
        sta     inmac
        rts
        endproc


;       set-up parameters & stack
;       -------------------------
parset  proc
        ldy     curaug          ; get cur ptr
        lda     (stack),y       ; para cnt of
        clc                     ; last macro
        adc     #1              ; include %0
        asl     a               ; four byte pushes
        asl     a
        adc     curaug          ; plus pointer
        bcs     fnest1          ; oops!

        sta     augx            ; save curaug ptr
        tay
        lda     #0
        sta     entdv           ; get from mem

        sta     (stack),y       ; init stack
        iny                     ; to zero
        sta     (stack),y       ; parameters
        iny
        lda     srcadr          ; macro name
        sta     (stack),y       ; adress
        iny
        lda     srcadr+1
        sta     (stack),y
        iny
        sty     tem0            ; save pointer

parget  jsr     gtokel          ; any parms?
        beq     pardone         ; nope, all done
        cmp     #@comment
        beq     pardone         ; nope, all done
        cmp     #@cma           ; a comma?
        beq     parget          ; yup, get next
        cmp     #@qu            ; string literal?
        bne     parexp          ; must be expression
        jsr     mstrlitr        ; string literal
        beq     parstrv         ; unc

fnest1  jmp     fnest           ; fatal nest er

parexp  dec     cox             ; pop index
        jsr     xscan           ; get result
parstrv pha                     ; save high byte
        txa
        ldy     tem0            ; get stack ptr
        sta     (stack),y
        iny
        pla                     ; push hi byte
        sta     (stack),y
        iny
        lda     srcadr          ; mv string adress
        sta     (stack),y
        iny
         lda    srcadr+1
         sta    (stack),y
         iny                    ; point to next
        sty     tem0
        beq     fnest           ; wrap-around!

        ldy     augx            ; bump %0
        lda     (stack),y       ; for this
        clc                     ; parameter
        adc     #1
        sta     (stack),y
        jmp     parget          ; loop until done

pardone lda     augx            ; set new curaug
        sta     curaug
        rts

fnest   lda     #nesm           ; fatal nesting er
        jmp     abort

;       macro string literal
;       --------------------
mstrlitr
        jsr     gtoke
        jsr     processlaa      ; set-up string
        lda     #eol            ; tack eol to string
        sta     lbuff,x         ; makes it unique!!

        inx                     ; bump indices
        inc     ltemp
        jsr     processla       ; do label
        inc     cox             ; skip other "

        ldx     ltemp           ; get string literal
        dex                     ; length - eol
        lda     #0              ; rtn length high, &
        sta     (svesa),y       ; strings always defined
        rts
        endproc


;       directive endm
;       --------------
@@endm  proc
        ldx     mcrsk           ; get stk ptr
        beq     me1             ; nesting err!
        dex                     ; pop stk ptr
        dex
        dex
        dex
        bne     act             ; mini pass still active

                        ;  mini-pass 3
        asl     mpf             ; test end pass 3
        beq     act             ; end pass 3
        lda     #$40            ; reset asm pass 2
        sta     eaf
        lda     svcpc           ; restore asmpc
        sta     asmpc
         lda    svcpc+1
         sta    asmpc+1
        lda     svstm           ; pnt back to 1st
        sta     stmcur          ; level macrox
         lda    svstm+1
         sta    stmcur+1
        lda     #0              ; clr errs
        sta     eix
        rts

act     stx     mcrsk
        lda     svcur+1,x       ; restore old
        sta     stmcur+1        ; pointers
        lda     svcur,x
        sta     stmcur
        lda     svaug,x
        sta     curaug
        lda     svedv,x
        sta     entdv

        jsr     spp3            ; list
        lda     mcrsk           ; at macro bottom?
        bne     ..cd            ; no
         sta    inmac           ; clear macrox
..cd    rts                     ; ret

me1     jmp     d4              ; nesting err
        endproc



;       directive macro defination
;       --------------------------
@@macro proc
        jsr     gtoke
        jsr     processlab      ; macro name
        jsr     wrtxref         ; references
        ldy     #3
        lda     (srcadr),y      ; ck for dup
        ldx     eaf             ; get pass #
        bpl     m0              ; br if pass 2
        and     #4
        bne     ddf             ; a dup macro def

        lda     #4              ; define macro
        sta     (srcadr),y
        dey
        lda     macend+1        ; adress of start
        sta     (srcadr),y      ; of macro exc
        dey
        lda     macend
        sta     (srcadr),y

                        ;   read in rest of defination
m4      jsr     gscl            ; get line
        bcc     m5              ; br if not eof
..xlmr  lda     #lmr            ; linked macro err
        jmp     abort

m5      sty     tem             ; save line len
        ldx     #macend         ; expand macro tab
        jsr     explow
        ldy     tem
m6      dey                     ; mv line to
        lda     (exbuf),y
        sta     (svesa),y       ; defination
        tya                     ; table
        bne     m6

        jsr     gtoke           ; look for .endm
        cmp     #@macro         ; ck nested def
        beq     fnest2
        cmp     #@endm          ; done?
        bne     m4
        rts
fnest2  jmp     fnest

                        ; macro defination pass 2
                        ;  just list out defination
ddf     lda     #$14            ; flag dup def
        sta     (srcadr),y
m0      and     #$10            ; a dup def?
        beq     m2
        lda     #dupm           ; tell user
        jsr     puterr
m2      jsr     spp3            ; list this line
m8      jsr     gscl            ; get nxt line
                        ;  this branch needed for a
                        ;  dup name and missing
        bcs     ..xlmr          ; .endm error!
        jsr     spp3            ; and list
        jsr     gtoke           ; look for .endm
        cmp     #@macro
        beq     fnest2
        cmp     #@endm
        bne     m8
        rts
        endproc



;       directive = and .=
;       ------------------
@@equal proc
        sec                     ; *** flag as equate ***
        bcs     eqv

@@.equal
        clc                     ; *** flag as set ***
        lda     #0              ; clear dup err
        sta     eix
eqv     lda     srcadr+1        ; set "labpt"
        sta     labpt+1
        lda     srcadr
        sta     labpt
        bcc     eqe             ; go if a set

        ldy     #3              ; set primary undefined
        lda     (labpt),y
        ora     #1
        sta     (labpt),y

eqe     lda     mpf             ; ck mini-pass
        beq     do              ; if not mini-pass

                        ; do internal symbol processing
        ldx     ltemp           ; get lab len
        lda     #'('            ; assign type char
        sta     lbuff,x
        inc     ltemp           ; bump lab len
        inx                     ; bump indx
        jsr     processla       ; handle symbol

        lda     (srcadr),y
        bit     mpf             ; get mini-pass #
        bmi     eqmp3           ; br if pass 3
        and     #8              ; lab already set?
        bne     do              ; ye, normal process

                        ; assign expression to label & label(
        lda     #8
        sta     (srcadr),y      ; set it
        jsr     pscan           ; execute exp
        ldy     #1              ; move value to
        txa
        sta     (srcadr),y      ; secondary label(
        iny
        lda     fr0+1           ; high
        sta     (srcadr),y
        jmp     dasgn           ; assign to primary

                        ; re-assign to primary
eqmp3   and     #8              ; lab reset?
        beq     do              ; yup, go normal
        lda     #0
        sta     (srcadr),y
        jsr     pscan           ; ex exp
        ldy     #1
        lda     (srcadr),y      ; restore old
        tax                     ; result
        iny
        lda     (srcadr),y
        sta     fr0+1           ; so list is correct
        stx     fr0
        jmp     dasgn           ; assign to primary


do      jsr     xscan           ; get #
dasgn   ldy     #2
        sta     (labpt),y       ; assign val
        txa
        dey
        sta     (labpt),y
        ldy     #3
        lda     (labpt),y
        and     #$fe            ; clr poss undef
        ora     res             ; set/rest def bit
        ora     #2              ; set equated bit
        sta     (labpt),y
        jmp     spp2            ; list

pscan   lda     srcadr          ; save secondary
        pha                     ; label pointer
        lda     srcadr+1
        pha
        jsr     xscan           ; ex exp!
        pla                     ; restore pointer
        sta     srcadr+1
        pla
        sta     srcadr
        rts
        endproc


;       directive if
;       -----------
@@if    proc
        jsr     xscan           ; eval exp
        ora     fr0             ; test result
        php                     ; save stat
        jsr     wrtxref         ; references
        jsr     spp2            ; list line
        plp
        beq     if1             ; else or endif

        lda     #qelse          ; else to stop
        jmp     rpush           ; assembly

oops    lda     eofflag
        cmp     #$40            ; ck end .include
        beq     rskmgr0         ; if .inc, not end
        jmp     enpass

if1     lda     #qendif         ; look for else or
        jsr     rpush           ; endif to ret
        jmp     rskmgr          ; go look for match

                        ; run-time stack handler
rskmgr0 lda     #0              ; clear .include eof
        sta     eofflag

rskmgr  jsr     gscl            ; get line
        bcs     oops            ; encountered eof
        jsr     gtoke           ; test this line
        bpl     rm7             ; br if no label

        jsr     ifs             ; skip past label
        jsr     gtoke           ; next token
rm7     cmp     #@if            ; nested if?
        bne     rm2             ; nope

        lda     #qdumy          ; push dummy type
        jsr     rpush

rm1     lda     optflag         ; test condt option
        and     #8              ; list conditional?
        beq     hclst           ; yup!
        jsr     bckup
        bne     rskmgr          ; unc

hclst   jsr     spp3            ; list this line
        jmp     rskmgr          ; get nxt

rm2     cmp     #@else          ; it else?
        bne     rm3             ; no
        ldy     rstk            ; stack ptr
        lda     runstk,y        ; get type
        beq     rm1             ; br if dummy
        cmp     #qendif         ; an exit?
        bne     rm4             ; nest err
        lda     #qselse
        sta     runstk,y        ; set else fnd
rm6     jmp     spp3            ; list & ret

rm4     lda     #ntpe           ; nesting err
        jsr     puterr
        jmp     rm1

rm3     cmp     #@endif         ; terminal?
        beq     @@endif         ; yup
        cmp     #@endm          ; macro end?
        bne     rm1

        lda     inmac           ; execute .endm only
        beq     rm1             ; if doing macrox
        jsr     @@endm          ; exec endm
        jmp     rskmgr
        endproc


;       directive endif
;       ---------------
@@endif proc
        ldy     rstk            ; get cur ptr
        beq     rm5             ; nesting err
        dec     rstk            ; pop stack
        beq     rm6             ; at bottom!

        lda     runstk,y        ; make sure nxt
        beq     ee1             ; isn't a dummy
        bne     rm6             ; unc

rm5     lda     #ntpe           ; nesting err
        jmp     sp3er
        endproc



;       directive else
;       --------------
@@else  proc
        ldy     rstk            ; get stk ptr
        beq     rm5             ; err!
        lda     runstk,y        ; get type
        beq     rm6             ; null
        cmp     #qselse         ; else set?
        beq     rm5             ; err!
ee1     jmp     rm1             ; look for endif
        endproc



;       push to run-time conditions stk
;       -------------------------------
rpush   proc
        ldy     rstk
        iny
        cpy     #skmx+1         ; at top of stk?
        bcs     cp1             ; yup, err
         sta    runstk,y
         sty    rstk
         rts
cp1     lda     #sofe           ; stack overflow
        jmp     puterr          ; and ret
        endproc



;       execute expression (rev 4/15/83)
;       --------------------------------
;       a = high, x = low, y = adrmode

xscan   proc
        ldy     #tose           ; init stack
        sty     aptr
        lda     #@eol           ; set pseudo end
        sta     argops,y
        ldy     #0
        sty     res             ; clear resolved
        sty     optr            ; init operator stack
        sty     topstk          ; fr0 push flag

                        ; simple expression? (faster)
        jsr     getn            ; try get value
        bcs     xs2             ; no, an op
        stx     fr0             ; possible result
        sta     fr0+1
        ldy     cox             ; test end exp
        lda     (exbuf),y
        cmp     #@pcmax         ; a terminator?
        bcs     quif            ; yup, all done ***
        dec     topstk          ; set push fr0 flag

                        ; not simple expression
xs9
xs1     jsr     getn            ; get #
        bcs     xs2

        bit     topstk          ; at top of stack?
        bpl     tsk
        pha                     ; push fr0 to stack
        ldy     optr
        lda     fr0+1
        sta     argops,y
        iny
        lda     fr0
        sta     argops,y
        iny
        sty     optr
        pla
tsk     stx     fr0             ; set new fr0
        sta     fr0+1
        dec     topstk
        jmp     xs1

xs2     sta     svop            ; save oper
        tax
        lda     onprec-@plus,x
        sta     svnp            ; save precedance
xs8     ldy     aptr
        ldx     argops,y        ; get old
        lda     ofprec-@plus,x
        cmp     svnp
        bcc     xs3             ; push oper
        tay
        beq     quit            ; done
xs7     inc     aptr
        jsr     xs4
        jmp     xs8
        endproc


;       execute operator
;       ----------------
xs4     proc
        lda     #high (@xplus-1)
        pha
         lda    opertrl-@plus,x
         pha
        rts
        endproc


xs3     proc
        lda     svop            ; push operator:
        dey                     ; bump index
        cpy     optr            ; stack overflow?
        bcc     quitr           ; yup!!
         sta    argops,y        ; ok, so save it
         sty    aptr
         jmp    xs1

quitr   lda     #6              ; exp complex err
        jsr     puterrasm
        jmp     quit

quif    inc     cox             ; adjust cox, fast exit
quit    lda     res             ; exp resolved?
        beq     xs5             ; yup
        lda     #5              ; else, undef error
        jsr     puterrasm

xs5     dec     cox             ; pop index
        ldy     cox             ; get adressing mode
        lda     (exbuf),y
        tax
        ldy     endm-@pcmax,x   ; mode in y

                        ; determind absolute or z page
noudf   ldx     fr0             ; res low
        lda     fr0+1           ; res hi
        ora     res             ; force non z page
        beq     exx0            ; z pg anyway
        cpy     #5              ; non z instr?
        bcc     exx0            ; no
        iny                     ; bump mode to abs
        iny
        iny
exx0    lda     fr0+1           ; restore high byte
        rts
        endproc



;       pop entry from argops
;       ---------------------
opop    proc
        ldy     optr            ; get stack pointer
        dey                     ; pop index
        ldx     argops,y        ; get low byte in x
        dey                     ; pop index
        lda     argops,y        ; get high byte
        sty     optr            ; save new pointer
        rts
        endproc


;       get num, a = high; x = low  cc
;       a = operator token cs
;       ------------------------------
getn:   proc
        ldy     cox             ; get next token
        inc     cox
        lda     (exbuf),y
        bmi     issymb          ; branch if label
getne:  cmp     #@apost+1       ; other type?
        bcs     gn1             ; yup, go test
nomp    lsr     a               ; status: word or byte
        ldy     cox             ; get low byte
        inc     cox
        lda     (exbuf),y
        tax                     ; in x
        bcs     gn2             ; if carry, then get high
        lda     #0              ; else, high =0
        rts
gn2     clc                     ; set no-operator status
        jmp     gtoke           ; get high

gn1     cmp     #@cpc           ; pc?
        beq     dopc
        bcc     domp            ; do macro param
        rts                     ; cs =operator

dopc:   clc
        lda     asmpc
        adc     objix           ; add in cur
        tax
        lda     asmpc+1
        adc     #0
        rts                     ; return, carry is clear

                        ; process symbol
issymb  jsr     processlab      ; v1.103
        lda     (srcadr),y      ; defined byte
        bcc     yesdf           ; go if not newly inserted
         lda    #$81            ; special for .ref
         clc
         bcc    yesrf           ; unc

yesdf   and     #1              ; just move status
        ora     res             ; to resolved flag

yesrf   sta     res
        ldy     #1              ; get value
        lda     (srcadr),y
        tax
        iny
        lda     (srcadr),y
        rts                     ; carry is clear
        endproc



b2      proc
        lda     #32             ; bad parameter
        jsr     puterr          ; err & ret
        lda     #0              ; ret zero
        tax
        clc
        rts
        endproc


;       get parameter from a macro call
;       -------------------------------
domp    proc
        jsr     gtoke
        cmp     #@lp            ; is it %(exp)?
        bne     b4              ; nope
         lda    #@elp           ; special operator
         rts                    ; carry set

b4      jsr     nomp            ; get param #
valuep  cmp     #0              ; no param >255
        bne     b2              ; val > 255
        txa
        ldy     curaug          ; get aug pointer
        cmp     (stack),y       ; test param range
        beq     b3
        bcs     b2              ; out of range

b3      ldx     mcrsk           ; ck for zero stk
        beq     b2
        asl     a               ; form stack indx
        asl     a               ; multiples of 4
        clc                     ; as offset from
        adc     curaug          ; cur aug ptr
        tay                     ; x
        lda     (stack),y       ; get low
        iny
        tax
        lda     (stack),y       ; get high
        pha                     ; save high
        iny
        lda     (stack),y       ; string adress
        sta     srcadr
        iny
        lda     (stack),y
        sta     srcadr+1

        ldy     #3              ; get defined stat
        lda     (srcadr),y
        ora     res
        and     #1              ; out other stuff
        sta     res
        pla                     ; restore high
        clc
        rts
        endproc



;       test for exp end:
;        cs = end, y = adr mode
;        cc = operator, or number type
;       ------------------------------
endexp? proc
        ldy     cox             ; get token
        lda     (exbuf),y
        cmp     #@pcmax         ; tst first term
        bcs     endp            ; br if terminal
        rts                     ; an operator

endp    tax                     ; indx to mode tab
        bmi     b1              ; test for label
        ldy     endm-@pcmax,x   ; get mode in y
        rts
b1      clc
        rts
        endproc



;       process labels
;       --------------
; returns y=3
; pointing to define byte

processlaa proc
        and #$7f                ; lab len
        sta     ltemp           ; sav
        ldx     #0              ; clr indx
        ldy     cox
labloop lda     (exbuf),y       ; move to
        sta     lbuff,x         ; text buff
        inx
        iny
        cpx     ltemp
        bcc     labloop
        sty     cox             ; point to nxt
        rts
        endproc


processlab proc
        jsr processlaa
        lda     lbuff           ; check local

        $if sorcim
         cmp    #':'
         bne    ..rc1
          lda   #'?'
          sta   lbuff
..rc1    ldy    lbuff-1,x       ; last a ":"?
         cpy    #':'
         bne    ..rc2
         dex            ; ignore the colon
         dec ltemp
..rc2
        $endif

        cmp     #'?'
        bne     processla       ; nope

        lda     localix
        beq     processla       ; first unchanged
         sta    lbuff           ; do local counter

processla
        txa                     ; form exp size
        clc
        adc     #6
        sta     labof

                        ; get label character address
        lda     lbuff           ; set-up
        cmp     #'?'            ; less, use "?"
        bcs     limith
         lda    #'?'
         bne    getaddr         ; unc
limith  cmp     #'Z'+1          ; test top
        bcc     getaddr
         lda    #'Z'

getaddr asl     a               ; double index
        tax
        ldy     spointer-'?'*2+1,x      ; high
        bmi     newlabel                ; yup, a first

                                ; search loop
        lda     spointer-'?'*2,x        ; low
nextlink
        clc
        adc     smbltp          ; add in absolute
        sta     srcadr
         tya
         adc    smbltp+1        ; add in high
         sta    srcadr+1

        ldx     #0              ; compare offsets
        lda     labof
        cmp     (srcadr,x)
        bne     testlink

        ldy     #6              ; do characters
sy1     lda     (srcadr),y
        eor     lbuff-6,y
        beq     sy2             ; a match

testlink
        ldy     #4              ; get link low
        lda     (srcadr),y
        tax
        iny
        lda     (srcadr),y
        tay
        txa
        cpy     #-1
        bne     nextlink
        beq     nointab

sy2     iny
        cpy     labof
        bcc     sy1
         clc
         jmp    rightone
        endproc


;       insert symbol in table
;       ----------------------
; x is index to "spointer"

newlabel proc
        lda     smblend
        sta     srcadr          ; set 'srcadr'
        sec
        sbc     smbltp          ; 'spointer' entry
        sta     spointer-'?'*2,x        ; link
         lda    smblend+1
         sta    srcadr+1
         sbc    smbltp+1
         sta    spointer-'?'*2+1,x      ; high

nointab lda     smblend         ; set "svesa"
        sta     svesa
        clc                     ; add low expand
        adc     labof
         tay                    ; for mem test
         lda    smblend+1
         sta    svesa+1
         adc    #0

        cmp     mtp+1           ; within?
        bcc     esafe
        cpy     mtp             ; low within?
        bcc     esafe           ; yes
        beq     esafe
        jmp     exxi            ; out of memory

esafe   sta     smblend+1
        sty     smblend

                        ;  link this entry to last entry
                        ;   "srcadr" points to last
        ldy     #4              ; subtract "smblpt"
        sec
        lda     svesa
        tax
        sbc     smbltp
        sta     (srcadr),y
         iny
         lda    svesa+1
         sbc    smbltp+1
         sta    (srcadr),y

        lda     svesa+1         ; make "svesa"="srcadr"
        sta     srcadr+1
        stx     srcadr

        ldy     #0
        lda     labof           ; offset to nxt lab
        sta     (svesa),y
        tya                     ; a =0
        iny                     ; y =1
        tax                     ; x =0
        sta     (svesa),y       ; zero adress
        iny
        sta     (svesa),y
        lda     #1
        iny
        sta     (svesa),y       ; set undef

        iny                     ; set last label link
        iny
        lda     #-1
        sta     (svesa),y       ; high byte is flag

mvelp   iny                     ; move symbol
        lda     lbuff,x
        sta     (svesa),y
        inx
        cpx     ltemp           ; past last one?
        bcc     mvelp           ; no, get next
rightone
        lda     refdv           ; build x-ref?
        bne     buildrf         ; yup
        ldy     #3
        rts


                        ; build x-reference data
buildrf php                     ; save carry
        bit     eaf             ; pass 2?
        bvc     noxloc          ; no
        ldy     #6
        lda     (srcadr),y      ; is it a local?
        cmp     #'?'+1

        $if dolocal
          bcs   noxloc          ; only locals
        $else
          bcc   noxloc          ; yup
        $endif

        ldx     xrindx          ; buffer full?
        cpx     #xrmax+1
        bcs     xrfer           ; yup

        sec
        lda     srcadr
        sbc     smbltp          ; dump table link
        sta     xrefbuf+1,x     ; format:
        lda     srcadr+1        ; high/low
        sbc     smbltp+1
        sta     xrefbuf,x

                        ; already referenced?
        ldy     xrindx
duprfx  dey                     ; pop to last entry
        dey
        bmi     newrfx          ; no match, new entry
        lda     xrefbuf,y       ; high
        cmp     xrefbuf,x
        bne     duprfx          ; try next entry
        lda     xrefbuf+1,y
        cmp     xrefbuf+1,x
        bne     duprfx          ; next entry
        beq     noxloc          ; don't update index

newrfx  inx                     ; point to next
        inx
        stx     xrindx
noxloc  ldy     #3
        plp                     ; get carry status
        rts

xrfer   lda     #xrerr          ; too many x-refs
        jsr     puterr
        jmp     noxloc
        endproc



;       label at start of line
;       ----------------------
labint  proc
        jsr     processlab
        lda     (srcadr),y
        and     #$fe            ; clr undef bit
        pha                     ; save
        bmi     lt1             ; already there
        lda     asmpc+1         ; define it
        dey
        sta     (srcadr),y
        lda     asmpc
        dey
        sta     (srcadr),y
        jmp     lt3

lt1     and     #2              ; equated?
        bne     lt3             ; yes
        lda     asmpc+1         ; must match pc
        dey
        cmp     (srcadr),y
        bne     lt2
        lda     asmpc
        dey
        cmp     (srcadr),y
        beq     lt3             ; ok
lt2     lda     #13             ; phase error
        jsr     puterrasm
lt3     pla                     ; get define byte
        pha
        asl     a
        bpl     lt4
         lda    #7              ; duplicate label
         jsr    puterrasm
lt4     ldy     #3
        pla
        bit     inmac           ; in macro exec?
        bpl     lt6             ; no, regular lab

        tax                     ; def byte must =0
        beq     lt5             ; don't set flag

lt6     ora     eaf             ; set pass #
lt5     sta     (srcadr),y
        lda     refdv           ; doing x-ref?
        beq     lt7             ; no
        bit     eaf             ; only if pass 2
        bmi     lt7
        bit     lbuff           ; locals will clear v

        $if dolocal
          bvs   lt7
        $else
          bvc   lt7             ; don't flag
        $endif

        lda     #rflab          ; flag labeled line
        jsr     xrout
lt7     rts
        endproc



@@local proc
        inc     localix         ; tough!!
        lda     localix         ; get fancy...
        cmp     #'?'            ; at max?
        bcc     lsafe           ; yup
        lda     #hlocer         ; tell user
        jsr     puterr
        lda     #'?'

lsafe   sta     fr0
        lda     #0
        sta     fr0+1
        jmp     spp2            ; list & return
        endproc


;       directive title
;       ---------------
@@title proc
        ldx     #0              ; string statement
        jsr     msrg
        bne     ..pp0           ; unc

;       directive error
;       ---------------
@@error ldx     #2
        jsr     msrg
        lda     #0              ; force error "0"
        jsr     puterrasm
..pp0   jmp     spp3
        endproc




;       put .page, .title, & .error strings in memory
;       ---------------------------------------------
msrg:   proc
        jsr     gtoke
        cmp     #@qu            ; string?
        beq     g0              ; yup
        rts

g0      clc
        lda     strings
        adc     mstb,x
        sta     srcadr
        lda     strings+1
        adc     #0
        sta     srcadr+1
        jsr     gtoke7          ; get str len
        tax
        lda     #0
        sta     temp
        cpx     #strmax+1       ; check length
        bcc     ptmov
        ldx     #strmax

ptmov   jsr     gtoke7          ; get 7 bit char
        ldy     temp
        inc     temp
        sta     (srcadr),y      ; to buffer
        dex
        bne     ptmov
        ora     #$80            ; flag last char
        sta     (srcadr),y
        rts
        endproc



;       directive tab
;       -------------
@@tab   proc
        jsr     xscan           ; get first #
        stx     tab1
        inc     cox             ; skip ,
        jsr     xscan           ; get second #
        stx     tab2
        inc     cox             ; skip ,
        jsr     xscan           ; get third #
        stx     tab3
        jsr     setztab         ; reset zp tabs
        jmp     spp3            ; go list
        endproc



;       test for save file & pop file size
;       ----------------------------------
validf: proc
        jsr     gobj            ; test ck byte
        cmp     #ckbyt
        bne     ifail           ; oops!
        jsr     gobj            ; test ck byte
        cmp     #ckbyt
        bne     ifail           ; oops!

popsizef:
        jsr     gobj            ; file size
        sta     filesize
        jsr     gobj
        sta     filesize+1      ; store & return
        rts

ifail   lda     #23             ; not save error
stopasm jmp     abort
        endproc


;       directive ds
;       ------------
@@ds    proc
        jsr     xscan           ; get number
        bit     pcdef           ; pc defined?
        bpl     ..ae9           ; no!
        txa                     ; calculate new pc
        clc
        adc     asmpc
        tax                     ; low in x
        lda     fr0+1
        adc     asmpc+1
        jmp     ..ae0           ; join *=

;       directive *=
;       ------------
@@aster jsr     xscan           ; get num
..ae0   ldy     res             ; defined?
        beq     ..ae1           ; yup
..ae9   lda     #14             ; stop!
        bne     stopasm

..ae1   dey                     ; set y =$ff
        sty     pcdef           ; set pc defined
        bit     eaf             ; get pass #
        bpl     ..ae2           ; branch if pass 2
        sta     asmpc+1         ; else, just set new
        stx     asmpc           ; pc adress
        jmp     spp1            ; all done

..ae2   pha                     ; save new pc
        txa
        pha
        jsr     write           ; write, if any
        jsr     spp1            ; list line
        pla                     ; get new pc
        sta     asmpc
        pla
        sta     asmpc+1
        jmp     setobuf         ; set obj pc & return
        endproc



;       writes object to device
;       -----------------------
write:  proc
        lda     objptr
        cmp     #4              ; test null
        beq     ..empt
        sec
        sbc     #5              ; for end adress
        clc                     ; compute end adress
        ldy     #0
        adc     (objbuf),y
        ldy     #2
        sta     (objbuf),y
        dey
        lda     (objbuf),y
        adc     #0
        ldy     #3
        sta     (objbuf),y
        pha                     ; save end adress
        dey
        lda     (objbuf),y
        pha

        jsr     runout          ; write it
        pla
        clc                     ; set new start adress
        adc     #1

                        ; high byte on stack, carry
                        ; must be clear at entry
..sovx  ldy     #0
        sta     (objbuf),y
        pla
        adc     #0
        iny
        sta     (objbuf),y
        ldy     #4              ; reset pointer
        sty     objptr
..empt  rts


;       set up object buffer
;       --------------------
setobuf
        lda     asmpc
        clc
        adc     offset
        tax
        lda     asmpc+1
        adc     offset+1
        pha
        txa
        clc
        bcc     ..sovx
        endproc
$title(MAC65/XL Cartridge  (c) 1988 ICD Inc)
$subtitle(Cartridge bank C for MAC65)

        seg     bankc

;       symbol table sorter
;       -------------------
sortsymbols: proc
sort0   lda     #0              ; clear exchange
        sta     swapsort
        lda     #'@'-1          ; init index
        sta     sindex

sort1   lda     sindex          ; get index
        inc     sindex          ; bump
        cmp     #'Z'+1          ; done a pass?
        bcc     sort2           ; no
        lda     swapsort        ; do another?
        bne     sort0           ; yup
        rts

sort2   asl     a               ; double index
        tax
        ldy     spointer-'?'*2+1,x      ; high link
        bmi     sort1           ; nothing

        jsr     formtop         ; form topsort

        clc                     ; form labelptr
        txa
        adc     #low (spointer-'?'*2-4)
        tax
        lda     #high (spointer-'?'*2-4)
        adc     #0

                        ; a = high  x = low
sort11  sta     bckptr+1
        stx     bckptr

        ldy     #5              ; last?
        lda     (labelptr),y
        bmi     sort1           ; yup

        jsr     formlab         ; form fwd pointer
        sty     fwdptr
        sta     fwdptr+1

                        ; comparison
        ldy     #0              ; compare offsets
        lda     (labelptr),y
        cmp     (fwdptr),y
        bcc     sort7
        beq     sort7
        lda     (fwdptr),y
        dey                     ; set y =$ff

sort7   sta     maxsort
        dec     maxsort
        sty     hilosort
        ldy     #6              ; compare chars
        bne     sort9           ; skip 1st char

sort8   iny                     ; bump to next
        lda     (labelptr),y
        cmp     (fwdptr),y
        bcc     notswap
        bne     yesswap

sort9   cpy     maxsort         ; at length?
        bcc     sort8           ; no
        lda     hilosort        ; by offset
        beq     notswap         ; smaller in place

yesswap ldy     #5              ; index for exchange
        sty     swapsort        ; show exch (<>0)

sort6   lda     (labelptr),y
        tax
        lda     (bckptr),y
        sta     (labelptr),y
        pha                     ; ff
        txa
        sta     (bckptr),y

        lda     (fwdptr),y
        sta     (labelptr),y
        pla                     ; ff
        sta     (fwdptr),y

        dey
        cpy     #4
        bcs     sort6

        ldx     fwdptr          ; save fwd pointer
        lda     fwdptr+1
        bne     sort11          ; unc

notswap
        ldx     labelptr        ; save topsort
        lda     labelptr+1

        ldy     fwdptr          ; point to new
        sty     labelptr        ; topsort
        ldy     fwdptr+1
        sty     labelptr+1
        bne     sort11          ; unc
        endproc


;=============================================
;       decode & move text line pointed
;       to by stmcur to inbuff
;       -------------------------------
_edtv:  proc
        sta     cmdlg           ; save line length
        stx     labpt           ; pointer to
        sty     labpt+1         ; tokenized line
        ldy     #0              ; clear text index
        sty     cix
        jsr     intlbf
        ldx     supr            ; do line #?
        bne     cov2            ; nope!

        iny                     ; move line # (y=1)
        lda     (labpt),y       ; to fr0 for
        sta     fr0+1           ; conversion to ascii
        lda     (labpt,x)       ; (x=0)
        sta     fr0
        jsr     inii            ; integer to ascii
        jsr     intlbf
        ldy     #-1

..cc1   iny                     ; find end of number
        lda     (inbuff),y
        bpl     ..cc1
        and     #$7f            ; mask 7th bit
        sta     (inbuff),y
        iny
        sty     cix             ; start of text index
        bit     system          ; bump colunm?
        bmi     ..cc2
         iny
..cc2   sty     rpsp            ; char past l. #
        lda     eaf             ; in asm mode?
        beq     ..regl          ; no

        lda     inmac           ; in macrox?
        beq     ..regl          ; no
        lda     optflag         ; option 'no num'?
        and     #$10
        beq     ..regl

        asl     a               ; a=$20, change
chli    sta     (inbuff),y      ; line # to spaces
        dey
        bpl     chli

..regl  jsr     p20             ; always leading space
cov2    ldy     #2              ; init labpt ptr
        sty     ttix

                        ; convert tokens to text
pu1     jsr     ifd             ; get token
        bcs     pu9             ; br if end
        beq     cver            ; convert error
        bpl     cvst            ; a statement
        jsr     cvlb            ; convert label
        beq     pu1             ; unc, get nxt

cvst    cmp     #@scomment      ; comment?
        beq     stmc            ; statement commt
        tax                     ; save code
        jsr     ifstu           ; handle if indent
                        ;    do 1st tab
..tb1   jsr     p20             ; pad spaces
        cpy     ztab1           ; at nxt colunm?
        bcc     ..tb1           ; no

        txa                     ; the code
        ldy     #high statement
        ldx     #low statement  ; go cvrt
        jsr     lookup1         ; this statement
        cpx     #@macrox        ; spec case
        bne     ..tb9           ; br, not macrox

        jsr     cvmr            ; macro name

..tb9                   ;  do 2nd tab
..tb2   jsr     p20             ; pad spaces
        cpy     ztab2           ; at nxt col?
        bcc     ..tb2           ; nope
        jsr     efstu           ; endif?

                        ; convert function loop
cvfunt  jsr     ifd             ; get code
        bcs     pu9             ; done!
        bpl     pu2             ; not a symbol
        jsr     cvlb            ; do symbol
        beq     cvfunt          ; unc

pu2     cmp     #ch+1           ; ck for num
        bcc     cvnm            ; convert num
        cmp     #@comment       ; comment?
        beq     docmtf          ; function commt

        jsr     lookup          ; cvrt function
        cpx     #@apost-ch-1    ; spec case?
        bne     cvfunt          ; get nxt

        jsr     ifd             ; for ', get char
        jsr     p58             ; put in buff
        bne     cvfunt          ; unc get nxt

pu9     lda     #eol            ; do eol & ret
        jmp     p58

cver    lda     #low eres       ; point to
        sta     fr0             ; error message
        lda     #high eres
        sta     fr0+1
        jsr     loo3            ; mv error string
        bmi     docmts          ; unc

docmtf                  ;   do 3rd tab
..tb3   jsr     p20             ; fill spaces
        cpy     tab3            ; at column?
        bcc     ..tb3           ; no
        bcs     docmts

stmc    ldx     cpad            ; pad with spaces?
        beq     docmts          ; no
        jsr     ifd             ; get leading char
        jsr     p58             ; out it
..stmc  jsr     p20             ; do padding now
        dex
        bne     ..stmc

docmts  jsr     ifd             ; out ascii text
        bcs     pu9             ; br if at end
        jsr     p58             ; put to buffer
        bne     docmts          ; unc, get nxt
        endproc


;       convert number tokens
;       ---------------------
cvnm    proc
        cmp     #tdw            ; dec or hex?
        php                     ; save answer
        jsr     gvl             ; get interger
        stx     fr0             ; low byte
        sta     fr0+1           ; high byte
        plp                     ; pop num type
        bcs     dodc            ; br if dec
         jsr    hex1            ; else, hex #
         jmp    cvfunt          ; get nxt

                        ;   decimal #
dodc                    ;   mv interger to
        sta     zte4            ; holding area
        stx     zte4+1          ; x=fr0
        lda     #0              ; clear result
        sta     fr0+1
        sta     fr0+2
        sta     fr0+3

        ldy     #15             ; convert to 3
        sed                     ; byte packed bcd
dc2     asl     zte4+1          ; format
        rol     zte4
        ldx     #3
dc3     lda     fr0,x
        adc     fr0,x
        sta     fr0,x
        dex
        bne     dc3
        dey
        bpl     dc2             ; if not 16 times
        cld                     ; clear dec mode

        inx                     ; x=1
        lda     fr0+1           ; speed up for #s
        ora     fr0+2           ; less than 100
        bne     unpack          ; no, go unpack
        inx                     ; point to last bcd

unph    inx                     ; high nibble
        lda     fr0,x
        lsr     a               ; shift in position
        lsr     a
        lsr     a
        lsr     a
        jsr     out

unpack  lda     fr0,x           ; low nibble
        cpx     #3              ; set end? status
        php                     ; save it
        bcc     cvr1            ; don't force 0
        ldy     #0              ; end, always do 0

cvr1    and     #$0f            ; mask off high
        jsr     out             ; make it ascii
        plp                     ; get end? status
        bcc     unph            ; no, more to do
        jmp     cvfunt          ; get nxt function

out     bne     ..o             ; if non-zero, else
        tya                     ; ck supress flag
        bne     ..k             ; br, skip leading 0

..o     ora     #$30            ; for ascii
        jsr     p58             ; to buffer
        ldy     #0              ; clear supress flag
..k     rts                     ; return
        endproc



;       hex #
;       -----
hex1    proc
        lda     #'$'            ; leading $ sign
        jsr     p58
        lda     fr0+1           ; skip high if 0
        beq     hxl
        jsr     hexc            ; high byte

hxl     lda     fr0             ; low byte

hexc    pha                     ; save byte
        lsr     a               ; shift high
        lsr     a               ; into position
        lsr     a
        lsr     a
        jsr     hxs             ; do nibble

        pla                     ; pop byte
        and     #$0f            ; mask high
hxs     tay                     ; index into convr
        lda     hext,y          ; tab, and to buff
        jmp     p58             ; unc
        endproc


;       increment list token pointer
;       ----------------------------
ifd     proc
        inc     ttix            ; bump index
        ldy     ttix            ; get index
        cpy     cmdlg           ; set end status
        lda     (labpt),y       ; get token
        rts
        endproc



;       get nums for cvnm
;       -----------------
gvl     proc
        lsr     a               ; single or double?
        php                     ; cc=single
        jsr     ifd             ; get token
        tax                     ; low byte
        plp                     ; x
        bcs     ifd             ; get high & rtn
        lda     #0              ; set high =0
        rts
        endproc


;       convert token symbols
;       ---------------------
cvmr    proc
        jsr     ifd             ; for macrox name
cvlb    and     #$7f            ; count +128 in a
        tax                     ; use as cntr
cvl1    jsr     ifd             ; get char
        jsr     p58             ; put char
        dex                     ; pop cnt
        bne     cvl1            ; br if not done
        rts
        endproc


;       do structered listing format
;       ----------------------------
ifstu   proc
        cpx     #@if            ; is it if?
        bne     ..ata           ; no
        inc     ztab1           ; bump tabs
        inc     ztab1
        inc     ztab2           ; by 2 spaces
        inc     ztab2
        rts

efstu   cpx     #@endif         ; endif?
        bne     ..ata
        lda     ztab1           ; tst for
        cmp     tab1            ; bottom
        beq     ..ata           ; don't pop tabs
        dec     ztab1
        dec     ztab1
        dec     ztab2
        dec     ztab2
..ata   rts
        endproc


;       Handle right margin
;       -------------------
rightmr: proc
        lda     cix             ; form line length
        clc
        adc     leftmar         ; left margin
        adc     #16             ; object field
        cmp     rightmar
        beq     ..x
        bcs     hrght           ; greater than, go process
..x     jmp     inbout          ; it's within margins

hrght   lda     rightmar        ; form lbuff index
        sbc     leftmar
        sbc     #16
        tay                     ; search back index
        pha                     ; in case fail
        ldx     #12             ; limit of twelve characters

loo6    dey
        dex
        bmi     loo7            ; no space!
        lda     (inbuff),y
        cmp     #$20
        bne     loo6

        pla                     ; balance stack
        lda     #eol            ; rep space with eol
        sta     (inbuff),y
        iny
        sty     ttix            ; character past eol
        jsr     inbout          ; print this portion

vect1   lda     ttix            ; save buffer index
        pha
        jsr     tstej           ; new line stuff
        jsr     intlbf          ; if eject, reset inbuff
        lda     leftmar
        clc
        adc     #17
        adc     tab3            ; begin at comment column
        jsr     space
        pla
        tay                     ; dump remainder of line
        jmp     inbout1

                        ; no space, just chop
loo7    pla                     ; saved index
        tay
        dey
        sty     ttix
        lda     (inbuff),y
        pha                     ; replace with eol
        lda     #eol
        sta     (inbuff),y
        jsr     inbout          ; print first segment
        pla                     ; restore clobbered char
        ldy     ttix
        sta     (inbuff),y
        jmp     vect1           ; go print rest
        endproc
$title(MAC65/XL Cartridge  (c) 1988 ICD Inc)
$subtitle(DDT debugging tool - for use with MAC/65)

        seg     m_ddt
;==================================
;       Flash & mac ddt entry
;       --------------------
nkybd   proc
        sta     bank3           ; insurance...
        jmp     _nkybd          ; go to it
        endproc


;       call (jsr) entry point
;       ----------------------
ddt:    proc                            ; save some mac system
        ldy     #$80-(runstk-xcii)      ; this # should be +
..zsav0 lda     xcii-($80-(runstk-xcii)),y
        sta     (lomem),y
         iny
         bpl    ..zsav0
                                ;  save upper half of page 0
..zsav1 lda     0,y
        sta     (lomem),y
        iny
        bne     ..zsav1
        sty     zpsflg          ; set ddt mode

        tya                     ; a=0
        ldx     #14
..cbp2  sta     brkp0,x
        dex
        bpl     ..cbp2

ddtwrm: SEI                     ; set up keyboard vector
        sta     bank3           ; ^^^^^^^^^ bank 3 ^^^^^^^^^
        jsr     sddtx           ; set brk vector

        ldx     #1
..rvec1 lda     vkeybd,x
        sta     vkeysv,x
        lda     keyld,x
        sta     vkeybd,x
         dex
         bpl    ..rvec1

        CLI
        jsr     save
        pla
        pla

        lda     #low quitddt
        sta     pcsav
        lda     #high quitddt
        jmp     ddtx1
        endproc



;       quit
;       ----
quitddt proc
        lda     #4              ; clear all bkps
..qclr  pha
        jsr     clbrk
        pla
        sec
        sbc     #1
        bpl     ..qclr

        jsr     usersc
        SEI
        lda     vkeysv
        sta     vkeybd
        lda     vkeysv+1
        sta     vkeybd+1
        CLI

        ldy     #$80-(runstk-xcii)      ; this # should be +
..zlod0 lda     (lomem),y
        sta     xcii-($80-(runstk-xcii)),y
         iny
         bpl    ..zlod0

;       ldy #$80
..zload lda     (lomem),y
        sta     0,y
         iny
         bne    ..zload

        jsr     cartrtn         ; some init
        jmp     warm            ; go to mac warm
        endproc



;       subroutine  ddt  (brk entry)
;       ----------------------------
ddtx    proc
        sta     bank3           ; ^^^^^^^^^ bank 3 ^^^^^^^^^
        pla                     ;restore acc
        plp                     ;and p reg
        jsr     save
        pla
        sbc     #2              ;correct for brk quirk
ddtx1a  sta     pcsav
        pla
        sbc     #0
ddtx1   sta     pcsav+1
        tsx
        stx     spsav
        CLI                     ;clear interrupt disable

        lda     #0
        sta     dscnfl          ;set flag to dbug screen off
        sta     windf
        sta     ssflg           ;clear ss auto flag
        jsr     clbrk           ;clear breakpoint 0
        jsr     pcsexa          ;set up window index
        jsr     sctog           ;turn on dbug screen
        sta     bank3           ; ^^^^^^^^^ bank 3 ^^^^^^^^^^^
        jmp     dbnd5
        endproc


mnctab  db      "BRADEAINAPHXPHY"
        db      "PLXPLYSTZSTZTRB"
        db      "TSBADCANDBITCMP"
        db      "EORJMPLDAORASBC"
        db      "STA"


gmnetab proc
        lda     vcount          ; so display doesn't jump
..ggm   cmp     vcount          ; tight timing!!!
        beq     ..ggm           ; not yet ready
                        ;   need 21 machine cycles
        pha
        pla
        pha
        pla
        pha
        pla
..gmn   sta     bank1           ; get chars quickly ^^^^^^^^^ bank 1 ^^^^^^^
        lda     mnetab+2,x
        pha
        lda     mnetab+1,x
        pha
        lda     mnetab,x
        sta     bank3           ; ^^^^^^^^^^^ bank 3 ^^^^^^^^^^
        jsr     shlet           ; show 1st
        pla
        jsr     shlet           ; show 2nd
        pla
        and     #$7f
                        ;  fall thru & return

;       show letter
;       -----------
; puts letter, in reverse video if overlayed by brk

shlet   jsr     intrep          ; get screen code
        ora     brkor
        sta     (dindir),y
        iny
        rts
        endproc


$eject
        seg     e_ddt
;============================================
;       Must align on an even boundary
;       ------------------------------
        db      $ff     ; this insures that high (symbol-1) on same page!
                        ;  INDX assumed to be first - because of $if directive

;       Addressing mode processors
;       --------------------------
indx    proc
        jsr     lpar
        jsr     nxby
        jsr     comx
        jmp     rpar
        endproc

indy    proc
        jsr     lpar
        jsr     nxby
        jsr     rpar
        jmp     comy
        endproc

imm     proc
        lda     #3              ; #
        sta     (dindir),y
        iny
        iny
        jsr     dollar
        lda     stemp
        jmp     shhex
        endproc

zx      proc
        lda     itemp
        cmp     #$96            ;ck if exception
        beq     zy
        cmp     #$b6
        beq     zy
        jsr     nxby
        jmp     comx
        endproc

zy      proc
        jsr     nxby
        jmp     comy
        endproc

ind     proc
        jsr     lpar
        jsr     nxad
        jmp     rpar
        endproc

absx    proc
        lda     itemp
        cmp     #$be
        beq     absy
        jsr     nxad
        jmp     comx
        endproc

absy    proc
        jsr     nxad
        jmp     comy
        endproc

rel     proc
        lda     stemp           ;now offset
        pha
        and     #$80
        bmi     ..relbk

        lda     #93             ; forward branch
        sta     (dindir),y
        ldy     #0
..rel0  pla
        clc
        adc     dndr2
        tax
        tya
        adc     dndr2+1
        tay
        txa
        clc
        adc     #2
        sta     stemp
        tya
        adc     #0
        ldy     #15
        jsr     shhex
        lda     stemp
        jmp     shhex

..relbk lda     #92             ; backward branch
        sta     (dindir),y
        ldy     #$ff
        bne     ..rel0
        endproc

acc     proc
        lda     #'A'-$20        ; A
        bne     setch

izpg    jsr     lpar
        jsr     nxby
        jmp     rpar

absix   jsr     lpar
        jsr     nxad
        jsr     comx
        jmp     rpar

z
nxby    jsr     dollar
        bne     nxby1

imp     rts

abs
nxad    jsr     dollar
        lda     stemp+1
        jsr     shhex
nxby1   lda     stemp
        jmp     shhex

        $if (nxad-indx) > 254
         $error "DDT addressing modes too long!"
        $endif


;       Output special chars
;       --------------------
lpar    proc
        lda     #'('-$20
        bne     setch
dollar  lda     #'$'-$20
        bne     setch
rpar    lda     #')'-$20
        bne     setch
com     lda     #','-$20
        bne     setch

comx    jsr     com
        lda     #'X'-$20
        bne     setch

comy    jsr     com
        lda     #'Y'-$20
setch   sta     (dindir),y
        iny
        rts
        endproc


$eject
        seg     c_ddt
;===============================================
;       Main (banked) portion of DDT code
;       ---------------------------------

;       this is the main active ddt loop
;       --------------------------------
dbnd5   proc
        jsr     setpu           ;clear the command buffer

        ldy     #14             ;num. of spaces to clear
        lda     #0
clrc1   sta     (dindir),y
        dey
        bpl     clrc1
        iny
        lda     #$80
        sta     (dindir),y
        sty     chcol

dbnd5a  lda     ssflg           ;ck if in auto mode
        beq     dbn56
        lda     brkkey          ;ck if break has been hit
        bne     dbn55
        inc     brkkey          ;turn off break indicator
dbn60   lda     #0
        sta     ssflg           ;ok, so turn off ss auto flag
        lda     crstat
        sta     crsinh          ;also restore cursor state
        lda     poksav
        ora     $7f
        and     pokmsk
        sta     pokmsk
        sta     irqen
        lda     dscnfl          ;set screen todbug
        bne     dbn56
        jsr     pcsexa
        jsr     sctog
        bne     dbn56
dbn55   jsr     gtnxin          ;ck next inst
        beq     dbn60
        jmp     sstep1          ;so, go do next instr.
dbn56   lda     #1              ;ck if start pressed
        bit     consol
        bne     dcon2
        jsr     deboun
        jmp     dcont


dcon2   lda     consol          ;ck if select
        and     #2
        bne     dcon1
        jsr     sctog           ;toggle screen
        jsr     deboun
        beq     dbnd5

dcon1   lda     #4
        bit     consol          ;ck if option
        bne     dcmdi
        jsr     deboun
        jmp     sstep

dcmdi   jsr     ckgetc          ;ck keypress
        bcs     dbnd5a
        ldx     #13             ;num of values to try
onli1   cmp     keytb,x         ;ck value
        beq     onli2
        dex                     ;if not it, get ready for next
        bpl     onli1
        jmp     dbnd5a

onli2   txa
        asl     a
        tay
        lda     keyru+1,y
        pha
        lda     keyru,y
        pha
        lda     cmdcd,x         ;get the internal rep.
        jsr     click           ;click
        jmp     putch           ;put char & call key routine
        endproc


dcont   proc
        jsr     usersc
        lda     #$c0
        sta     nmien           ;restore nmi s
        ldx     spsav           ;restore stack pointer
        txs
        lda     pcsav+1         ;now restore state
        pha
        lda     pcsav
        pha
        lda     psav
        pha
        ldx     xsav
        ldy     ysav
        lda     #83
        sta     ddtact          ;enable flash entry
        lda     asav
        rti
        endproc


keytb   db 42,61,21,40,58,15,13
        db 46,14,62,7,35,$2f
        db $25

cmdcd   db 37,39,34,50,36,93,41
        db 55,92,51,10,46,49
        db $2d


keyru   dw exam-1,go-1,stbrk-1,regis-1,depos-1
        dw pullwi-1,interp-1,chwin-1
        dw pushw-1,serch-1,pcset-1,next-1,quitddt-1
        dw move-1


;       flash (keyboard) entry
;       ----------------------
_nkybd  proc
        lda     kbcode          ;get keyboard character
        cmp     #dbchr
        bne     nkyb3           ;if not dbug entry char. continue
        lda     #$ff
        sta     chr
        lda     ddtact          ;ck if ddt already active
        cmp     #69             ;already there
        bne     nkyb4
        lda     #dbchr
        bne     nkyb3
nkyb4   pla
        plp
        jsr     save
        pla
        jmp     ddtx1a

nkyb3   jmp     (vkeysv)

keyld   dw nkybd
        endproc



;       save caller's status
;       (except stack register)
;       -----------------------
save    proc
        sta     asav            ;a
        php
        lda     #69
        sta     ddtact          ;disable flash entry
        pla
        sta     psav            ;p
        stx     xsav            ;x
        sty     ysav            ;y
        cld                     ;no decimal mode
        sec                     ;for pc manipulation
        rts                     ;return
        endproc


savad   db      low vvblki, low vvbkih, low vvblkd
        db      low vvbkdh, low sdmctl, low sdlstl, low sdlsth
        db      low gprior, low color1, low color2, low color4
        db      low chact,  low chbas


;       set display list
;       ----------------
setdl   proc
        ldx     windf
        lda     dlsttb,x
        sta     sdlstl
         lda    #high ddtlst
         sta    sdlstl+1
        rts
        endproc


dif     equ     dscd3-dscd2

septb   db 4,9,14,19,28
        db dif+4, dif+7, dif+10
        db dif+13, dif+16


;       set breakpoint
;       --------------
; in:
;       acc     = breakpoint #
;       btemp   = breakpoint loc.

setbp   proc
        jsr     times3
        lda     brkp0,x         ;ck if set
        ora     brkp0+1,x
        beq     ..setb2
        lda     btemp
        pha                     ;save temp
        lda     btemp+1
        pha
        lda     brkp0,x         ;if not, first restore
        sta     btemp
        lda     brkp0+1,x
        sta     btemp+1
        txa
        pha
        jsr     restor
        pla
        tax
        pla                     ;now restore btemp
        sta     btemp+1
        pla
        sta     btemp

..setb2 ldy     #0              ;set=0 for dindir
        lda     btemp+1
        sta     dindir+1
        sta     brkp0+1,x       ;set up pointer to breakpoint
        lda     btemp
        sta     dindir
        sta     brkp0,x
        lda     (dindir),y      ;get current value
        sta     brkp0+2,x       ;and save
        lda     #0
        sta     (dindir),y      ;set breakpoint

sddtx   lda     #low ddtx
        sta     vbreak
         lda    #high ddtx
         sta    vbreak+1
rste    rts
        endproc


;       clear breakpoint
;       ----------------
; in:
;       acc     = brkpt to clear

clbrk   proc
        jsr     times3
        lda     brkp0,x
        sta     btemp
        lda     brkp0+1,x
        sta     btemp+1

                        ;  restore breakpoint
restor  lda     btemp
        sta     dindir
        lda     btemp+1
        sta     dindir+1
        jsr     stbpp           ;set pointer
        bcs     rste            ;if not found, exit

rest6   lda     brkp0+2,x       ;get proper instruction
        ldy     #0              ;set up y for dindirect
        sta     (dindir),y      ;restore
        tya                     ;move 0 to acc
        sta     brkp0,x
        sta     brkp0+1,x       ;clear out breakpoint
        sta     brkp0+2,x
        rts
        endproc


;       set breakpoint pointer
;       ----------------------
stbpp   proc
        ldx     #0              ;get ready to restore instruction
..r0    lda     btemp           ;load a with low byte of brkp loc
        cmp     brkp0,x         ;this should equal one of the brkps
        beq     ..r1            ;is this it?
..r3    inx                     ; no. increment x and try again
        inx
        inx
        cpx     #15             ;ck if at end of table
        bne     ..r0
        sec                     ;set not found flag
        rts

                        ;  first byte was equal
..r1    lda     btemp+1         ;get high byte of pc
        cmp     brkp0+1,x       ;is this it?
        bne     ..r3            ;no, try again

        clc             ;  second byte equal
        rts
        endproc



;       multiply by 3 (put in X)
;       ------------------------
times3  proc
        sta     itbyte          ;save value temporarily
        asl     a               ;multiply *2
        adc     itbyte
        tax
        rts
        endproc


;       move pcsav to examwi
;       --------------------
pcsexa  proc
        lda     pcsav
        sta     examwi
         lda    pcsav+1
         sta    examwi+1
user1   rts
        endproc


;       set user screen
;       ---------------
usersc  proc
        lda     dscnfl          ;ck state of display
        beq     user1           ;toggle screen if necessary

                        ; toggles between ddt screen &
                        ;  the normal one
sctog   ldx     #0
        stx     nmien           ; zap all interrupts
        lda     dscnfl
        beq     dbsc1           ;ck what state the screen is in
         stx    dscnfl

        ldx     #12             ; restore graphics
..rsgr  ldy     savad,x
        lda     grsav,x
        sta     $0200,y
        dex
        bpl     ..rsgr

        $if 0                   ; not needed???
         lda    sdlstl
         sta    dlistl
          lda   sdlstl+1
          sta   dlistl+1
        $endif

        lda     sdmctl
;       sta     dmactl          ; not needed???
        and     #$0c
        lsr     a
        lsr     a
        sta     gractl
        lda     #$c0
        sta     nmien           ;allow dma
        rts

dbsc1   inc     dscnfl
        ldx     #13             ; save graphics
..svgp  ldy     savad-1,x
        lda     $0200,y
        sta     grsav-1,x
        dex
        bne     ..svgp
                                ;  x=0
        stx     gprior          ;set priority

                        ; clear and set up the screen
        txa                     ; set a=0
..cdsc  sta     dscd1,x
        sta     dscd1+(284-256),x
         dex
         bne    ..cdsc

                        ; now draw separator lines
        ldx     #10
        $if text=10
         stx    color1
        $endif

        lda     #124
..dli3  ldy     septb-1,x
        sta     dscd2,y
        dex
        bne     ..dli3          ; x=0

                        ; and set up dbug color regs
        $if bordr<>4
         lda    #bordr
         sta    color4
        $endif

        $if text<>10
         lda    #text
         sta    color1
        $endif

        $if bkgnd<>0
         lda    #bkgnd
         sta    color2
        $endif

        lda     #low kchact
        sta     chact
         lda    #high kchact
         sta    chact+1
        jsr     setdl
        ldx     #4
        lda     #0

        $if bordr=4
         stx    color4
        $endif

        $if bkgnd=0
         sta    color2
        $endif

        sta     gractl
..tog1  sta     grafp0,x
        dex
        bpl     ..tog1

        lda     #$21            ; narrow screen
        sta     sdmctl
;       sta     dmactl          ; not needed???

        lda     #low kvvbki
        sta     vvblki
         ldx    #high kvvbki
         stx    vvblki+1
        lda     #low kvvbkd
        sta     vvblkd

        $if high kvvbkd <> high kvvbki
         ldx    #high kvvbkd
        $endif

        stx     vvblkd+1
        lda     #$40
        sta     nmien
                        ;    and update dynamic parts
                        ;    show screen
sscrn   lda     #low dscd2
        sta     dindir          ;pt to screen brk area
         lda    #high dscd2
         sta    dindir+1

        ldx     #3
        lda     #0
        ldy     #29
shbk0   sta     (dindir),y
        iny
        dex
        bne     shbk0
        tay
shbk1   lda     brkp0+4,x       ;get first byte of address
        jsr     shhex
        lda     brkp0+3,x
        jsr     shhex
        inx
        inx
        inx
        iny
        cpy     #20
        bne     shbk1
        jsr     shstat          ;and register state

                        ;  and fall into dwind

;       display window
;       --------------
dwind   lda     examwi
;       sec
;       sbc     #2
        sta     dndr2           ;back window up 2
         lda    examwi+1
;        sbc    #0
         sta    dndr2+1

        lda     #low (dscd1+6)
        sta     dindir          ;pt to 1st data area
         lda    #high (dscd1+6)
         sta    dindir+1
        lda     #$1e
        sta     dscd1+4
        lda     windf           ;ck which filter is in
        beq     dwin1
         jmp    dhexw

dwin1   lda     #10
        sta     tbyte           ;bytes to show
dis1    lda     #0
        sta     mprot           ;bytes to protect initially
dis11   jsr     clin            ;clear line
        lda     dndr2+1         ;high byte of address loc
        jsr     shhex
        lda     dndr2
        jsr     shhex
        ldy     #2
        lda     (dndr2),y       ;now get 3 values
        sta     stemp+1
        dey
        lda     (dndr2),y
        sta     stemp
        dey
        lda     (dndr2),y
        sta     itemp
        ldy     #6
        jsr     shhex

                        ; now do the disassembly
        lda     mprot           ;check if byte is protected
        bne     nxins
        sta     undff           ;clear undefined flag
        sta     brkor           ;and brk display flag
        lda     itemp
        bne     dis2            ;make sure it's not a brk
        jsr     brins
        sta     itemp
        lda     #$80
        sta     brkor
        lda     itemp

dis2    jsr     detlen          ;get length of ins.
        sta     mprot
        ldy     #$09
        bcs     undef
        ldx     #$9e            ;try to find opcode pattern
patl1   lda     itemp           ;get opcode
        and     pattab,x        ;mask off don't cares
        cmp     pattab+1,x      ;see if it matches
        beq     patfd
        dex
        dex
        jmp     patl1

patfd   txa                     ;yep, this is it
        lsr     a               ;divide by 2
        cmp     #56
        bcs     pat65c
        jsr     times3
        jsr     gmnetab         ; get table bank
patjn   jsr     shwad           ;and the address mode
nxins   jsr     nxline
        inc     dndr2           ;and memory address
        bne     nxi2
        inc     dndr2+1
nxi2    dec     mprot
        dec     tbyte
        beq     sscr9
        jmp     dis11
sscr9   rts

undef   lda     #1
        sta     mprot
        ldx     #3
undf1   lda     #'?'
        jsr     shlet
        dex
        bne     undf1
        beq     nxins

pat65c  sbc     #56
        jsr     times3
        lda     mnctab,x        ;show mnem.
        jsr     shlet
        lda     mnctab+1,x
        jsr     shlet
        lda     mnctab+2,x
;       and     #$7f            ;strip off steve's bit 7
        jsr     shlet
        jmp     patjn

dhexw   lda     dndr2
        and     #3
        eor     #$ff
        clc
        adc     #5
        jsr     dhlin           ;show current line
        beq     dhxw0
         rts

dhxw0   lda     #9
        sta     stemp
dhxw1   jsr     nxline
        lda     #4              ;show 4 bytes from here on
        jsr     dhlin
        dec     stemp
        bne     dhxw1
        rts
        endproc


dhlin   proc
        sta     tbyte           ;byte to show
        sta     mprot
        jsr     clin
        lda     dndr2+1
        pha
        jsr     shhex
        lda     dndr2
        pha
        jsr     shhex
        ldy     #6
        ldx     #0
dhln1   lda     (dndr2,x)       ;this gets next value to show
        jsr     shhex
        iny
        inc     dndr2
        bne     dhln15
         inc    dndr2+1

dhln15  dec     tbyte
        bne     dhln1           ;if any more show
        lda     mprot
        sta     tbyte
        ldy     #18             ;now do atascii decoding
        pla
        sta     dndr2
        pla
        sta     dndr2+1
dhln2   lda     (dndr2,x)
        jsr     intrep
        sta     (dindir),y
        iny
        inc     dndr2
        bne     dhln25
         inc    dndr2+1

dhln25  dec     tbyte
        bne     dhln2
        rts
        endproc



;       next line
;       ---------
nxline  proc
        clc
        lda     dindir
        adc     #28
        sta     dindir
        bcc     nxl1
         inc    dindir+1
nxl1    rts
        endproc




;       clear display window line
;       -------------------------
clin    proc
        lda     #0
        tay
..clin1 sta     (dindir),y
        iny
        cpy     #26
        bne     ..clin1
        tay
        rts
        endproc


;       debounces the console
;       ---------------------
deboun  proc
        lda     rtclok
        clc
        adc     #$20
        sta     itbyte
..debos lda     #8
        sta     consol
        lda     consol
        cmp     #7
        beq     ..deb
         lda    rtclok
         cmp    itbyte
         bne    ..debos
        rts

..deb   lda     rtclok
        clc
        adc     #3
..db2   cmp     rtclok
        bne     ..db2
        lda     consol
        cmp     #7
        bne     deboun
        rts
        endproc


;       this routine retrieves the proper breakpoint instruction
;       --------------------------------------------------------
brins   proc
        jsr     dndbte
        jsr     stbpp
        bcc     bri2
         lda    #0
         rts
bri2    lda     brkp0+2,x
        rts
        endproc



;       determine length
;       ----------------
; this routine determines the length of the opcode passed in the acc

detlen  proc
        pha
        lsr     a               ; this computes proper table byte
        lsr     a
        tax
        pla
        and     #3              ; mask off doublet counter
        tay
        lda     lim,x           ;get proper byte from table
        cpy     #0              ;check if 0 doublet selected
        beq     detl2
detl1   lsr     a
        lsr     a               ;shift doublet to 0'th position
        dey                     ;decrement doublet ctr
        bne     detl1           ;loop if necessary
detl2   clc
        and     #3              ;mask off the 0'th doublet
        bne     detl3
        lda     #1
        sec
detl3   tay                     ;needed for sstep
        rts
        endproc


lim     db $09,$2a,$19,$3f
        db $2a,$2a,$1d,$3f
        db $0b,$2a,$19,$3f
        db $2a,$2a,$1d,$3f
        db $09,$28,$19,$3f
        db $2a,$28,$1d,$3c
        db $09,$2a,$19,$3f
        db $2a,$2a,$1d,$3f
        db $0a,$2a,$11,$3f
        db $2a,$2a,$1d,$3f
        db $2a,$2a,$19,$3f
        db $2a,$2a,$1d,$3f
        db $0a,$2a,$19,$3f
        db $2a,$28,$1d,$3c
        db $0a,$2a,$19,$3f
        db $2a,$28,$1d,$3c


;       move dndr2 to btemp
;       -------------------
dndbte  proc
        lda     dndr2
        sta     btemp
         lda    dndr2+1
         sta    btemp+1
        rts
        endproc


;       vsync
;       -----
vsync   proc
        lda     #$7b
..vsyn1 cmp     vcount
        bne     ..vsyn1
        rts
        endproc



;       internal representation
;       -----------------------
intrep  proc
        pha
        and     #$80
        sta     itbyte
        pla
        and     #$7f
        cmp     #32
        bmi     ..intr1
        cmp     #96
        bpl     inram
        sec
        sbc     #32
        bpl     inram
..intr1 ora     #64
inram   ora     itbyte
        rts
        endproc



;       show state
;       ----------
; this routine is used to
; show the values of the
; register states

shstat  proc
        lda     #low dscd3
        sta     dindir          ;ptr to dbug screen
         lda    #high dscd3
         sta    dindir+1
        ldy     #0
        lda     pcsav+1         ;now show values
        jsr     shhex
        lda     pcsav
        jsr     shhex
        iny
        ldx     #0
..shst  lda     asav,x
        jsr     shhex
        iny
        inx
        cpx     #4
        bne     ..shst

        lda     #low (dscd2+20)
        sta     dindir
         lda    #high (dscd2+20)
         sta    dindir+1

        ldy     #0
        lda     psav
        ldx     #8              ;set up counter for digits
..shb   asl     a               ;shift left to get bit in carry
        pha                     ;save temp
        lda     #16             ;load value for 0
        adc     #0              ;value is 17 for 1
        sta     (dindir),y
        iny
        pla
        dex
        bne     ..shb
        rts
        endproc



;       show hex
;       --------
; this routine is used to
;   show a value in the acc
;   in hex notation
; the routine assumes dindir
;   & y are set to point to
;   location for show

shhex   proc
        pha                     ;save temp
        lsr     a
        lsr     a
        lsr     a               ;get left nibble to right
        lsr     a
        jsr     hxdig
        pla
        and     #$0f            ;now mask off right nibble

                        ;   and fall thru into hxdig
hxdig   clc
        adc     #16             ;add offset for numbers
        cmp     #26             ;ck if > 10
        bcc     ..hx0
         clc
         adc    #7              ;add hex offset
..hx0   sta     (dindir),y
        iny
        rts
        endproc




;       show address mode
;       -----------------
shwad   proc
        lda     undff           ;check if instruction undefined
        beq     ..shad1
         rts

..shad1 ldx     #0
..shad2 lda     itemp           ;get op  code
        and     adpat,x         ;mask off dont cares
        inx
        cmp     adpat,x         ;does it match the mandatory part?
        beq     ..adfnd         ;yes
        inx
        inx
        bne     ..shad2         ;no, try next

..adfnd inx
        lda     adpat,x         ;get address mode
        tax
        lda     #high indx
        pha
        lda     adrou,x         ;get low byte
        pha
        iny
        rts                     ;go to mode routine
        endproc



pattab  db $ff,$20,$df,$4c,$e7,$c6,$e7,$e6,$e3,$a2
        db $e3,$a0,$e7,$86,$e7,$84,$f3,$e0,$f3,$c0
        db $f7,$24,$ff,$00,$ff,$18,$ff,$d8,$ff,$58
        db $ff,$b8,$ff,$ca,$ff,$88,$ff,$e8,$ff,$c8
        db $ff,$ea,$ff,$48,$ff,$08,$ff,$68,$ff,$28
        db $ff,$40,$ff,$60,$ff,$38,$ff,$f8,$ff,$78
        db $ff,$aa,$ff,$a8,$ff,$ba,$ff,$8a,$ff,$9a
        db $ff,$98,$ff,$90,$ff,$b0,$ff,$f0,$ff,$30
        db $ff,$d0,$ff,$10,$ff,$50,$ff,$70,$e3,$01
        db $e3,$21,$e3,$41,$e3,$61,$e3,$81,$e3,$a1
        db $e3,$c1,$e3,$e1,$e3,$02,$e3,$22,$e3,$42
        db $e3,$62

; ** start of 6502c patterns  **
        db $ff,$80,$ff,$3a,$ff,$1a,$ff,$da,$ff,$5a
        db $ff,$fa,$ff,$7a,$fd,$9c,$ef,$64,$f7,$14,$f7,$04

; and the funny address modes
        db $ff,$72,$ff,$32,$f7,$34,$ff,$d2,$ff,$52,$ff,$7c
        db $ff,$b2,$ff,$12,$ff,$f2,$ff,$92


;       Addressing mode handlers
;       ------------------------
adrou   db      low indx-1, low indy-1, low imm-1, low zx-1
        db      low zy-1,   low ind-1,  low acc-1, low rel-1
        db      low absx-1, low absy-1, low abs-1, low z-1
        db      low imp-1,  low izpg-1, low absix-1

adpat   db $1f,$01,$00, $1f,$11,$01
        db $1f,$09,$02, $ff,$a2,$02
        db $bf,$a0,$02, $ff,$c0,$02
        db $1f,$15,$03, $df,$94,$03
        db $1f,$16,$03, $df,$96,$04
        db $ff,$6c,$05, $9f,$0a,$06
        db $1f,$10,$07, $1f,$1d,$08
        db $1f,$1e,$08, $ff,$bc,$08
        db $1f,$19,$09, $ff,$be,$09
        db $1f,$0d,$0a, $1f,$0e,$0a
        db $1f,$0c,$0a, $ff,$20,$0a
        db $1f,$05,$0b, $1f,$06,$0b
        db $7f,$24,$0b, $bf,$84,$0b
        db $ff,$e4,$0b, $bf,$00,$0c
        db $ff,$60,$0c, $0f,$08,$0c
        db $8f,$8a,$0c

;  ** start of 6502c patterns  **
        db $ff,$80,$07,$ff,$3a,$06
        db $ff,$1a,$06,$ff,$da,$0c
        db $ff,$5a,$0c,$ff,$fa,$0c
        db $ff,$7a,$0c,$ff,$9c,$0a
        db $ff,$9e,$08,$ff,$64,$0b
        db $ff,$74,$03,$ff,$1c,$0a
        db $ff,$14,$0b,$ff,$0c,$0a
        db $ff,$04,$0b,$1f,$12,$0d
        db $ff,$7c,$0e,$ff,$3c,$08
        db $ff,$34,$03


;       get next instruction
;       --------------------
gtnxin  proc
        lda     pcsav
        sta     dndr2
         lda    pcsav+1
         sta    dndr2+1
        ldy     #0
        lda     (dndr2),y       ;get instruction
        rts
        endproc


;       single step
;       -----------
sstep   proc
        jsr     gtnxin
        bne     sstep1          ;make sure it's not a brk
        jsr     brins           ;otherwise, get it
        bne     sstep1
        jmp     ssrt2
sstep1  sta     itemp
        ldx     #14
        lda     itemp
tfor2   cmp     forb,x
        bne     nfnd            ;ck if op code in on forbidden list
        cpx     #9
        bcc     ssbc            ;simulate conditional branch
        lda     ssrouh-9,x      ;get address of routine
        pha
        lda     ssroul-9,x
        pha
        rts                     ;goto routine
        endproc

ssroul  db      low (ssjmp-1), low (ssjmi-1)
        db      low (ssjsr-1), low (ssrti-1), low (ssrts-1)
        db      low (ssjii-1)

ssrouh  db      high (ssjmp-1), high (ssjmi-1)
        db      high (ssjsr-1), high (ssrti-1), high (ssrts-1)
        db      high (ssjii-1)

ssbc    proc
        lda     pcsav
        adc     #2
        sta     pcsav
        bcc     ..ssbc
         inc    pcsav+1

..ssbc  jsr     ptbed
        lda     itemp
        sta     tbed            ;put in test bed
         lda    #4              ;offset to 2nd jmp ins.
         sta    tbed+1
        bne     sst2            ;unconditional
        endproc


forb    db      $90,$b0,$f0,$d0,$30,$10,$50,$70,$80
        db      $4c,$6c,$20,$40,$60,$7c

nfnd    proc
        dex
        bpl     tfor2
        jsr     ptbed           ; prepare test bed
        lda     itemp
        jsr     detlen          ;get length
        bcc     sst01
        inc     pcsav
        bne     ssrt2
         inc    pcsav+1
        bne     ssrt2           ;unconditional

sst01   tya
        tax
        dex
        dey
        clc
        adc     pcsav           ;set pc after instr.
        sta     pcsav
        bcc     ..sst1
         inc    pcsav+1

..sst1  lda     (dndr2),y       ;move instr. to test bed
        sta     tbed,x
         dex
         dey
         bpl    ..sst1
        lda     itemp           ;just in case it was a brk
        sta     tbed

sst2    ldx     spsav
        txs
        ldx     xsav            ;restore register state
        ldy     ysav
        lda     psav
        pha
        lda     asav
        plp
        jmp     tbed            ;execute instruction
        endproc

ssret   proc
        sta     asav            ;save register state again
        stx     xsav
        sty     ysav
        php
        pla
        sta     psav
        cld                     ;force binary mode
        cli
ssrt1   tsx
        stx     spsav
ssrt2   lda     dscnfl          ;ck if we need to update
        beq     ssrt3
        jsr     vsync
        jsr     shstat          ;update register state
        jsr     pcsexa          ;set memory window index
        lda     #0
        sta     windf
        jsr     vsync
        jsr     sscrn
ssrt3   jmp     dbnd5a          ;go back and check console
        endproc


;       Prepare test bed
;       ----------------
ptbed   proc
        ldx     #8
..ptb1  lda     tbedx,x
        sta     tbed,x
        dex
        bpl     ..ptb1
        rts
        endproc


ssjmp   proc
        jsr     pcad            ;set pc to next add
        jmp     ssrt2
        endproc

ssjmi   proc
        jsr     sssind
ssjm2   lda     (dindir),y
        sta     pcsav+1
        dey
        lda     (dindir),y
        sta     pcsav
        jmp     ssrt2
        endproc

ssjii   proc
        jsr     sssind
        lda     dindir
        clc
        adc     xsav
        sta     dindir
        bcc     ssjm2
         inc    dindir+1
        bcs     ssjm2           ; always
        endproc


pcad    proc
        ldy     #1
        lda     (dndr2),y
        sta     pcsav
         iny
         lda    (dndr2),y
         sta    pcsav+1
        rts
        endproc

ssrts   proc
        pla                     ;note: carry is set...
        adc     #0              ;correct for jsr weirdness
        sta     pcsav
        pla
        adc     #0
        sta     pcsav+1
        jmp     ssrt1
        endproc


ssjsr   proc
        jsr     pcad
        clc
        lda     dndr2
        adc     #2
        tay
         lda    dndr2+1
         adc    #0
         pha
        tya
        pha
        jmp     ssrt1
        endproc


ssbc2   proc
        cld
        cli
        ldy     #1
        lda     (dndr2),y
        bpl     ssbc3           ;this sets for for pos & neg offset
         dey
ssbc3   dey
        clc
        adc     pcsav
        sta     pcsav
         tya
         adc    pcsav+1
         sta    pcsav+1
        jmp     ssrt2
        endproc

ssrti   proc
        ldx     #-3
..ssrt  pla
        sta     psav+3,x
        inx
        bne     ..ssrt
        jmp     ssrt1
        endproc


;       set single step indirect
;       ------------------------
sssind  proc
        ldy     #2
        lda     (dndr2),y
        sta     dindir+1
         dey
         lda    (dndr2),y       ;set dindir to next ad
         sta    dindir
        rts
        endproc


getch   proc
        jsr     ckgetc
        bcs     getch
        ldx     #15
getc0   cmp     keycd,x         ;ck if in key code table
        beq     getc2
        dex
        bpl     getc0
        ldx     #4              ;ck if in delimiter table
getc3   cmp     delcd,x
        beq     getc4
        dex
        bpl     getc3
        cmp     #52             ;ck if delete
        bne     getch
        lda     #0              ;set delete indicator
        sec                     ;set  control flag
        bcs     getc7

getc2   txa
        clc
        adc     #16             ;convert to hex
        cmp     #26
        bcc     getc7
        clc
        adc     #7
        bne     getc7
getc4   sec                     ;set delimiter flag
getc7
                        ;  note fall through

;       click routine
;       -------------
click   php
        ldx     #$7e
        pha
..clik1 stx     consol
        lda     vcount
..clik2 cmp     vcount
        beq     ..clik2
        dex
        dex
        bpl     ..clik1
        pla
        plp
        rts
        endproc


;       check and get character
;       -----------------------
ckgetc  proc
        lda     chr
        cmp     #$ff
        bne     ckgt1
         rts

ckgt1   ldx     #$ff
        stx     chr
        rts
        endproc


;       constant tables
;       ---------------
keycd   db 50,31,30,26,24,29,27,51,53,48
;          0  1  2  3  4  5  6  7  8  9
        db 63,21,18,58,42,56
;          a  b  c  d  e  f

delcd   db 33,32,12,55,7
;          sp  , cr  >  *


;       putch
;       -----
putch   proc
        pha
        jsr     setpu
        ldy     chcol
        pla
        sta     (dindir),y
        iny
        lda     #$80
        sta     (dindir),y
        sty     chcol
        rts
        endproc


;       setpu
;       -----
setpu   proc
        lda     #low (dscd3+17)
        sta     dindir          ;set ptr to buf
         lda    #high (dscd3+17)
         sta    dindir+1
        rts
        endproc


;       pull memory window down
;       -----------------------
pullwi  proc
        lda     windf
        cmp     #1
        bne     ..pull1
        inc     examwi
        bne     pull3
         inc    examwi+1
        clc
        bcc     pull3

..pull1 jsr     einsiz
        clc
        adc     examwi
        sta     examwi
        bcc     pull3
         inc    examwi+1

pull3   jsr     dwind
        jmp     dbnd5
        endproc



;       pushw
;       -----
; this routine pushes the window up 1 byte

pushw   proc
        lda     examwi
        bne     ..push0
         dec    examwi+1
..push0 dec     examwi
        jmp     pull3
        endproc


;       exam window instruction size
;       ----------------------------
einsiz  proc
        jsr     exadnd
        ldy     #0
        lda     (dndr2),y
        bne     ..ein1
         jsr    brins
..ein1  jmp     detlen
        endproc



;       move examwi to dndr2
;       --------------------
exadnd  proc
        lda     examwi
        sta     dndr2
         lda    examwi+1
         sta    dndr2+1
        rts
        endproc



;       interp
;       ------
; turn the interpretive flag on

interp  proc
        inc     ssflg
        lda     #0
        sta     windf
        lda     crsinh
        sta     crstat
        lda     pokmsk
        sta     poksav
        ora     #$80
        sta     pokmsk
        sta     irqen
        jmp     dbnd5
        endproc


;       next
;       ----
next    proc
        jsr     einsiz          ;get size of inst in examwi
        clc
        adc     pcsav
        sta     btemp
         lda    pcsav+1
         adc    #0
         sta    btemp+1
        lda     #0              ;for internal bkp
        jsr     setbp
        jmp     dcont
        endproc



;       exam (examine)
;       --------------
exam    proc
        jsr     ogtadd          ;space over 1 and get add
        bcs     ..exam1         ;ck if null address
        lda     hexval
        sta     examwi          ;move address to exam window
         lda    hexval+1
         sta    examwi+1
        jsr     dwind           ;show window
..exam1 jmp     dbnd5
        endproc


;       getval (get value)
;       ------------------
getval  proc
        lda     #0
        sta     cc              ;set char. counter to 0
..gtv1  jsr     getch           ;get char.
        bcs     delim           ;if delimiter
        ldx     cc
        cpx     fsize
        beq     ..gtv1
        jsr     putch           ;put char on screen
        inc     cc              ;increment char. count
        bne     ..gtv1

delim   cmp     #0
        beq     ..deli2
        rts

..deli2 ldx     cc
        bne     ..deli3
        pla                     ;clear rts value
        pla
        pla
        pla
        jmp     dbnd5

..deli3 dec     cc
        ldy     chcol
        sta     (dindir),y
        dey
        lda     #$80
        sta     (dindir),y
        dec     chcol
        bne     ..gtv1
        endproc


;       gtadd (get address)
;       -------------------
gtadd   proc
        jsr     putch
        lda     #4
        sta     fsize           ;set field size
        jsr     getval          ;get address value
        jsr     convrt
        bcc     gtadx
        cmp     #$37            ;ck for >
        beq     gtaw            ;yes, branch
        cmp     #$07            ;ck for *
        bne     gtadx           ;no, exit (carry set)
        lda     pcsav           ;pc -> result
        sta     hexval
        lda     pcsav+1
        jmp     gtaw1

gtaw    lda     examwi          ;window pos'n -> result
        sta     hexval
        lda     examwi+1
gtaw1   sta     hexval+1
        clc
gtadx   rts
        endproc


;       convert atascii to hex
;       ----------------------
convrt  proc
        ldx     cc
        bne     ..conv2
         sec
         rts                    ;this flags null value
..conv2 lda     chcol
        sec
        sbc     cc
        tay
        jsr     setpu           ;set pointer
        lda     #0
        sta     hexval
        sta     hexval+1
conv4   lda     (dindir),y      ;get hex val
        cmp     #$20
        bcc     ..gth1
         clc
         adc    #9
..gth1  and     #$0f
        pha
        iny                     ;set y to next character
        ldx     #4
..rotv1 asl     hexval
        rol     hexval+1
        dex
        bne     ..rotv1

        pla
        ora     hexval
        sta     hexval
        dec     cc
        bne     conv4
        clc
        rts
        endproc



;       space over 1 and get add
;       ------------------------
ogtadd  proc
        lda     #0              ;space over 1 char.
        jmp     gtadd
        endproc


;       search routine
;       --------------
smtch   proc
        ldy     #0
..smt2  lda     sstr,y
        cmp     (dindir),y
        bne     ..smno
        iny
        cpy     sssiz
        bmi     ..smt2
        sec
        rts

..smno  clc
        rts
        endproc



go      proc
        jsr     ogtadd          ;get address value
        bcc     ..gg
         jmp    go1

..gg    lda     hexval
        sta     pcsav
         lda    hexval+1
         sta    pcsav+1
        jmp     dcont
        endproc


;       deposit
;       -------
depx.x  proc
        lda     #low sstr
        sta     dndr2
         lda    #high sstr
         sta    dndr2+1
        lda     #1
        bne     depxx

depos   jsr     exadnd          ;set up indirect pointer
        lda     #0
depxx   sta     depsw
        lsr     a               ; sets a=0
        jsr     putch           ;get hex string
        lda     #12
        sta     fsize
        jsr     getval
        lda     chcol
        sec
        sbc     cc
        sta     chcol
        lda     cc
        sta     tbyte
        lsr     a               ;compute string size
        adc     #0
        sta     sssiz

depo0   lda     tbyte
        sec
        sbc     #2
        bmi     depo2
        sta     tbyte
        lda     #2
        sta     cc
depo4   clc
        adc     chcol
        sta     chcol
        jsr     convrt
        ldy     #0
        sta     (dndr2),y
        inc     dndr2
        bne     depo45
         inc    dndr2+1
depo45  jmp     depo0

depo2   clc
        adc     #1
        bmi     depo3
        sta     tbyte
        lda     #1
        sta     cc
        bne     depo4

depo3   lda     depsw
        beq     depo5
        rts
        endproc



;       serch
;       -----
serch   proc
        lda     sssiz
        pha
        jsr     depx.x
        lda     sssiz
        bne     ..ser0
        pla                     ;get old val
        beq     go1
        sta     sssiz
        bne     ..ser05
..ser0   pla

..ser05 lda     examwi
        clc
        adc     #1
        sta     dindir
         lda    examwi+1
         adc    #0
         sta    dindir+1
..ser1  jsr     smtch
        bcs     ..sfnd
        inc     dindir
        bne     ..ser1
         inc    dindir+1
         lda    dindir+1
         cmp    #$d0
         bne    ..ser1
        beq     go1             ; go dbnd5

..sfnd  lda     dindir
        ldx     dindir+1
        sta     examwi
        stx     examwi+1
depo5   jsr     sscrn
go1     jmp     dbnd5
        endproc


;       move (from addr, to addr, length)
;       ---------------------------------
move    proc
        jsr     depx.x
        ldx     sssiz
        cpx     #6              ; must be 3 words!
        bne     stbr6           ; go "dbnd5"

        dex                     ; make low-high pairs
..lh    lda     sstr,x          ; get low
        ldy     sstr-1,x        ; get high
        sta     sstr-1,x        ; store low
        sty     sstr,x          ; store high
        dex                     ; point to next
        dex
        bpl     ..lh

$if 1
        ldx     sstr+5          ; CART version
        bmi     stbr6
..top   ldy     #0
        txa
        bne     ..dmv1
         ldy    sstr+4
         beq    depo5
..dmv1  sty     sstr+5
        jsr     aff0
         inc    sstr+1
         inc    sstr+3
         dex
        bpl     ..top
        bmi     depo5
$else
        lda     sstr+4          ; get length
        tay
        ora     sstr+5
        beq     stbr6           ; no zero lengths!
        ldx     sstr+5          ; get high
        inx
..dmv1  dey
        lda     (sstr),y
        sta     (sstr+2),y
        tya
        bne     ..dmv1
         inc    sstr+1
         inc    sstr+3
         dex
         bne    ..dmv1
        beq     depo5           ; all done
$endif
        endproc


;       stbrk (set breakpoint)
;       ----------------------
stbrk   proc
        lda     #0
        jsr     gtbyt           ;get value
        bcs     stbr6
        lda     hexval
        beq     stbr6
        cmp     #5
        bcs     stbr6
        pha                     ;save brkp #
        lda     #12             ;for ,
        jsr     gtadd
        bcc     stbr4
        pla
        jsr     clbrk           ;clear brkpt
        jmp     stbr5

stbr4   lda     hexval
        sta     btemp
         lda    hexval+1
         sta    btemp+1
        pla                     ;get brkp # back
        jsr     setbp
stbr5   jsr     sscrn
stbr6   jmp     dbnd5
        endproc



;       regis (register change)
;       -----------------------
regis   proc
        lda     #0
        sta     cc
        jsr     putch           ;space over
..reg1  jsr     ckgetc
        bcs     ..reg1

        ldx     #2
..reg11 cmp     delcd,x
        beq     ..reg5
         dex
         bpl    ..reg11

        cmp     #52             ;?delete
        beq     ..rdel
        tax
        lda     cc
        bne     ..reg1          ;already typed one char.

        txa
        ldx     #4
..reg2  cmp     regtb,x         ;ck if chac. in table
        beq     ..reg3
        dex
        bpl     ..reg2
        bmi     ..reg1

..reg3  stx     itbyte
        lda     regcd,x
        jsr     click
        jsr     putch
        inc     cc
        bne     ..reg1

..rdel  jsr     click
        lda     cc
        beq     stbr6
        dec     cc
        lda     #0
        ldy     chcol
        sta     (dindir),y
         dey
         lda    #$80
         sta    (dindir),y
        dec     chcol
        bne     ..reg1

..reg5  lda     cc
        beq     stbr6
        jsr     click
        lda     #12             ;load with value for ,
        jsr     gtbyt
        bcs     stbr6
        lda     hexval
        ldx     itbyte
        sta     asav,x
        jsr     shstat
        jmp     dbnd5
        endproc


;       set pc special for *
;       --------------------
pcset   proc
        jsr     ogtadd
        bcs     ..reg9
        lda     hexval
        sta     pcsav
         lda    hexval+1
         sta    pcsav+1

..reg9  jsr     shstat
        ldx     pcsav+1
        lda     pcsav
         sta    examwi
         stx    examwi+1
        lda     #0
        beq     scpc            ; change window!
        endproc


regtb   db 63,22,43,62,10
regcd   db 33,56,57,51,48


;       change window
;       -------------
chwin   proc
        lda     windf
        eor     #1
scpc    sta     windf
        jsr     setdl
        jmp     pull3           ;jsr dwind ...jmp dbnd5
        endproc


;       get byte
;       --------
gtbyt   proc
        jsr     putch
        lda     #2
        sta     fsize
        jsr     getval
        jmp     convrt
        endproc


;       test bed
;       --------
tbedx   proc
        nop
        nop
        nop
        jmp     ssret           ;back to ddt after execution
        jmp     ssbc2           ;back after conditional branch
        endproc


;       display list
;       ------------
dlsttb  db      low ddtlst, low ddtls1

ddtlst  proc
        db      $70,$70,$70
        db      66
        dw      tidis           ;diss. title
        db      1               ; this blanks 1 line
        dw      ..ddtl2

ddtls1  db      $70,$70,$70
        db      66
        dw      tihex           ;hex title
        db      0               ; this blanks 1 line

..ddtl2 db      66
        dw      ..ddtc
        db      66
        dw      dscd1           ;display line 2
        db      66
        dw      dscd1+(1*28)
        db      66
        dw      dscd1+(2*28)
        db      66
        dw      dscd1+(3*28)
        db      66
        dw      dscd1+(4*28)
        db      66
        dw      dscd1+(5*28)
        db      66
        dw      dscd1+(6*28)
        db      66
        dw      dscd1+(7*28)
        db      66
        dw      dscd1+(8*28)
        db      66
        dw      dscd1+(9*28)
        db      $40,66
        dw      dscc2           ;display line 17
        db      66
        dw      dscd2           ;display line 19
        db      $40,66
        dw      dscc3           ;display line 20
        db      66
        dw      dscd3           ;display line 21
        db      65
        dw      ddtlst          ;vertical retrace

;       DDT screen constant data
;       ------------------------
dscc2   db      $A2, $AB, $B0, $91, $FC         ; "BKP1|"
        db      $A2, $AB, $B0, $92, $FC         ; "BKP2|"
        db      $A2, $AB, $B0, $93, $FC         ; "BKP3|"
        db      $A2, $AB, $B0, $94, $FC         ; "BKP4|"
        db      $AE, $B6, $80, $A2, $A4, $A9    ; "NV BDI"
        db      $BA, $A3, $FC, $80, $80, $80    ; "ZC|   "

tidis   equ     $-3
        db      $80, $80
        db      $80, $AC, $AF, $A3, $8E, $80    ; " LOC. "
        db      $80, $B6, $A1, $AC, $80, $A9    ; " VAL I"
        db      $AE, $B3, $B4, $B2, $B5, $A3    ; "NSTRUC"
        db      $B4, $A9, $AF, $AE              ; "TION"

tihex
        db      $80, $80, $80, $80
        db      $80, $80, $AC, $AF, $A3, $8E    ; "  LOC."
        db      $80, $80, $80, $80, $80, $B6    ; "     V"
        db      $A1, $AC, $B5, $A5, $B3, $80    ; "ALUES "
        db      $80, $80, $A1, $B4, $A1, $8E    ; "  ATA."
        db      $80, $80, $80                   ; "   "

        db      $80

;       Dunion's copyright
;       ------------------
..ddtc  equ     $-2
        db      $A4, $A4, $B4, $80, $88, $A3    ; "DDT (C"
        db      $89, $80, $91, $99, $98, $94    ; ") 1984"
        db      $80, $AA, $A1, $AD, $A5, $B3    ; " JAMES"
        db      $80, $AA, $8E, $80, $A4, $B5    ; " J. DU"
        db      $AE, $A9, $AF, $AE, $80, $80    ; "NION  "

;       display line 21
;       ---------------
dscc3   equ     $-1

        db      $B0, $A3, $80, $FC, $A1, $80    ; "PC |A "
        db      $FC, $B8, $80, $FC, $B9, $80    ; "|X |Y "
        db      $FC, $B3, $80, $FC, $80, $A5    ; "|S | E"
        db      $AE, $B4, $A5, $B2, $80, $A3    ; "NTER C"
        db      $AF, $AD, $AD, $A1, $AE, $A4    ; "OMMAND"
        db      $80                             ; " "


$if 1
        ds      $52
aff0    proc
        ldy     #0
..lp    lda     (sstr),y        ; for CART version!!
        sta     (sstr+2),y
        iny
        cpy     sstr+5
        bne     ..lp
        rts
        endproc
$endif
